@@ -14,7 +14,7 @@ open OpamProcess.Job.Op
 open OpamStateTypes
 open Cmdliner
 
-type command = unit Cmdliner.Term.t * Cmdliner.Term.info
+type command = unit Cmdliner.Term.t * Cmdliner.Cmd.info
 
 let checked_repo_root () =
   let repo_root = OpamFilename.cwd () in
@@ -141,7 +141,8 @@ let cache_urls repo_root repo_def =
   repo_dl_cache @ global_dl_cache
 
 (* Downloads all urls of the given package to the given cache_dir *)
-let package_files_to_cache repo_root cache_dir cache_urls ?link (nv, prefix) =
+let package_files_to_cache repo_root cache_dir cache_urls
+    ~recheck ?link (nv, prefix) =
   match
     OpamFileTools.read_opam
       (OpamRepositoryPath.packages repo_root prefix nv)
@@ -153,24 +154,34 @@ let package_files_to_cache repo_root cache_dir cache_urls ?link (nv, prefix) =
         OpamPackage.to_string nv ^
         OpamStd.Option.to_string ((^) "/") name
       in
-      match OpamFile.URL.checksum urlf with
+      let checksums =
+        OpamHash.sort (OpamFile.URL.checksum urlf)
+      in
+      match checksums with
       | [] ->
         OpamConsole.warning "[%s] no checksum, not caching"
           (OpamConsole.colorise `green label);
         Done errors
-      | (first_checksum :: _) as checksums ->
-        OpamRepository.pull_file_to_cache label
-          ~cache_urls ~cache_dir
-          checksums
-          (OpamFile.URL.url urlf :: OpamFile.URL.mirrors urlf)
-        @@| fun r -> match OpamRepository.report_fetch_result nv r with
-        | Not_available (_,m) ->
+      | best_chks :: _ ->
+        let cache_file =
+          OpamRepository.cache_file cache_dir best_chks
+        in
+        let error_opt =
+          if not recheck && OpamFilename.exists cache_file then Done None
+          else
+            OpamRepository.pull_file_to_cache label
+              ~cache_urls ~cache_dir
+              checksums
+              (OpamFile.URL.url urlf :: OpamFile.URL.mirrors urlf)
+            @@| fun r -> match OpamRepository.report_fetch_result nv r with
+            | Not_available (_,m) -> Some m
+            | Up_to_date () | Result () -> None
+        in
+        error_opt @@| function
+        | Some m ->
           OpamPackage.Map.update nv (fun l -> m::l) [] errors
-        | Up_to_date () | Result () ->
+        | None ->
           OpamStd.Option.iter (fun link_dir ->
-              let target =
-                OpamRepository.cache_file cache_dir first_checksum
-              in
               let name =
                 OpamStd.Option.default
                   (OpamUrl.basename (OpamFile.URL.url urlf))
@@ -179,7 +190,7 @@ let package_files_to_cache repo_root cache_dir cache_urls ?link (nv, prefix) =
               let link =
                 OpamFilename.Op.(link_dir / OpamPackage.to_string nv // name)
               in
-              OpamFilename.link ~relative:true ~target ~link)
+              OpamFilename.link ~relative:true ~target:cache_file ~link)
             link;
           errors
     in
@@ -226,8 +237,15 @@ let cache_command cli =
       "JOBS" "Number of parallel downloads"
       OpamArg.positive_integer 8
   in
-  let cmd global_options cache_dir no_repo_update link jobs () =
+  let recheck_arg =
+    OpamArg.mk_flag ~cli OpamArg.(cli_from cli2_2) ["check-all"; "c"]
+      "Run a full integrity check on the existing cache. If this is not set, \
+       only missing cache files are handled."
+  in
+  let cmd global_options cache_dir no_repo_update link jobs recheck () =
     OpamArg.apply_global_options cli global_options;
+    (* this option was the default until 2.1 *)
+    let recheck = recheck || OpamCLIVersion.Op.(cli @< OpamArg.cli2_2) in
     let repo_root = checked_repo_root () in
     let repo_file = OpamRepositoryPath.repo repo_root in
     let repo_def = OpamFile.Repo.safe_read repo_file in
@@ -239,7 +257,8 @@ let cache_command cli =
       OpamParallel.reduce ~jobs
         ~nil:OpamPackage.Map.empty
         ~merge:(OpamPackage.Map.union (fun a _ -> a))
-        ~command:(package_files_to_cache repo_root cache_dir cache_urls ?link)
+        ~command:(package_files_to_cache repo_root cache_dir cache_urls
+                    ~recheck ?link)
         (List.sort (fun (nv1,_) (nv2,_) ->
              (* Some pseudo-randomisation to avoid downloading all files from
                 the same host simultaneously *)
@@ -274,7 +293,137 @@ let cache_command cli =
   in
   OpamArg.mk_command  ~cli OpamArg.cli_original command ~doc ~man
     Term.(const cmd $ global_options cli $
-          cache_dir_arg $ no_repo_update_arg $ link_arg $ jobs_arg)
+          cache_dir_arg $ no_repo_update_arg $ link_arg $ jobs_arg $
+          recheck_arg)
+
+let packages_with_prefixes repo_root packages =
+  let pkgs_map = OpamRepository.packages_with_prefixes repo_root in
+  if packages = [] then pkgs_map else
+  let pkgs_map, missing_pkgs =
+    List.fold_left (fun (map, missing) (n,vo)->
+        match vo with
+        | Some v ->
+          let nv = OpamPackage.create n v in
+          (match OpamPackage.Map.find_opt nv pkgs_map with
+           | Some pre ->( OpamPackage.Map.add nv pre map), missing
+           | None -> map, (n,vo)::missing)
+        | None ->
+          let n_map = OpamPackage.packages_of_name_map pkgs_map n in
+          if OpamPackage.Map.is_empty n_map then
+            map, (n,vo)::missing
+          else
+            (OpamPackage.Map.union (fun _nv _nv' ->
+                 OpamStd.Sys.exit_because `Internal_error)
+                n_map map),
+            missing
+      ) (OpamPackage.Map.empty, []) packages
+  in
+  if missing_pkgs <> [] then
+    OpamConsole.warning "Not found package%s %s. Ignoring them."
+      (match missing_pkgs with | [_] -> "" | _ -> "s")
+      (OpamStd.Format.pretty_list
+         (List.map
+            (fun (n,vo) ->
+               OpamConsole.colorise `underline
+                 (match vo with
+                  | Some v -> OpamPackage.to_string (OpamPackage.create n v)
+                  | None -> OpamPackage.Name.to_string n))
+            (List.rev missing_pkgs)));
+  pkgs_map
+
+let update_extrafiles_command_doc =
+  "Add extra-files to an opam repository."
+let update_extrafiles_command cli =
+  let command = "update-extrafiles" in
+  let doc = update_extrafiles_command_doc in
+  let man = [
+    `S Manpage.s_description;
+    `P "This command scans through package definitions, and add extrafiles \
+        field as requested."
+  ]
+  in
+  let hash_kinds = [`MD5; `SHA256; `SHA512] in
+  let hash_type_arg =
+    OpamArg.mk_opt ~cli OpamArg.(cli_from cli2_2) ["hash"]
+      "HASH_ALGO" "The hash, or hashes to be added"
+      Arg.(some (enum
+                   (List.map (fun k -> OpamHash.string_of_kind k, k)
+                      hash_kinds))) None
+  in
+  let packages =
+    OpamArg.mk_opt ~cli OpamArg.(cli_from cli2_2) ["p";"packages"]
+      "PACKAGES" "Only add extra files for the given packages"
+      Arg.(list OpamArg.package) []
+  in
+  let cmd global_options hash_type packages () =
+    OpamArg.apply_global_options cli global_options;
+    let repo_root = checked_repo_root () in
+    let pkg_prefixes = packages_with_prefixes repo_root packages in
+    let compute ?kind file =
+      OpamHash.compute ?kind (OpamFilename.to_string file)
+    in
+    let has_error =
+      OpamPackage.Map.fold (fun nv prefix has_error ->
+          let opam_file = OpamRepositoryPath.opam repo_root prefix nv in
+          let opam = OpamFile.OPAM.read opam_file in
+          let has_error =
+            if OpamFile.exists (OpamRepositoryPath.url repo_root prefix nv) then
+              (OpamConsole.warning "Not updating external URL file at %s"
+                 (OpamFile.to_string
+                    (OpamRepositoryPath.url repo_root prefix nv));
+               true)
+            else has_error
+          in
+          let files_dir = OpamRepositoryPath.files repo_root prefix nv in
+          if OpamFilename.exists_dir files_dir then
+            (let files =
+               OpamFilename.rec_files files_dir
+               |> List.map (fun file ->
+                   file,
+                   OpamFilename.Base.of_string
+                     (OpamFilename.remove_prefix files_dir file))
+             in
+             let extra_files =
+               OpamStd.Option.default [] (OpamFile.OPAM.extra_files opam)
+             in
+             let extra_files = OpamFilename.Base.Map.of_list extra_files in
+             let xfiles =
+               List.fold_left (fun extra (file, base) ->
+                   match OpamFilename.Base.Map.find_opt base extra_files with
+                   | Some hash ->
+                     let kind = OpamHash.kind hash in
+                     let xhash = compute ~kind file in
+                     if OpamHash.compare hash xhash <> 0 then
+                       OpamFilename.Base.Map.add base
+                         (OpamStd.Option.map_default (fun kind ->
+                              compute ~kind file) xhash hash_type)
+                         extra
+                     else
+                       (match hash_type with
+                        | Some kind ->
+                          OpamFilename.Base.Map.add base (compute ~kind file)
+                            extra
+                        | None -> extra)
+                   | None ->
+                     OpamFilename.Base.Map.add base
+                       (compute ?kind:hash_type file) extra)
+                 extra_files files
+             in
+             if not (OpamFilename.Base.Map.equal
+                       OpamHash.equal extra_files xfiles) then
+               let xfiles = OpamFilename.Base.Map.bindings xfiles in
+               let opam1 = OpamFile.OPAM.with_extra_files xfiles opam in
+               OpamFile.OPAM.write_with_preserved_format opam_file opam1);
+          has_error)
+        pkg_prefixes false
+    in
+    if has_error then OpamStd.Sys.exit_because `Sync_error
+    else OpamStd.Sys.exit_because `Success
+  in
+  OpamArg.mk_command  ~cli OpamArg.(cli_from cli2_2) command ~doc ~man
+    Term.(const cmd $ global_options cli
+          $ hash_type_arg $ packages)
+
 
 let add_hashes_command_doc =
   "Add archive hashes to an opam repository."
@@ -321,10 +470,10 @@ let add_hashes_command cli =
               let t_mapping = Hashtbl.create 187 in
               (OpamStd.Option.default [] (OpamFile.Lines.read_opt cache_file)
                |> List.iter @@ function
-                | [src; dst] ->
-                  Hashtbl.add t_mapping
-                    (OpamHash.of_string src) (OpamHash.of_string dst)
-                | _ -> failwith ("Bad cache at "^OpamFile.to_string cache_file));
+               | [src; dst] ->
+                 Hashtbl.add t_mapping
+                   (OpamHash.of_string src) (OpamHash.of_string dst)
+               | _ -> failwith ("Bad cache at "^OpamFile.to_string cache_file));
               Hashtbl.add t (k1,k2) (cache_file, t_mapping);
             ))
           hash_kinds
@@ -396,38 +545,7 @@ let add_hashes_command cli =
       cache_urls repo_root
         (OpamFile.Repo.safe_read (OpamRepositoryPath.repo repo_root))
     in
-    let pkg_prefixes =
-      let pkgs_map = OpamRepository.packages_with_prefixes repo_root in
-      if packages = [] then pkgs_map
-      else
-        (let pkgs_map, missing_pkgs =
-           List.fold_left (fun ((map: string option OpamPackage.Map.t),error)  (n,vo)->
-               match vo with
-               | Some v ->
-                 let nv = OpamPackage.create n v in
-                 (match OpamPackage.Map.find_opt nv pkgs_map with
-                  | Some pre ->( OpamPackage.Map.add nv pre map), error
-                  | None -> map, (n,vo)::error)
-               | None ->
-                 let n_map = OpamPackage.packages_of_name_map pkgs_map n in
-                 if OpamPackage.Map.is_empty n_map then
-                   map, (n,vo)::error
-                 else
-                   (OpamPackage.Map.union (fun _nv _nv' -> assert false) n_map map),
-                   error
-             ) (OpamPackage.Map.empty, []) packages
-         in
-         if missing_pkgs <> [] then
-           OpamConsole.warning "Not found package%s %s. Ignoring them."
-             (if List.length missing_pkgs = 1 then "" else "s")
-             (OpamStd.List.concat_map ~left:"" ~right:"" ~last_sep:" and " ", "
-                (fun (n,vo) ->
-                   OpamConsole.colorise `underline
-                     (match vo with
-                      | Some v -> OpamPackage.to_string (OpamPackage.create n v)
-                      | None -> OpamPackage.Name.to_string n)) missing_pkgs);
-         pkgs_map)
-    in
+    let pkg_prefixes = packages_with_prefixes repo_root packages in
     let has_error =
       OpamPackage.Map.fold (fun nv prefix has_error ->
           let opam_file = OpamRepositoryPath.opam repo_root prefix nv in
@@ -440,30 +558,30 @@ let add_hashes_command cli =
             else has_error
           in
           let process_url has_error urlf =
-              let hashes = OpamFile.URL.checksum urlf in
-              let hashes =
-                if replace then
-                  List.filter (fun h -> List.mem (OpamHash.kind h) hash_types)
-                    hashes
-                else hashes
-              in
-              let has_error, hashes =
-                List.fold_left (fun (has_error, hashes) kind ->
-                    if List.exists (fun h -> OpamHash.kind h = kind) hashes
-                    then has_error, hashes else
-                    match get_hash cache_urls kind hashes
-                            (OpamFile.URL.url urlf)
-                    with
-                    | Some h -> has_error, hashes @ [h]
-                    | None ->
-                      OpamConsole.error "Could not get hash for %s: %s"
-                        (OpamPackage.to_string nv)
-                        (OpamUrl.to_string (OpamFile.URL.url urlf));
-                      true, hashes)
-                  (has_error, hashes)
-                  hash_types
-              in
-              has_error, OpamFile.URL.with_checksum hashes urlf
+            let hashes = OpamFile.URL.checksum urlf in
+            let hashes =
+              if replace then
+                List.filter (fun h -> List.mem (OpamHash.kind h) hash_types)
+                  hashes
+              else hashes
+            in
+            let has_error, hashes =
+              List.fold_left (fun (has_error, hashes) kind ->
+                  if List.exists (fun h -> OpamHash.kind h = kind) hashes
+                  then has_error, hashes else
+                  match get_hash cache_urls kind hashes
+                          (OpamFile.URL.url urlf)
+                  with
+                  | Some h -> has_error, hashes @ [h]
+                  | None ->
+                    OpamConsole.error "Could not get hash for %s: %s"
+                      (OpamPackage.to_string nv)
+                      (OpamUrl.to_string (OpamFile.URL.url urlf));
+                    true, hashes)
+                (has_error, hashes)
+                hash_types
+            in
+            has_error, OpamFile.URL.with_checksum hashes urlf
           in
           let has_error, url_opt =
             match OpamFile.OPAM.url opam with
@@ -720,7 +838,7 @@ let pattern_list_arg =
 
 let env_arg cli =
   OpamArg.mk_opt ~cli OpamArg.cli_original ["environment"]
-    "VAR=VALUE[;VAR=VALUE]"
+    "VAR=VALUE[,VAR=VALUE]"
     (Printf.sprintf
        "Use the given opam environment, in the form of a list of \
         comma-separated 'var=value' bindings, when resolving variables. This \
@@ -775,6 +893,7 @@ let get_virtual_switch_state repo_root env =
         load ~lock_kind:`Lock_read !r.root_dir) +!
                                 OpamFile.Config.empty);
     global_variables = OpamVariable.Map.empty;
+    global_state_to_upgrade = { gtc_repo = false; gtc_switch = false; };
   } in
   let singl x = OpamRepositoryName.Map.singleton repo.repo_name x in
   let repos_tmp =
@@ -985,10 +1104,15 @@ let add_constraint_command cli =
        $(b,<2)). The default in this case is to print a warning and keep \
        the existing constraint unchanged."
   in
-  let cmd global_options force atom () =
+  let packages =
+    OpamArg.mk_opt ~cli OpamArg.(cli_from cli2_2) ["p";"packages"]
+      "PACKAGES" "Only add constraints for the given packages"
+      Arg.(list OpamArg.package) []
+  in
+  let cmd global_options force atom packages () =
     OpamArg.apply_global_options cli global_options;
     let repo_root = checked_repo_root () in
-    let pkg_prefixes = OpamRepository.packages_with_prefixes repo_root in
+    let pkg_prefixes = packages_with_prefixes repo_root packages in
     let name, cstr_opt = atom in
     let cstr = match cstr_opt with
       | Some (relop, v) ->
@@ -1057,7 +1181,7 @@ let add_constraint_command cli =
       pkg_prefixes
   in
   OpamArg.mk_command  ~cli OpamArg.cli_original command ~doc ~man
-  Term.(pure cmd $ global_options cli $ force_arg $ atom_arg)
+  Term.(const cmd $ global_options cli $ force_arg $ atom_arg $ packages)
 
 (* HELP *)
 let help =
@@ -1082,8 +1206,8 @@ let help =
           List.iter (OpamConsole.msg "%s\n") cmds; `Ok ()
       | `Ok t -> `Help (man_format, Some t) in
 
-  Term.(ret (const help $Term.man_format $Term.choice_names $topic)),
-  Term.info "help" ~doc ~man
+  Term.(ret (const help $Arg.man_format $Term.choice_names $topic)),
+  Cmd.info "help" ~doc ~man
 
 let admin_subcommands cli =
   let index_command = index_command cli in
@@ -1097,6 +1221,7 @@ let admin_subcommands cli =
     filter_command cli;
     add_constraint_command cli;
     add_hashes_command cli;
+    update_extrafiles_command cli;
     help;
   ]
 
@@ -1126,7 +1251,7 @@ let default_subcommand cli =
       upgrade_command_doc
   in
   Term.(const usage $ global_options cli),
-  Term.info "opam admin"
+  Cmd.info "opam admin"
     ~version:(OpamVersion.to_string OpamVersion.current)
     ~sdocs:OpamArg.global_option_section
     ~doc:admin_command_doc
