@@ -54,7 +54,7 @@ let preinit_environment_variables =
       "see options `--debug' and `--debug-level'.";
       "YES", (fun v -> YES (env_bool v)),
       "see options `--yes' and `--confirm-level`. \
-       $(b,OPAMYES) has has priority over $(b,OPAMNO) and is \
+       $(b,OPAMYES) has priority over $(b,OPAMNO) and is \
        ignored if $(b,OPAMCONFIRMLEVEL) is set.";
     ] in
   let client =
@@ -122,7 +122,7 @@ let environment_variables =
       "STATUSLINE", cli_original, (fun v -> STATUSLINE (env_when v)),
       ("display a dynamic status line showing what's currently going on on \
         the terminal. (one of "^string_of_enum when_enum^")");
-      "USEOPENSSL", cli_original, (fun v -> USEOPENSSL (env_bool v)),
+      "USEOPENSSL", cli_between cli2_0 cli2_2, (fun _v -> OpamStd.Config.E.REMOVED),
       "force openssl use for hash computing.";
       "UTF8", cli_original, (fun v -> UTF8 (env_when_ext v)),
       (Printf.sprintf "use UTF8 characters in output (one of %s). By default \
@@ -130,7 +130,7 @@ let environment_variables =
          (string_of_enum when_enum));
       "UTF8MSGS", cli_original, (fun v -> UTF8MSGS (env_bool v)),
       "use extended UTF8 characters (camels) in opam messages. Implies \
-       $(i,OPAMUTF8). This is set by default on OSX only.";
+       $(i,OPAMUTF8). This is set by default on macOS only.";
       "VERBOSE", cli_original, (fun v -> VERBOSE (env_level v)),
       "see option `--verbose'.";
     ] in
@@ -167,7 +167,7 @@ let environment_variables =
       "EXTERNALSOLVER", cli_original, (fun v -> EXTERNALSOLVER (env_string v)),
       "see option `--solver'.";
       "FIXUPCRITERIA", cli_original, (fun v -> FIXUPCRITERIA (env_string v)),
-      "same as $(i,OPAMUPGRADECRITERIA), but specific to fixup.";
+      "same as $(i,OPAMUPGRADECRITERIA), but specific to fixup and reinstall.";
       "NOASPCUD", cli_original, (fun v -> NOASPCUD (env_bool v)),
       "Deprecated.";
       "PREPRO", cli_original, (fun v -> PREPRO (env_bool v)),
@@ -209,6 +209,12 @@ let environment_variables =
        'download-command' value from the main config file.";
       "NOCHECKSUMS", cli_original, (fun v -> NOCHECKSUMS (env_bool v)),
       "enables option --no-checksums when available.";
+      "REPOSITORYTARRING", cli_from cli2_2,
+      (fun b -> REPOSITORYTARRING (env_bool b)),
+      "internally store the repositories as tar.gz files. This can be much \
+       faster on filesystems that don't cope well with scanning large trees \
+       but have good caching in /tmp. However this is slower in the \
+       general case.";
       "REQUIRECHECKSUMS", cli_original,
       (fun v -> REQUIRECHECKSUMS (env_bool v)),
       "Enables option `--require-checksums' when available \
@@ -250,6 +256,8 @@ let environment_variables =
        by `opam env --switch=SWITCH --set-switch'.";
       "UNLOCKBASE", cli_original, (fun v -> UNLOCKBASE (env_bool v)),
       "see install option `--unlock-base'.";
+      "WITHDEVSETUP", cli_from cli2_2, (fun v -> WITHDEVSETUP (env_bool v)),
+      "see install option `--with-dev-setup'.";
       "WITHDOC", cli_original, (fun v -> WITHDOC (env_bool v)),
       "see install option `--with-doc'.";
       "WITHTEST", cli_original, (fun v -> WITHTEST (env_bool v)),
@@ -409,8 +417,10 @@ let help_sections cli =
     `P (Printf.sprintf "See https://opam.ocaml.org/doc.");
 
     `S Manpage.s_authors;
+    `P "David Allsopp      <david@tarides.com>"; `Noblank;
     `P "Vincent Bernardoff <vb@luminar.eu.org>"; `Noblank;
     `P "Raja Boujbel       <raja.boujbel@ocamlpro.com>"; `Noblank;
+    `P "Kate Deplaix       <kit-ty-kate@outlook.com>"; `Noblank;
     `P "Roberto Di Cosmo   <roberto@dicosmo.org>"; `Noblank;
     `P "Thomas Gazagnaire  <thomas@gazagnaire.org>"; `Noblank;
     `P "Louis Gesbert      <louis.gesbert@ocamlpro.com>"; `Noblank;
@@ -504,7 +514,9 @@ let apply_global_options cli o =
     (* ?skip_version_checks:bool *)
     (* ?all_parens:bool *)
     (* - core options - *)
-    ?debug_level:(if o.safe_mode then Some 0 else o.debug_level)
+    ?debug_level:(match o.debug_level with
+        | Some x when x < 0 -> o.debug_level
+        | _ -> if o.safe_mode then Some 0 else o.debug_level)
     ?verbose_level:(if o.quiet then Some 0 else
                     if o.verbose = 0 then None else Some o.verbose)
     ?color:o.color
@@ -513,6 +525,7 @@ let apply_global_options cli o =
     ?confirm_level:o.confirm_level
     ?yes
     ?safe_mode:(flag o.safe_mode)
+    (* ?cygbin:string option *)
     (* ?lock_retries:int *)
     (* ?log_dir:OpamTypes.dirname *)
     (* ?keep_log_dir:bool *)
@@ -530,6 +543,7 @@ let apply_global_options cli o =
     (* ?solver_preferences_best_effort_prefix: *)
     (* - state options - *)
     ?root_dir:o.opt_root
+    ?original_root_dir:o.opt_root
     ?current_switch:(o.opt_switch >>| OpamSwitch.of_string)
     ?switch_from:(o.opt_switch >>| fun _ -> `Command_line)
     (* ?jobs: int *)
@@ -560,8 +574,55 @@ let apply_global_options cli o =
     OpamJson.append "opam-version" (`String OpamVersion.(to_string (full ())));
     OpamJson.append "command-line"
       (`A (List.map (fun s -> `String s) (Array.to_list Sys.argv)))
-  )
-
+  );
+  (* We need to retrieve very early cygwin root path to set up 'cygbin' config
+     field and git binary path.
+     It is retrieved from config file, and we use a low level reading of
+     that file instead of OpamStateConfig.safe_load to avoid multiple error
+     messages displayed if an error is detected in the config file. If there is
+     an error, or best effort notification, it will be highlighted after
+     anyway. *)
+  try
+    let opamfile =
+      OpamPath.config OpamStateConfig.(!r.root_dir)
+      |> OpamFile.to_string
+      |> OpamParser.FullPos.file
+    in
+    List.iter OpamParserTypes.FullPos.(function
+        | { pelem = Variable ({ pelem = "sys-pkg-manager-cmd"; _},
+                              {pelem = List { pelem = elements; _}; _}); _} ->
+          let rec aux last elements =
+            match last, elements with
+            | _, [] -> ()
+            | Some { pelem = String "cygwin"; _},
+              { pelem = String cygcheck; _}::_  ->
+              let cygbin = Filename.dirname cygcheck in
+              OpamCoreConfig.update ~cygbin ()
+            | _, element::elements -> aux (Some element) elements
+          in
+          aux None elements
+        | { pelem = Variable ({ pelem = "global-variables"; _},
+                              {pelem = List { pelem = elements; _}; _}); _} ->
+          let rec aux last elements =
+            match last, elements with
+            | _, [] -> ()
+            | Some { pelem = Ident "os-distribution"; _},
+              { pelem = String "msys2"; _}::_  ->
+              let cygbin =
+                OpamStd.Option.map Filename.dirname
+                  (OpamSystem.resolve_command "cygcheck")
+              in
+              OpamCoreConfig.update ?cygbin ()
+            | _, element::elements -> aux (Some element) elements
+          in
+          aux None elements
+        | { pelem = Variable ({ pelem = "git-location"; _},
+                              {pelem = String git_location; _}); _} ->
+          OpamCoreConfig.update ~git_location ()
+        | _ -> ())
+      opamfile.file_contents
+  with
+  | Sys_error _ | Not_found -> ()
 
 (** Build options *)
 
@@ -574,6 +635,7 @@ type build_options = {
   req_checksums : bool;
   build_test    : bool;
   build_doc     : bool;
+  dev_setup       : bool;
   show          : bool;
   dryrun        : bool;
   fake          : bool;
@@ -589,14 +651,14 @@ type build_options = {
 
 let create_build_options
     keep_build_dir reuse_build_dir inplace_build make no_checksums
-    req_checksums build_test build_doc show dryrun skip_update
+    req_checksums build_test build_doc dev_setup show dryrun skip_update
     fake jobs ignore_constraints_on unlock_base locked lock_suffix
     assume_depexts no_depexts
     =
   {
     keep_build_dir; reuse_build_dir; inplace_build; make; no_checksums;
-    req_checksums; build_test; build_doc; show; dryrun; skip_update; fake;
-    jobs; ignore_constraints_on; unlock_base; locked; lock_suffix;
+    req_checksums; build_test; build_doc; dev_setup; show; dryrun; skip_update;
+    fake; jobs; ignore_constraints_on; unlock_base; locked; lock_suffix;
     assume_depexts; no_depexts;
   }
 
@@ -617,6 +679,7 @@ let apply_build_options cli b =
     (* ?no_base_packages:(flag o.no_base_packages) -- handled globally *)
     ?build_test:(flag b.build_test)
     ?build_doc:(flag b.build_doc)
+    ?dev_setup:(flag b.dev_setup)
     ?dryrun:(flag b.dryrun)
     ?makecmd:(b.make >>| fun m -> lazy m)
     ?ignore_constraints_on:
@@ -697,8 +760,12 @@ let existing_filename_dirname_or_dash =
   in
   parse, print
 
-let _subpath_conv =
-  (fun str -> `Ok (OpamStd.String.remove_prefix ~prefix:"./" str)), pr_str
+let subpath_conv =
+  let parse str =
+    `Ok (OpamFilename.SubPath.of_string str)
+  in
+  let print ppf sb = pr_str ppf (OpamFilename.SubPath.to_string sb) in
+  parse, print
 
 let package_name =
   let parse str =
@@ -716,7 +783,7 @@ let package_version =
   let print ppf ver = pr_str ppf (OpamPackage.Version.to_string ver) in
   parse, print
 
-let positive_integer : int Arg.converter =
+let positive_integer : int Arg.conv =
   let (parser, printer) = Arg.int in
   let parser s =
     match parser s with
@@ -843,13 +910,13 @@ let variable_bindings =
 
 let warn_selector =
   let parse str =
-    let sep = Re.(compile (set "+-")) in
+    let sep = Re.(compile (set "+-@")) in
     let sel = Re.(compile @@
                   seq [bos; group (rep1 digit);
                        opt @@ seq [str ".."; group (rep1 digit)];
                        eos]) in
     let rec seq i j =
-      if i = j then [i]
+      if Int.equal i j then [i]
       else if i < j then i :: seq (i+1) j
       else j :: seq (j+1) i
     in
@@ -861,8 +928,13 @@ let warn_selector =
           try seq i (int_of_string (Re.Group.get g 2))
           with Not_found -> [i]
         in
-        let enabled = Re.Group.get d 0 = "+" in
-        let acc = List.fold_left (fun acc n -> (n, enabled) :: acc) acc nums in
+        let state = match Re.Group.get d 0 with
+          | "+" -> `Enable
+          | "-" -> `Disable
+          | "@" -> `EnableError
+          | _ -> assert false
+        in
+        let acc = List.fold_left (fun acc n -> (n, state) :: acc) acc nums in
         aux acc r
       | [] -> acc
       | _ -> raise Not_found
@@ -873,8 +945,13 @@ let warn_selector =
   in
   let print ppf warns =
     pr_str ppf @@
-    OpamStd.List.concat_map "" (fun (num,enable) ->
-        Printf.sprintf "%c%d" (if enable then '+' else '-') num)
+    OpamStd.List.concat_map "" (fun (num,state) ->
+        let state = match state with
+          | `Enable -> '+'
+          | `Disable -> '-'
+          | `EnableError -> '@'
+        in
+        Printf.sprintf "%c%d" state num)
       warns
   in
   parse, print
@@ -974,9 +1051,33 @@ let opamlist_columns =
   in
   parse, print
 
+let mk_flag ~cli validity ?(section=Manpage.s_options) flags name =
+  mk_flag ~cli validity ~section flags name
+
+let mk_flag_replaced ~cli ?(section=Manpage.s_options) flags doc =
+ mk_flag_replaced ~cli ~section flags doc
+
+let mk_opt ~cli validity ?(section=Manpage.s_options) ?vopt flags value doc
+    kind default =
+  mk_opt ~cli validity ~section ?vopt flags value doc kind default
+
+let mk_opt_all ~cli validity ?(section=Manpage.s_options) ?vopt ?default flags
+    value doc kind =
+  mk_opt_all ~cli validity ~section ?vopt ?default flags value doc kind
+
+let mk_vflag ~cli ?(section=Manpage.s_options) default flags =
+  mk_vflag ~cli ~section default flags
+
+let mk_vflag_all ~cli ?(section=Manpage.s_options) ?default flags =
+  mk_vflag_all ~cli ~section ?default flags
+
+let mk_enum_opt ~cli validity ?(section=Manpage.s_options) flags value states
+    doc =
+  mk_enum_opt ~cli validity ~section flags value states doc
+
 let term_info ~cli title ~doc ~man =
   let man = man @ help_sections cli in
-  Term.info ~sdocs:global_option_section ~docs:Manpage.s_commands
+  Cmd.info ~sdocs:global_option_section ~docs:Manpage.s_commands
     ~doc ~man title
 
 let mk_command ~cli validity name ~doc ~man =
@@ -987,7 +1088,9 @@ let mk_command_ret ~cli validity name ~doc ~man =
 
 let make_command_alias ~cli cmd ?(options="") name =
   let term, info = cmd in
-  let orig = Term.name info in
+  (* TODO: Remove that horror whenever https://github.com/dbuenzli/cmdliner/pull/161 is merged *)
+  let name_of_new_api_info info = Cmd.name (Cmd.v info (Term.const ())) in
+  let orig = name_of_new_api_info info in
   let doc = Printf.sprintf "An alias for $(b,%s%s)." orig options in
   let man = [
     `S Manpage.s_description;
@@ -999,7 +1102,7 @@ let make_command_alias ~cli cmd ?(options="") name =
   ] @ help_sections cli
   in
   term,
-  Term.info name
+  Cmd.info name
     ~docs:"COMMAND ALIASES" ~sdocs:global_option_section
     ~doc ~man
 
@@ -1014,28 +1117,31 @@ let nonempty_arg_list name doc kind =
 
 (** Common flags *)
 
-let print_short_flag cli validity =
-  mk_flag ~cli validity ["s";"short"]
+let print_short_flag ?section cli validity =
+  mk_flag ~cli validity ?section ["s";"short"]
     "Output raw lists of names, one per line, skipping any details."
 
-let shell_opt cli validity =
+let shell_opt ?section cli validity =
   let enum = [
-    "bash",SH_bash;
-    "sh",SH_sh;
-    "csh",SH_csh;
-    "zsh",SH_zsh;
-    "fish",SH_fish;
-  ] |> List.map (fun (s,v) -> cli_original, s, v)
+    None,"bash",SH_bash;
+    None,"sh",SH_sh;
+    None,"csh",SH_csh;
+    None,"zsh",SH_zsh;
+    None,"fish",SH_fish;
+    Some cli2_2,"pwsh",SH_pwsh Powershell_pwsh;
+    Some cli2_2,"cmd",SH_cmd;
+    Some cli2_2,"powershell",SH_pwsh Powershell
+  ] |> List.map (fun (c,s,v) -> OpamStd.Option.map_default cli_from cli_original c, s, v)
   in
-  mk_enum_opt ~cli validity ["shell"] "SHELL" enum
+  mk_enum_opt ~cli validity ?section ["shell"] "SHELL" enum
     (Printf.sprintf
        "Sets the configuration mode for opam environment appropriate for \
         $(docv). One of %s. Guessed from the parent processes and the \\$SHELL \
         variable by default."
        (string_of_enum enum))
 
-let dot_profile_flag cli validity =
-  mk_opt ~cli validity ["dot-profile"]
+let dot_profile_flag ?section cli validity =
+  mk_opt ~cli validity ?section ["dot-profile"]
     "FILENAME"
     (Printf.sprintf
       "Name of the configuration file to update instead of \
@@ -1043,7 +1149,7 @@ let dot_profile_flag cli validity =
       dir_sep dir_sep)
     (Arg.some filename) None
 
-let repo_kind_flag cli validity =
+let repo_kind_flag ?section cli validity =
   let main_kinds = [
     "http" , `http;
     "local", `rsync;
@@ -1058,17 +1164,24 @@ let repo_kind_flag cli validity =
     "rsync", `rsync;
   ] |> List.map (fun (s,v) -> cli_original, s, v)
   in
-  mk_enum_opt ~cli validity ["k";"kind"] "KIND" (main_kinds @ aliases_kinds)
+  mk_enum_opt ~cli validity ?section ["k";"kind"] "KIND" (main_kinds @ aliases_kinds)
     (Printf.sprintf "Specify the kind of the repository to be used (%s)."
        (string_of_enum main_kinds))
 
-let jobs_flag cli validity =
-  mk_opt ~cli validity ["j";"jobs"] "JOBS"
+let jobs_flag ?section cli validity =
+  mk_opt ~cli validity ?section ["j";"jobs"] "JOBS"
     "Set the maximal number of concurrent jobs to use. The default value is \
     calculated from the number of cores. You can also set it using the \
     $(b,\\$OPAMJOBS) environment variable."
     Arg.(some positive_integer) None
 
+let formula_flag ?section cli =
+  mk_opt ~cli (cli_from ~experimental:true cli2_2) ?section ["formula"] "FORMULA"
+    "Specify a dependency formula to be verified by the solution, in addition \
+     to the switch invariant. The format is the same as for expressing \
+     dependencies in package definition files, e.g. '\"foo\" {>= \"1.1\"}'"
+    dep_formula OpamFormula.Empty
+
 let name_list =
   arg_list "PACKAGES" "List of package names." package_name
 
@@ -1150,7 +1263,7 @@ let global_options cli =
                 This is equivalent to setting $(b,\\$OPAMSWITCH) to $(i,SWITCH)."
       Arg.(some string) None in
   let yes =
-    mk_vflag_all ~cli [
+    mk_vflag_all ~cli ~section [
       cli_original, true, ["y";"yes"],
       "Answer yes to all opam yes/no questions without prompting. \
        See also $(b,--confirm-level). \
@@ -1204,7 +1317,7 @@ let global_options cli =
   let solver_preferences =
     mk_opt ~cli cli_original ~section ["criteria"] "CRITERIA"
       ("Specify user $(i,preferences) for dependency solving for this run. \
-        Overrides both $(b,\\$OPAMCRITERIA) and $(b,\\$OPAMUPGRADECRITERIA). \
+        Overrides $(b,\\$OPAMCRITERIA), $(b,\\$OPAMFIXUPCRITERIA) and $(b,\\$OPAMUPGRADECRITERIA). \
         For details on the supported language, and the external solvers available, see \
         $(i, http://opam.ocaml.org/doc/External_solvers.html). \
         A general guide to using solver preferences can be found at \
@@ -1213,8 +1326,10 @@ let global_options cli =
   let cudf_file =
     mk_opt ~cli cli_original ~section ["cudf"] "FILENAME"
       "Debug option: Save the CUDF requests sent to the solver to \
-       $(docv)-<n>.cudf."
-      Arg.(some string) None in
+       $(docv)-<n>.cudf where <n> is the n-th call to the solver during \
+       the opam run."
+      Arg.(some string) None
+  in
   let best_effort =
     mk_flag ~cli cli_original ~section ["best-effort"]
       "Don't fail if all requested packages can't be installed: try to install \
@@ -1254,9 +1369,9 @@ let global_options cli =
        version-controlled directory, update to the current working state of \
        their source instead of the last committed state, or the ref they are \
        pointing to. As source directory is copied as it is, if it isn't clean \
-       it may result on a opam build failure.\
-       This only affects packages explicitly listed on the command-line.\
-       It can also be set with $(b,\\$OPAMWORKINGDIR). "
+       it may result on a opam build failure. This only affects packages \
+       explicitly listed on the command-line. It can also be set with \
+       $(b,\\$OPAMWORKINGDIR)."
   in
   let ignore_pin_depends =
     mk_flag ~cli cli_original ~section ["ignore-pin-depends"]
@@ -1274,8 +1389,8 @@ let global_options cli =
         $d_no_aspcud $cli_arg)
 
 (* lock options *)
-let locked ?(section=Manpage.s_options) cli =
-  mk_flag ~cli cli_original ~section ["locked"]
+let locked ?section cli =
+  mk_flag ~cli cli_original ?section ["locked"]
     "In commands that use opam files found from pinned sources, if a variant \
      of the file with an added .$(i,locked) extension is found (e.g. \
      $(b,foo.opam.locked) besides $(b,foo.opam)), that will be used instead. \
@@ -1285,8 +1400,8 @@ let locked ?(section=Manpage.s_options) cli =
      of the dependencies currently installed on the host. This is equivalent \
      to setting the $(b,\\$OPAMLOCKED) environment variable. Note that this \
      option doesn't generally affect already pinned packages."
-let lock_suffix ?(section=Manpage.s_options) cli =
-  mk_opt ~cli (cli_from cli2_1) ~section ["lock-suffix"] "SUFFIX"
+let lock_suffix ?section cli =
+  mk_opt ~cli (cli_from cli2_1) ?section ["lock-suffix"] "SUFFIX"
     "Set locked files suffix to $(i,SUFFIX)."
     Arg.(string) ("locked")
 
@@ -1343,6 +1458,10 @@ let build_options cli =
        the command-line. This is equivalent to setting $(b,\\$OPAMWITHDOC) \
        (or the deprecated $(b,\\$OPAMBUILDDOC)) to \"true\"."
   in
+  let dev_setup =
+    mk_flag ~cli (cli_from cli2_2) ["with-dev-setup"] ~section
+      "Include development tools only dependencies."
+  in
   let make =
     mk_opt ~cli (cli_between cli2_0 cli2_1
                    ~replaced:"opam config set[-global] make MAKE")
@@ -1377,9 +1496,9 @@ let build_options cli =
     mk_opt ~cli cli_original ~section ["ignore-constraints-on"] "PACKAGES"
       "Forces opam to ignore version constraints on all dependencies to the \
        listed packages. This can be used to test compatibility, but expect \
-       builds to break when using this. Note that version constraints on \
-       optional dependencies and conflicts are unaffected. This is equivalent \
-       to setting $(b,\\$OPAMIGNORECONSTRAINTS)."
+       builds to break when using this. Note that availability of the packages \
+       as well as version constraints on optional dependencies and conflicts \
+       are unaffected. This is equivalent to setting $(b,\\$OPAMIGNORECONSTRAINTS)."
       Arg.(some (list package_name)) None ~vopt:(Some [])
   in
   let unlock_base =
@@ -1408,13 +1527,14 @@ let build_options cli =
   in
   Term.(const create_build_options
         $keep_build_dir $reuse_build_dir $inplace_build $make
-        $no_checksums $req_checksums $build_test $build_doc $show $dryrun
-        $skip_update $fake $jobs_flag cli cli_original $ignore_constraints_on
-        $unlock_base $locked $lock_suffix $assume_depexts $no_depexts)
+        $no_checksums $req_checksums $build_test $build_doc $dev_setup $show
+        $dryrun $skip_update $fake $jobs_flag ~section cli cli_original
+        $ignore_constraints_on $unlock_base $locked $lock_suffix
+        $assume_depexts $no_depexts)
 
 (* Option common to install commands *)
-let assume_built cli =
-  mk_flag ~cli cli_original  ["assume-built"]
+let assume_built ?section cli =
+  mk_flag ~cli cli_original  ?section ["assume-built"]
     "For use on locally-pinned packages: assume they have already \
      been correctly built, and only run their installation \
      instructions, directly from their source directory. This \
@@ -1423,27 +1543,42 @@ let assume_built cli =
      No locally-pinned packages will be skipped."
 
 (* Options common to all path based/related commands, e.g. (un)pin, upgrade,
-   remove, (re)install.
-   Disabled *)
-let recurse _cli = Term.const false
+   remove, (re)install. *)
 
-(*
-  mk_flag ~cli (cli_from cli2_2) ["recursive"]
-    "Allow recursive lookups of (b,*.opam) files. Cf. $(i,--subpath) also."
-*)
+let recurse ?section cli =
+  mk_flag ~cli (cli_from cli2_2) ?section ["recursive"]
+    "Allow recursive lookups of $(b,*.opam) files. Cf. $(i,--subpath) also."
 
-let subpath _cli = Term.const None
-(*
-  mk_opt ~cli (cli_from cli2_2) ["subpath"] "PATH"
+let subpath ?section cli =
+  mk_opt ~cli (cli_from cli2_2) ?section ["subpath"] "PATH"
     "$(b,*.opam) files are retrieved from the given sub directory instead of \
       top directory. Sources are then taken from the targeted sub directory, \
       internally only this subdirectory is copied/fetched.  It can be combined \
       with $(i,--recursive) to have a recursive lookup on the subpath."
     Arg.(some subpath_conv) None
-*)
 
 let package_selection_section = "PACKAGE SELECTION OPTIONS"
 
+let post ?(section=package_selection_section) cli =
+  mk_flag ~cli cli_original ["post"]  ~section
+    "Include dependencies tagged as $(i,post)."
+
+let dev ?(section=package_selection_section) cli =
+  mk_flag ~cli cli_original ["dev"]  ~section
+    "Include development packages in dependencies."
+
+let doc_flag ?(section=package_selection_section) cli =
+  mk_flag ~cli cli_original ["doc";"with-doc"] ~section
+    "Include doc-only dependencies."
+
+let test ?(section=package_selection_section) cli =
+  mk_flag ~cli cli_original ["t";"test";"with-test"] ~section
+    "Include test-only dependencies."
+
+let dev_setup ?(section=package_selection_section) cli =
+  mk_flag ~cli (cli_from cli2_2) ["with-dev-setup"] ~section
+    "Include developer only dependencies."
+
 let package_selection cli =
   let section = package_selection_section in
   let depends_on =
@@ -1496,22 +1631,6 @@ let package_selection cli =
     mk_flag ~cli cli_original ["nobuild"]  ~section
       "Exclude build dependencies (they are included by default)."
   in
-  let post =
-    mk_flag ~cli cli_original ["post"]  ~section
-      "Include dependencies tagged as $(i,post)."
-  in
-  let dev =
-    mk_flag ~cli cli_original ["dev"]  ~section
-      "Include development packages in dependencies."
-  in
-  let doc_flag =
-    mk_flag ~cli cli_original ["doc";"with-doc"] ~section
-      "Include doc-only dependencies."
-  in
-  let test =
-    mk_flag ~cli cli_original ["t";"test";"with-test"] ~section
-      "Include test-only dependencies."
-  in
   let field_match =
     mk_opt_all ~cli cli_original ["field-match"] "FIELD:PATTERN" ~section
       "Filter packages with a match for $(i,PATTERN) on the given $(i,FIELD)"
@@ -1539,13 +1658,14 @@ let package_selection cli =
       Arg.string
   in
   let filter
-      depends_on required_by conflicts_with coinstallable_with resolve recursive
-      depopts nobuild post dev doc_flag test field_match has_flag has_tag
+      depends_on required_by conflicts_with coinstallable_with resolve
+      recursive depopts nobuild post dev doc_flag test dev_setup field_match
+      has_flag has_tag
     =
     let dependency_toggles = {
       OpamListCommand.
-      recursive; depopts; build = not nobuild; post; test; doc = doc_flag;
-      dev
+      recursive; depopts; build = not nobuild; post; test; dev_setup;
+      doc = doc_flag; dev
     } in
     List.map (fun flag -> OpamListCommand.Flag flag) has_flag @
     List.map (fun tag -> OpamListCommand.Tag tag) has_tag @
@@ -1573,8 +1693,9 @@ let package_selection cli =
   in
   Term.(const filter $
         depends_on $ required_by $ conflicts_with $ coinstallable_with $
-        resolve $ recursive $ depopts $ nobuild $ post $ dev $ doc_flag $
-        test $ field_match $ has_flag $ has_tag)
+        resolve $ recursive $ depopts $ nobuild $ post cli $ dev cli $
+        doc_flag cli $ test cli $ dev_setup cli $ field_match $ has_flag $
+        has_tag)
 
 let package_listing_section = "OUTPUT FORMAT OPTIONS"
 
