@@ -209,7 +209,7 @@ let opam_init ?root_dir ?strict ?solver =
   let open OpamStd.Option.Op in
 
   (* (i) get root dir *)
-  let root = OpamStateConfig.opamroot ?root_dir () in
+  let root_from, root = OpamStateConfig.opamroot_with_provenance ?root_dir () in
 
   (* (ii) load conf file and set defaults *)
   (* the init for OpamFormat is done in advance since (a) it has an effect on
@@ -233,7 +233,7 @@ let opam_init ?root_dir ?strict ?solver =
     | Some conf ->
       let criteria kind =
         let c = OpamFile.Config.criteria conf in
-        try Some (List.assoc kind c) with Not_found -> None
+        OpamStd.(List.assoc_opt Compare.equal kind c)
       in
       OpamSolverConfig.update
         ?solver
@@ -261,5 +261,5 @@ let opam_init ?root_dir ?strict ?solver =
   OpamCoreConfig.initk ?log_dir |>
   OpamRepositoryConfig.initk |>
   OpamSolverConfig.initk ?solver |>
-  OpamStateConfig.initk ~root_dir:root |>
+  OpamStateConfig.initk ~root_dir:root ~root_from |>
   initk
