@@ -9,7 +9,6 @@
 (*                                                                        *)
 (**************************************************************************)
 
-open OpamCompat
 open OpamParserTypes.FullPos
 open OpamTypes
 open OpamStateTypes
@@ -25,6 +24,7 @@ type dependency_toggles = {
   build: bool;
   post: bool;
   test: bool;
+  dev_setup: bool;
   doc: bool;
   dev: bool;
 }
@@ -35,6 +35,7 @@ let default_dependency_toggles = {
   build = true;
   post = false;
   test = false;
+  dev_setup = false;
   doc = false;
   dev = false;
 }
@@ -81,7 +82,7 @@ let string_of_selector =
   | Any -> "any" % `cyan
   | Installed -> "installed" % `cyan
   | Root -> "root" % `cyan
-  | Compiler -> "base" % `cyan
+  | Compiler -> "invariant" % `cyan
   | Available -> "available" % `cyan
   | Installable -> "installable" % `cyan
   | Pinned -> "pinned" % `cyan
@@ -152,7 +153,7 @@ let package_dependencies st tog nv =
   get_opam st nv |>
   OpamPackageVar.all_depends
     ~build:tog.build ~post:tog.post
-    ~test:tog.test ~doc:tog.doc ~dev:tog.dev
+    ~test:tog.test ~doc:tog.doc ~dev_setup:tog.dev_setup ~dev:tog.dev
     ~depopts:tog.depopts
     st
 
@@ -167,10 +168,10 @@ let get_universe st ?requested tog =
   let requested =
     match requested with
     | Some r -> r
-    | None -> OpamPackage.names_of_packages st.packages
+    | None -> st.packages
   in
   OpamSwitchState.universe st
-    ~test:tog.test ~doc:tog.doc ~force_dev_deps:tog.dev
+    ~test:tog.test ~doc:tog.doc ~dev_setup:tog.dev_setup ~force_dev_deps:tog.dev
     ~requested Query
 
 let rec value_strings value =
@@ -214,24 +215,23 @@ let apply_selector ~base st = function
   | Any -> base
   | Installed -> st.installed
   | Root -> st.installed_roots
-  | Compiler -> st.compiler_packages
+  | Compiler -> OpamSwitchState.invariant_root_packages st
   | Available -> Lazy.force st.available_packages
   | Installable ->
     OpamSolver.installable_subset
-      (OpamSwitchState.universe st ~requested:OpamPackage.Name.Set.empty Query)
+      (OpamSwitchState.universe st ~requested:OpamPackage.Set.empty Query)
       base
   | Pinned -> OpamPinned.packages st
   | (Required_by ({recursive=true; _} as tog, atoms)
     | Depends_on ({recursive=true; _} as tog, atoms)) as direction ->
     let deps_fun = match direction with
-      | Required_by _ -> OpamSolver.dependencies
-      | Depends_on _ -> OpamSolver.reverse_dependencies
+      | Required_by _ -> OpamSwitchState.dependencies
+      | Depends_on _ -> OpamSwitchState.reverse_dependencies
       | _ -> assert false
     in
     deps_fun ~depopts:tog.depopts ~build:tog.build ~post:tog.post
       ~installed:false ~unavailable:true
-      (get_universe st tog)
-      (packages_of_atoms st atoms)
+      st (packages_of_atoms st atoms)
   | Required_by (tog, atoms) ->
     atom_dependencies st tog atoms |>
     OpamFormula.packages base
@@ -253,7 +253,6 @@ let apply_selector ~base st = function
     let universe =
       let requested =
         OpamFormula.packages_of_atoms st.packages atoms
-        |> OpamPackage.names_of_packages
       in
       get_universe st tog ~requested
     in
@@ -262,7 +261,7 @@ let apply_selector ~base st = function
         with u_installed = OpamPackage.Set.empty;
              u_installed_roots = OpamPackage.Set.empty }
     in
-    (match OpamSolver.resolve universe ~orphans:OpamPackage.Set.empty
+    (match OpamSolver.resolve universe
              (OpamSolver.request ~install:atoms ()) with
     | Success s -> OpamSolver.new_packages s
     | Conflicts cs ->
@@ -448,7 +447,7 @@ let raw_field_names =
 let string_of_field ?(raw=false) = function
   | Field s -> if raw then s ^":" else s
   | Raw_field s -> s ^":"
-  | f -> List.assoc f field_names
+  | f -> OpamStd.(List.assoc Compare.equal f field_names)
 
 let field_of_string ~raw =
   let names_fields = List.map (fun (a,b) -> b, a) field_names in
@@ -461,7 +460,7 @@ let field_of_string ~raw =
       Raw_field (OpamStd.String.remove_suffix ~suffix:":" s)
     else
     try
-      List.assoc s names_fields
+      OpamStd.List.assoc String.equal s names_fields
     with Not_found ->
     match OpamStd.List.find_opt (fun x -> s = x) opam_fields with
     | Some f -> Field f
@@ -540,16 +539,35 @@ let detail_printer ?prettify ?normalise ?(sort=false) st nv =
     +! ""
   | Raw_field f | Field f ->
     (try
-       List.assoc f (OpamFile.OPAM.to_list (get_opam st nv)) |>
-       mini_field_printer ?prettify ?normalise
+       let open OpamStd.Option.Op in
+       let opam = get_opam st nv in
+       let value =
+         match f with
+         | "url.swhid" ->
+           (match OpamFile.(OPAM.url opam >>= URL.swhid) with
+            | Some f -> OpamTypesBase.nullify_pos (String (OpamSWHID.to_string f))
+            | None -> raise Not_found)
+         | "url.mirrors" ->
+           OpamFile.(OPAM.with_url_opt
+                       (OPAM.url opam >>| URL.with_swhid_opt None)
+                       opam
+                     |> OPAM.to_list
+                     |> OpamStd.List.assoc String.equal f)
+         | _ -> OpamStd.List.assoc String.equal f (OpamFile.OPAM.to_list opam)
+       in
+       mini_field_printer ?prettify ?normalise value
      with Not_found -> "")
   | Installed_version ->
     (try OpamPackage.package_of_name st.installed nv.name |> fun inst_nv ->
          OpamPackage.version_to_string inst_nv |> fun s ->
          if OpamPackage.Set.mem inst_nv st.pinned then s % [`blue] else
          if OpamPackage.has_name st.pinned nv.name then s % [`bold;`red] else
-         if nv <> inst_nv &&
-            not (OpamPackage.Set.mem inst_nv st.compiler_packages)
+         (* If package is not installed one, check if it is part of invariant
+            formula package, and if it checks formula *)
+         if nv <> inst_nv
+         && (not (OpamPackage.Set.mem inst_nv
+                    (OpamFormula.packages st.installed st.switch_invariant))
+             || OpamFormula.verifies st.switch_invariant nv)
          then s % [`bold;`yellow] else
            s % [`magenta]
      with Not_found -> "--" % [`cyan])
@@ -599,7 +617,10 @@ let detail_printer ?prettify ?normalise ?(sort=false) st nv =
       OpamPath.Switch.changes st.switch_global.root st.switch nv.name
     in
     (match OpamFile.Changes.read_opt changes_f with
-     | None -> ""
+     | None ->
+       OpamConsole.error_and_exit `Not_found
+         "Cannot list installed files: %s is not installed"
+         (OpamPackage.Name.to_string nv.name)
      | Some c ->
        OpamStd.Format.itemize ~bullet:""
          (fun (file, status) ->
@@ -662,7 +683,7 @@ let display st format packages =
     if format.order = `Dependency then
       let universe =
         OpamSwitchState.universe st
-          ~requested:(OpamPackage.names_of_packages packages)
+          ~requested:packages
           Query
       in
       OpamSolver.dependency_sort ~depopts:true ~build:true ~post:false
@@ -740,6 +761,7 @@ let info st ~fields ~raw ~where ?normalise ?(show_empty=false)
     Pinning_target;
     Source_hash;
     Field "url.src";
+    Field "url.swhid";
     Field "url.checksum";
     Field "homepage";
     Field "doc";
@@ -754,6 +776,7 @@ let info st ~fields ~raw ~where ?normalise ?(show_empty=false)
     Field "depopts";
     Field "conflicts";
     Field "conflict-class";
+    Field "depexts";
     Synopsis;
     Description;
   ] in
