@@ -24,7 +24,8 @@ exception Fetch_fail of string
 
 let post_message ?(failed=false) st action =
   match action, failed with
-  | `Remove _, _ | `Reinstall _, _ | `Build _, false | `Fetch _, _ -> ()
+  | `Remove _, _ | `Reinstall _, _ | `Build _, false
+  | `Fetch _, _ -> ()
   | `Build pkg, true | `Install pkg, _ | `Change (_,_,pkg), _ ->
     let opam = OpamSwitchState.opam st pkg in
     let messages = OpamFile.OPAM.post_messages opam in
@@ -37,8 +38,16 @@ let post_message ?(failed=false) st action =
       OpamVariable.Map.add (OpamVariable.of_string "failure")
         (Some (B failed)) local_variables
     in
+    let local_variables =
+      OpamVariable.Map.add (OpamVariable.of_string "with-dev-setup")
+        (Some (B OpamStateConfig.(!r.dev_setup))) local_variables
+    in
     let messages =
       let filter_env = OpamPackageVar.resolve ~opam ~local:local_variables st in
+      (if OpamFile.OPAM.has_flag Pkgflag_Deprecated opam then
+        ["Note: This package is deprecated."]
+      else
+        []) @
       OpamStd.List.filter_map (fun (message,filter) ->
           if OpamFilter.opt_eval_to_bool filter_env filter then
             Some (OpamFilter.expand_string ~default:(fun _ -> "")
@@ -154,7 +163,8 @@ let check_availability ?permissive t set atoms =
            (OpamFormula.short_string_of_atom atom)
            msg
            (OpamStd.Option.map_default (Printf.sprintf "\n%s.") ""
-              (OpamSysInteract.repo_enablers ())))
+              (OpamSysInteract.repo_enablers
+                 ~env:t.switch_global.global_variables t.switch_global.config)))
     | None -> None
   in
   let check_atom (name, cstr as atom) =
@@ -171,11 +181,10 @@ let check_availability ?permissive t set atoms =
     then Some (OpamSwitchState.not_found_message t atom)
     else
     let f = name, match cstr with None -> Empty | Some c -> Atom c in
-    Some (Printf.sprintf "%s %s"
+    Some (Printf.sprintf "%s: %s"
             (OpamFormula.to_string (Atom f))
             (OpamSwitchState.unavailable_reason
-               ~default:"unavailable for unknown reasons (this may be a bug in \
-                         opam)"
+               ~default:"the package no longer exists"
                t f)) in
   let errors = OpamStd.List.filter_map check_atom atoms in
   if errors <> [] then
@@ -197,22 +206,26 @@ let fuzzy_name t name =
   | [name] -> name
   | _ -> name
 
-let sanitize_atom_list ?(permissive=false) t atoms =
+let sanitize_atom_list ?(permissive=false) ?(installed=false) t atoms =
   let atoms = List.map (fun (name,cstr) -> fuzzy_name t name, cstr) atoms in
+  let open OpamPackage.Set.Op in
   if permissive then
     check_availability ~permissive t
-      (OpamPackage.Set.union t.packages t.installed) atoms
+      (t.packages ++ t.installed) atoms
   else
     check_availability t
-      (OpamPackage.Set.union (Lazy.force t.available_packages) t.installed)
+      (if installed then t.installed ++ Lazy.force t.available_packages
+       else Lazy.force t.available_packages)
       atoms;
   atoms
 
 (* Pretty-print errors *)
 let display_error (n, error) =
-  let f action nv =
+  let f action nvs =
     let disp =
-      OpamConsole.header_error "while %s %s" action (OpamPackage.to_string nv) in
+      OpamConsole.header_error "while %s %s" action
+        (OpamStd.Format.pretty_list (List.map OpamPackage.to_string nvs))
+    in
     match error with
     | Sys.Break | OpamParallel.Aborted -> ()
     | Failure s -> disp "%s" s
@@ -223,13 +236,13 @@ let display_error (n, error) =
         OpamConsole.errmsg "%s" (OpamStd.Exn.pretty_backtrace e)
   in
   match n with
-  | `Change (`Up, _, nv)   -> f "upgrading to" nv
-  | `Change (`Down, _, nv) -> f "downgrading to" nv
-  | `Install nv        -> f "installing" nv
-  | `Reinstall nv      -> f "recompiling" nv
-  | `Remove nv         -> f "removing" nv
-  | `Build nv          -> f "compiling" nv
-  | `Fetch nv          -> f "fetching sources for" nv
+  | `Change (`Up, _, nv)   -> f "upgrading to" [nv]
+  | `Change (`Down, _, nv) -> f "downgrading to" [nv]
+  | `Install nv        -> f "installing" [nv]
+  | `Reinstall nv      -> f "recompiling" [nv]
+  | `Remove nv         -> f "removing" [nv]
+  | `Build nv          -> f "compiling" [nv]
+  | `Fetch nvs         -> f "fetching sources for" nvs
 
 module Json = struct
   let output_request request user_action =
@@ -239,9 +252,12 @@ module Json = struct
     in
     let j = `O [
         "action", `String (string_of_user_action user_action);
-        "install", `A (atoms request.wish_install);
+        "install", `A (List.map (fun f ->
+            `String OpamFormula.(string_of_formula short_string_of_atom f))
+            (OpamFormula.ands_to_list request.wish_install));
         "remove", `A (atoms request.wish_remove);
         "upgrade", `A (atoms request.wish_upgrade);
+        "all", `A (atoms request.wish_all);
         "criteria", `String (OpamSolverConfig.criteria request.criteria);
       ]
     in
@@ -406,7 +422,7 @@ let parallel_apply t
          (let spkgs = OpamSysPkg.Set.Op.(bypass -- !bypass_ref) in
           OpamConsole.note
             "Requirement for system package%s %s overridden in this switch. Use \
-             `opam option depext-bypass-=%s' to revert."
+             `opam option 'depext-bypass-=%S'' to revert."
             (if OpamSysPkg.Set.cardinal spkgs > 1 then "s" else "")
             (OpamStd.Format.pretty_list
                (List.map OpamSysPkg.to_string
@@ -446,7 +462,19 @@ let parallel_apply t
   let sources_needed =
     let sources_needed = OpamAction.sources_needed t action_graph in
     if not OpamClientConfig.(!r.working_dir) then sources_needed else
-      OpamPackage.Set.Op.(sources_needed -- requested)
+    let no_sources = OpamPackage.Set.Op.(requested %% t.pinned) in
+    let no_sources =
+      OpamPackage.Set.filter (fun nv ->
+          OpamStd.Option.Op.(OpamSwitchState.primary_url t nv
+                             >>= OpamUrl.local_dir) <> None)
+        no_sources
+    in
+    if OpamPackage.Set.is_empty no_sources then
+      (OpamConsole.note
+         "--working-dir is given but no requested package is pinned";
+       sources_needed)
+    else
+      OpamPackage.Set.Op.(sources_needed -- no_sources)
   in
 
   (* 1/ process the package actions (fetch, build, installations and removals) *)
@@ -454,10 +482,41 @@ let parallel_apply t
   let action_graph = (* Add build and fetch actions *)
     let noop_remove nv =
       OpamAction.noop_remove_package t nv in
-    PackageActionGraph.explicit
-      ~noop_remove
-      ~sources_needed:(fun p -> OpamPackage.Set.mem p sources_needed)
-      action_graph
+    let sources_needed =
+      (* used to factorise fetches: creates [`Fetch] nodes with several
+         packages that download once the archive and copies its content in
+         source directories. Handles only http and rsync backend with at least
+         one checksum. *)
+      let shared_source =
+        OpamPackage.Set.fold (fun nv url_nvs ->
+            match OpamSwitchState.url t nv with
+            | Some url ->
+              let checksums = OpamFile.URL.checksum url in
+              let url = OpamFile.URL.url url in
+              (match url.OpamUrl.backend, checksums with
+               | (`http | `rsync ) , _::_
+                 when OpamFilename.is_archive
+                     (OpamFilename.of_string url.OpamUrl.path) ->
+                 OpamUrl.Map.update url
+                   (OpamPackage.Set.add nv)
+                   OpamPackage.Set.empty url_nvs
+               | _ -> url_nvs)
+            | None -> url_nvs)
+          sources_needed OpamUrl.Map.empty
+        |> OpamUrl.Map.values
+        |> List.filter (fun s -> not @@ OpamPackage.Set.is_singleton s)
+      in
+      log "Regroup shared source packages: %s"
+        (OpamStd.List.to_string OpamPackage.Set.to_string shared_source);
+      fun p ->
+        try
+          List.find (OpamPackage.Set.mem p) shared_source
+          |> OpamPackage.Set.elements
+        with Not_found ->
+        try [ OpamPackage.Set.find (OpamPackage.equal p) sources_needed ]
+        with Not_found -> []
+    in
+    PackageActionGraph.explicit ~noop_remove ~sources_needed action_graph
   in
   let action_graph =
     if download_only then
@@ -533,16 +592,13 @@ let parallel_apply t
             (fun name _ -> OpamPackage.Set.exists (fun pkg -> OpamPackage.Name.equal name pkg.name) visible_installed)
             !t_ref.conf_files; }
     in
-    let nv = action_contents action in
-    let opam = OpamSwitchState.opam t nv in
-    let source_dir =
+    let source_dir nv =
+      let opam = OpamSwitchState.opam t nv in
       let raw = OpamSwitchState.source_dir t nv in
       match OpamFile.OPAM.url opam with
       | None -> raw
-      | Some url ->
-        match OpamFile.URL.subpath url with
-        | None -> raw
-        | Some subpath -> OpamFilename.Op.(raw / subpath) in
+      | Some url -> OpamFilename.SubPath.(raw /? OpamFile.URL.subpath url)
+    in
     if OpamClientConfig.(!r.fake) then
       match action with
       | `Build _ | `Fetch _ -> Done (`Successful (installed, removed))
@@ -554,16 +610,32 @@ let parallel_apply t
       | `Remove nv ->
         remove_from_install nv;
         Done (`Successful (installed, OpamPackage.Set.add nv removed))
-      | _ -> assert false
+      | `Change _ | `Reinstall _ -> assert false
     else
     match action with
-    | `Fetch nv ->
-      log "Fetching sources for %s" (OpamPackage.to_string nv);
-      (OpamAction.download_package t nv @@+ function
-        | None ->
-          store_time (); Done (`Successful (installed, removed))
-        | Some (_short_error, long_error) ->
-          Done (`Exception (Fetch_fail long_error)))
+    | `Fetch nvs ->
+      log "Fetching sources for %s"
+        (OpamStd.Format.pretty_list (List.map OpamPackage.to_string nvs));
+      ((match nvs with
+          | [nv] ->
+            OpamAction.download_package t nv
+          | _ ->
+            let url =
+              match OpamStd.List.filter_map (OpamSwitchState.url t) nvs with
+              | u::_ as urls ->
+                let checksums =
+                  List.map OpamFile.URL.checksum urls
+                  |> List.flatten
+                  |> List.sort_uniq OpamHash.compare
+                in
+                Some (OpamFile.URL.with_checksum checksums u)
+              | [] -> None
+            in
+            OpamAction.download_shared_source t url nvs) @@+ function
+       | None ->
+         store_time (); Done (`Successful (installed, removed))
+       | Some (_short_error, long_error) ->
+         Done (`Exception (Fetch_fail long_error)))
 
     | `Build nv ->
       if assume_built && OpamPackage.Set.mem nv requested then
@@ -582,12 +654,13 @@ let parallel_apply t
             OpamFilename.rmdir dir;
           false, dir
       in
-      let test =
-        OpamStateConfig.(!r.build_test) && OpamPackage.Set.mem nv requested
-      in
-      let doc =
-        OpamStateConfig.(!r.build_doc) && OpamPackage.Set.mem nv requested
+      let test, doc, dev_setup =
+        let found = OpamPackage.Set.mem nv requested in
+        OpamStateConfig.(!r.build_test) && found,
+        OpamStateConfig.(!r.build_doc) && found,
+        OpamStateConfig.(!r.dev_setup) && found
       in
+      let source_dir = source_dir nv in
       (if OpamFilename.exists_dir source_dir
        then (if not is_inplace then
                OpamFilename.copy_dir ~src:source_dir ~dst:build_dir)
@@ -595,29 +668,32 @@ let parallel_apply t
        OpamAction.prepare_package_source t nv build_dir @@+ function
        | Some exn -> store_time (); Done (`Exception exn)
        | None ->
-         OpamAction.build_package t ~test ~doc build_dir nv @@+ function
+         OpamAction.build_package t ~test ~doc ~dev_setup build_dir nv
+         @@+ function
          | Some exn -> store_time (); Done (`Exception exn)
          | None -> store_time (); Done (`Successful (installed, removed)))
     | `Install nv ->
-      let test =
-        OpamStateConfig.(!r.build_test) && OpamPackage.Set.mem nv requested
-      in
-      let doc =
-        OpamStateConfig.(!r.build_doc) && OpamPackage.Set.mem nv requested
+      let test, doc, dev_setup =
+        let found = OpamPackage.Set.mem nv requested in
+        OpamStateConfig.(!r.build_test) && found,
+        OpamStateConfig.(!r.build_doc) && found,
+        OpamStateConfig.(!r.dev_setup) && found
       in
       let build_dir = OpamPackage.Map.find_opt nv inplace in
-      (OpamAction.install_package t ~test ~doc ?build_dir nv @@+ function
-        | Left conf ->
-          add_to_install nv conf;
-          store_time ();
-          Done (`Successful (OpamPackage.Set.add nv installed, removed))
-        | Right exn ->
-          store_time ();
-          Done (`Exception exn))
+      (OpamAction.install_package t ~test ~doc ~dev_setup ?build_dir nv
+       @@+ function
+       | Left conf ->
+         add_to_install nv conf;
+         store_time ();
+         Done (`Successful (OpamPackage.Set.add nv installed, removed))
+       | Right exn ->
+         store_time ();
+         Done (`Exception exn))
     | `Remove nv ->
       (if OpamAction.removal_needs_download t nv then
          let d = OpamPath.Switch.remove t.switch_global.root t.switch nv in
          OpamFilename.rmdir d;
+         let source_dir = source_dir nv in
          if OpamFilename.exists_dir source_dir
          then OpamFilename.copy_dir ~src:source_dir ~dst:d
          else OpamFilename.mkdir d;
@@ -630,7 +706,7 @@ let parallel_apply t
         nv;
       store_time ();
       `Successful (installed, OpamPackage.Set.add nv removed)
-    | _ -> assert false
+    | `Change _ | `Reinstall _ -> assert false
   in
 
   let action_results =
@@ -679,7 +755,7 @@ let parallel_apply t
         (* Report download failures *)
         let failed_downloads = List.fold_left (fun failed (a, err) ->
             match (a, err) with
-            | `Fetch pkg, `Exception (Fetch_fail long_error) ->
+            | `Fetch [pkg], `Exception (Fetch_fail long_error) ->
               OpamPackage.Map.add pkg long_error failed
             | _ ->
               failed
@@ -694,7 +770,9 @@ let parallel_apply t
               match a with
               | `Fetch _ -> acc
               | _ ->
-                let r = match List.assoc a results with
+                let r =
+                  match OpamStd.List.assoc
+                          PackageActionGraph.Parallel.G.V.equal a results with
                   | `Successful _ -> `String "OK"
                   | `Exception e -> Json.exc e
                   | `Error (`Aborted deps) ->
@@ -770,7 +848,7 @@ let parallel_apply t
           if not OpamClientConfig.(!r.keep_build_dir) then
             OpamFilename.rmdir build_dir
         | `Remove _ | `Install _ | `Build _ | `Fetch _ -> ()
-        | _ -> assert false)
+        | `Change _ | `Reinstall _  -> assert false)
       graph
   in
   let t =
@@ -832,9 +910,11 @@ let parallel_apply t
       (* Cleanup build/install actions when one of them failed, it's verbose and
          doesn't add information *)
       let successful =
+        let was_successful p = not @@ List.mem (`Install p) failed in
         List.filter (function
-            | `Fetch p | `Build p when List.mem (`Install p) failed -> false
-            | _ -> true)
+            | `Fetch ps -> List.for_all was_successful ps
+            | `Build p -> was_successful p
+            | `Change _ | `Install _ | `Reinstall _ | `Remove _ -> true)
           successful
       in
       let remaining =
@@ -842,18 +922,24 @@ let parallel_apply t
             | `Remove p | `Install p
               when List.mem (`Build p) failed -> false
             | `Remove p | `Install p | `Build p
-              when List.mem (`Fetch p) failed -> false
-            | _ -> true)
+              when List.exists (function
+                  | `Fetch ps -> List.mem p ps
+                  | _ -> false) failed
+              -> false
+            | `Build _ | `Change _ | `Fetch _ | `Install _
+            | `Reinstall _ | `Remove _ -> true)
           remaining
       in
       let removes_missing_source =
         List.filter (function
             | `Remove p as rem ->
-              let fetch = `Fetch p in
-              List.mem fetch failed &&
-              PackageActionGraph.mem_edge action_graph fetch rem
-            | _ -> false
-          )
+              let is_fetch = function `Fetch ps -> List.mem p ps | _ -> false in
+              List.exists is_fetch failed
+              && PackageActionGraph.fold_edges (fun v v' mem ->
+                  mem || (is_fetch v && PackageAction.equal v' rem))
+                action_graph false
+            | `Build _ | `Change _ | `Fetch _ | `Install _
+            | `Reinstall _ -> false)
           successful
       in
       let failed =
@@ -864,7 +950,8 @@ let parallel_apply t
               let succ = PackageActionGraph.succ action_graph a in
               not (List.for_all (fun a -> List.mem a removes_missing_source)
                      succ)
-            | _ -> true)
+            | `Build _ | `Change _ | `Install _ | `Reinstall _
+            | `Remove _ -> true)
           failed
       in
       let filter_graph l =
@@ -899,7 +986,7 @@ let parallel_apply t
                (fun x -> x)
                (List.map (String.concat " ") @@
                 OpamStd.Format.align_table
-                  (PackageAction.to_aligned_strings actions)))
+                  (PackageAction.to_aligned_strings ~explicit:true actions)))
             (colorise tint
                (Printf.sprintf "%s%s "
                   (utf8_symbol Symbols.box_drawings_light_up_and_right "+")
@@ -919,9 +1006,9 @@ let parallel_apply t
          OpamConsole.warning
                  "The sources of the following couldn't be obtained, they may be \
                   uncleanly removed:\n%s"
-                 (OpamStd.Format.itemize
-                    (fun rm -> OpamPackage.to_string (action_contents rm))
-                    removes_missing_source));
+                 (OpamStd.Format.itemize OpamPackage.to_string
+                    (List.flatten
+                       (List.map action_contents removes_missing_source))));
       OpamConsole.msg "\n";
       OpamConsole.header_msg "Error report";
       if OpamConsole.debug () || OpamConsole.verbose () then
@@ -940,7 +1027,7 @@ let parallel_apply t
         ~empty:"No changes have been performed"
         successful;
       t, err
-    | _ -> assert false
+    | Nothing_to_do | OK _  -> assert false
 
 let simulate_new_state state t =
   let installed =
@@ -956,22 +1043,20 @@ let simulate_new_state state t =
       t state.installed in
   { state with installed }
 
+let dry_run state solution =
+  simulate_new_state state (OpamSolver.get_atomic_action_graph solution)
+
 (* Ask confirmation whenever the packages to modify are not exactly
    the packages in the user request *)
 let confirmation ?ask requested solution =
   OpamCoreConfig.answer_is_yes () ||
-  match ask with
-  | Some false -> true
-  | Some true -> OpamConsole.confirm "Do you want to continue?"
-  | None ->
-    let open PackageActionGraph in
-    let solution_packages =
-      fold_vertex (fun v acc ->
-          OpamPackage.Name.Set.add (OpamPackage.name (action_contents v)) acc)
-        solution
-        OpamPackage.Name.Set.empty in
-    OpamPackage.Name.Set.equal requested solution_packages
-    || OpamConsole.confirm "Do you want to continue?"
+  ask = Some false ||
+  let solution_packages =
+    OpamPackage.names_of_packages (OpamSolver.all_packages solution)
+  in
+  ask <> Some true && OpamPackage.Name.Set.equal requested solution_packages ||
+  let stats = OpamSolver.stats solution in
+  OpamConsole.confirm "\nProceed with %s?" (OpamSolver.string_of_stats stats)
 
 let run_hook_job t name ?(local=[]) ?(allow_stdout=false) w =
   let shell_env = OpamEnv.get_full ~set_opamroot:true ~set_opamswitch:true ~force_path:true t in
@@ -987,7 +1072,7 @@ let run_hook_job t name ?(local=[]) ?(allow_stdout=false) w =
     | [] -> None
   in
   let env v =
-    try Some (List.assoc v local)
+    try Some (OpamStd.List.assoc OpamVariable.Full.equal v local)
     with Not_found -> OpamPackageVar.resolve_switch t v
   in
   let rec iter_commands = function
@@ -1029,12 +1114,25 @@ let print_depext_msg (avail, nf) =
 
 (* Gets depexts from the state, without checking again, unless [recover] is
    true. *)
-let get_depexts ?(recover=false) t packages =
+let get_depexts ?(force=false) ?(recover=false) t packages =
+  if not force && OpamStateConfig.(!r.no_depexts) then OpamSysPkg.Set.empty else
   let sys_packages =
     if recover then
       OpamSwitchState.depexts_status_of_packages t packages
     else
-      Lazy.force t.sys_packages
+      let base = Lazy.force t.sys_packages in
+      (* workaround: st.sys_packages is not always updated with added
+         packages *)
+      let more_pkgs =
+        OpamPackage.Set.filter (fun nv ->
+            (* dirty heuristic: recompute for all non-canonical packages *)
+            OpamPackage.Map.find_opt nv t.repos_package_index
+            <> OpamSwitchState.opam_opt t nv)
+          packages
+      in
+      if OpamPackage.Set.is_empty more_pkgs then base else
+      OpamPackage.Map.union (fun _ x -> x) base
+        (OpamSwitchState.depexts_status_of_packages t more_pkgs)
   in
   let avail, nf =
     OpamPackage.Set.fold (fun pkg (avail,nf) ->
@@ -1048,29 +1146,141 @@ let get_depexts ?(recover=false) t packages =
   print_depext_msg (avail, nf);
   avail
 
-let install_depexts ?(force_depext=false) ?(confirm=true) t packages =
-  let sys_packages =
-    if force_depext || OpamFile.Config.depext t.switch_global.config then
-      get_depexts ~recover:force_depext t packages
+let install_sys_packages ~map_sysmap ~confirm env config sys_packages t =
+  let rec entry_point t sys_packages =
+    if OpamClientConfig.(!r.fake) then
+      (print_command sys_packages; t)
+    else if OpamFile.Config.depext_run_installs config then
+      if confirm then menu t sys_packages else auto_install t sys_packages
     else
-      OpamSysPkg.Set.empty
-  in
-  if OpamSysPkg.Set.is_empty sys_packages ||
-     OpamClientConfig.(!r.show) ||
-     OpamClientConfig.(!r.assume_depexts)
-  then t else
-  let print () =
+      manual_install t sys_packages
+  and menu t sys_packages =
+    let answer =
+      let pkgman =
+        OpamConsole.colorise `yellow
+          (OpamSysInteract.package_manager_name ~env config)
+      in
+      OpamConsole.menu ~unsafe_yes:`Yes ~default:`Yes ~no:`Quit
+        "opam believes some required external dependencies are missing. opam \
+         can:"
+        ~options:[
+          `Yes, Printf.sprintf
+            "Run %s to install them (may need root/sudo access)" pkgman;
+          `No, Printf.sprintf
+            "Display the recommended %s command and wait while you run it \
+             manually (e.g. in another terminal)" pkgman;
+          `Ignore, "Continue anyway, and, upon success, permanently register \
+                    that this external dependency is present, but not \
+                    detectable";
+          `Quit, "Abort the installation";
+        ]
+    in
+    OpamConsole.msg "\n";
+    match answer with
+    | `Yes -> auto_install t sys_packages
+    | `No ->
+      OpamConsole.note "Use 'opam option depext-run-installs=false' \
+                        if you don't want to be prompted again.";
+      OpamConsole.msg "\n";
+      print_command sys_packages;
+      OpamConsole.pause "Standing by, press enter to continue when done.";
+      OpamConsole.msg "\n";
+      check_again t sys_packages
+    | `Ignore -> bypass t
+    | `Quit -> give_up_msg (); OpamStd.Sys.exit_because `Aborted
+  and print_command sys_packages =
+    (* Ensure that setup-x86_64.exe exists, so that an invalid command is not
+       displayed to the user. *)
+    if OpamSysPoll.os_distribution env = Some "cygwin" then
+      OpamSysInteract.Cygwin.check_setup ~update:false;
     let commands =
-      OpamSysInteract.install_packages_commands sys_packages
-      |> List.map (fun (c,a) -> c::a)
+      OpamSysInteract.install_packages_commands ~env config sys_packages
+      |> List.map (fun ((`AsAdmin c | `AsUser c), a) -> c::a)
     in
     OpamConsole.formatted_msg
       (match commands with
        | [_] -> "This command should get the requirements installed:\n"
        | _ -> "These commands should get the requirements installed:\n");
     OpamConsole.msg "\n    %s\n\n"
-      (OpamStd.List.concat_map "\n    " (String.concat " ") commands)
+      (OpamConsole.colorise `bold
+         (OpamStd.List.concat_map "\n    " (String.concat " ") commands))
+  and manual_install t sys_packages =
+    print_command sys_packages;
+    let answer =
+      OpamConsole.menu ~default:`Continue ~no:`Quit "Would you like opam to:"
+        ~options:[
+          `Continue, "Check again, as the package is now installed";
+          `Ignore, "Continue anyway, and, upon success, permanently register \
+                    that this external dependency is present, but not \
+                    detectable";
+          `Quit, "Abort the installation";
+        ]
+    in
+    OpamConsole.msg "\n";
+    match answer with
+    | `Continue -> check_again t sys_packages
+    | `Ignore -> bypass t
+    | `Quit -> give_up ()
+  and auto_install t sys_packages =
+    try
+      if OpamSysPoll.os_distribution env = Some "cygwin" then
+        OpamSysInteract.Cygwin.check_setup ~update:true;
+      OpamSysInteract.install ~env config sys_packages; (* handles dry_run *)
+      map_sysmap (fun _ -> OpamSysPkg.Set.empty) t
+    with Failure msg ->
+      OpamConsole.error "%s" msg;
+      check_again t sys_packages
+  and check_again t sys_packages =
+    let open OpamSysPkg.Set.Op in
+    let needed, notfound =
+      OpamSysInteract.packages_status ~env config sys_packages
+    in
+    let still_missing = needed ++ notfound in
+    let installed = sys_packages -- still_missing in
+    let t =
+      map_sysmap (fun sysp -> OpamSysPkg.Set.diff sysp installed) t
+    in
+    if OpamSysPkg.Set.is_empty still_missing then t else
+    if OpamSysPkg.Set.(not (is_empty notfound) && is_empty needed) then
+      (OpamConsole.error
+         "These packages are still missing and not found via \
+          your system package manager: %s\n"
+         (syspkgs_to_string notfound);
+       manual_install t still_missing)
+    else
+      (OpamConsole.error "These packages are still missing: %s\n"
+         (syspkgs_to_string sys_packages);
+       if OpamStd.Sys.tty_in then entry_point t sys_packages
+       else give_up ())
+  and bypass t =
+    OpamConsole.note
+      "Run 'opam option depext=false' if you wish to permanently disable \
+       handling of system packages.\n";
+    t
+  and give_up_msg () =
+    OpamConsole.note
+      "You can retry with '--assume-depexts' to skip this check, or run 'opam \
+       option depext=false' to permanently disable handling of system \
+       packages.\n%s"
+      (if OpamStd.Sys.tty_in || OpamCoreConfig.answer_is `unsafe_yes then ""
+       else "Running the system package manager non-interactively requires \
+             '--confirm-level=unsafe-yes'.\n")
+  and give_up () =
+    give_up_msg ();
+    OpamStd.Sys.exit_because `Aborted
   in
+  if OpamSysPkg.Set.is_empty sys_packages ||
+     OpamClientConfig.(!r.show) ||
+     OpamClientConfig.(!r.assume_depexts) then
+    t
+  else
+  try
+    OpamConsole.header_msg "Handling external dependencies";
+    OpamConsole.msg "\n";
+    entry_point t sys_packages
+  with Sys.Break as e -> OpamStd.Exn.finalise e give_up_msg
+
+let install_depexts ?(force_depext=false) ?(confirm=true) t packages =
   let map_sysmap f t =
     let sys_packages =
       OpamPackage.Set.fold (fun nv sys_map ->
@@ -1086,80 +1296,55 @@ let install_depexts ?(force_depext=false) ?(confirm=true) t packages =
     in
     { t with sys_packages = lazy sys_packages }
   in
-  let recheck t sys_packages =
-    let needed, _notfound = OpamSysInteract.packages_status sys_packages in
-    let installed = OpamSysPkg.Set.diff sys_packages needed in
-    map_sysmap (fun sysp -> OpamSysPkg.Set.diff sysp installed) t, needed
+  let confirm =
+    confirm && not (OpamSysInteract.Cygwin.is_internal t.switch_global.config)
   in
-  let rec wait msg sys_packages =
-    let give_up () =
-      OpamConsole.formatted_msg
-        "You can retry with '--assume-depexts' to skip this check, or run \
-         'opam option depext=false' to permanently disable handling of \
-         system packages altogether.\n";
-      OpamStd.Sys.exit_because `Aborted
-    in
-    if not (OpamStd.Sys.tty_in && OpamCoreConfig.answer_is `ask) then
-      give_up ()
-    else if OpamConsole.confirm
-        "%s\nWhen you are done: check again and continue?"
-        msg
-    then
-      let t, to_install = recheck t sys_packages in
-      if OpamSysPkg.Set.is_empty to_install then t else
-      let msg =
-        Printf.sprintf
-          "\nThe following remain to be installed: %s"
-          (syspkgs_to_string to_install)
-      in
-      wait msg to_install
-    else if
-      OpamConsole.confirm ~default:false
-        "Do you want to attempt to proceed anyway?"
-    then t
-    else give_up ()
+  let sys_packages =
+    get_depexts ~force:force_depext ~recover:force_depext t packages
   in
-  OpamConsole.header_msg "Handling external dependencies";
-  if not (OpamFile.Config.depext_run_installs t.switch_global.config) then
-    (print ();
-     wait "You may now install the packages on your system."
-       sys_packages)
-  else if OpamClientConfig.(!r.fake) then (print (); t)
-  else if
-    not confirm
-    || OpamConsole.confirm ~require_unsafe_yes:true
-      "Let opam run your package manager to install the required system \
-       packages?\n(answer 'n' for other options)"
-  then
-    try
-      OpamSysInteract.install sys_packages; (* handles dry_run *)
-      map_sysmap (fun _ -> OpamSysPkg.Set.empty) t
-    with Failure msg ->
-      OpamConsole.error "%s" msg;
-      if not confirm then (print (); t) else
-        wait "You can now try to get them installed manually."
-          sys_packages
-  else
-    (OpamConsole.note "Use 'opam option depext-run-installs=false' \
-                       if you don't want to be prompted again.";
-     print ();
-     wait
-       "You may now install the packages manually on your system."
-       sys_packages)
+  let env = t.switch_global.global_variables in
+  let config = t.switch_global.config in
+  install_sys_packages ~map_sysmap ~confirm env config sys_packages t
+
+let install_sys_packages ~confirm =
+  install_sys_packages ~map_sysmap:(fun _ () -> ()) ~confirm
 
 (* Apply a solution *)
-let apply ?ask t ~requested ?add_roots ?(assume_built=false)
-    ?(download_only=false) ?force_remove solution =
+let apply ?ask t ~requested ?print_requested ?add_roots
+    ?(skip=OpamPackage.Map.empty)
+    ?(assume_built=false)
+    ?(download_only=false) ?force_remove solution0 =
+  let names = OpamPackage.names_of_packages requested in
+  let print_requested = OpamStd.Option.default names print_requested in
   log "apply";
+  let solution =
+    OpamSolver.filter_solution ~recursive:false
+      (fun nv -> not (OpamPackage.Map.mem nv skip))
+      solution0
+  in
   if OpamSolver.solution_is_empty solution then
-    (* The current state satisfies the request contraints *)
+    (* The current state satisfies the request contraints,
+       but there might be depexts missing *)
+    let virt_inst =
+      OpamPackage.Set.Op.((t.installed %% requested) ++ (OpamPackage.keys skip))
+    in
+    let t =
+      if OpamClientConfig.(!r.show) then
+        let _ = get_depexts t virt_inst in t
+        (* Prints the msg about additional depexts to install *)
+      else install_depexts t virt_inst
+    in
     t, Nothing_to_do
   else (
     (* Otherwise, compute the actions to perform *)
-    let stats = OpamSolver.stats solution in
     let show_solution = ask <> Some false in
     let action_graph = OpamSolver.get_atomic_action_graph solution in
     let new_state = simulate_new_state t action_graph in
+    let new_state0 =
+      { new_state with installed =
+                         OpamPackage.Set.union new_state.installed
+                           (OpamPackage.keys skip) }
+    in
     OpamPackage.Set.iter
       (fun p ->
          try OpamFile.OPAM.print_errors (OpamSwitchState.opam new_state p)
@@ -1184,26 +1369,33 @@ let apply ?ask t ~requested ?add_roots ?(assume_built=false)
           else None
         )  messages in
       let append nv =
-        (* mark pinned packages with a star *)
-        if OpamPackage.Set.mem nv t.pinned then "*"
-        else ""
+        let pinned =
+          if OpamPackage.Set.mem nv t.pinned then " (pinned)"
+          else ""
+        and deprecated =
+          let opam = OpamSwitchState.opam new_state nv in
+          if OpamFile.OPAM.has_flag Pkgflag_Deprecated opam then " (deprecated)"
+          else ""
+        in
+        pinned ^ deprecated
       in
       OpamSolver.print_solution ~messages ~append
-        ~requested ~reinstall:(Lazy.force t.reinstall)
-        solution;
-      let total_actions = sum stats in
-      if total_actions >= 2 then
-        OpamConsole.msg "===== %s =====\n" (OpamSolver.string_of_stats stats);
+        ~requested:print_requested ~reinstall:(Lazy.force t.reinstall)
+        ~available:(Lazy.force t.available_packages)
+        ~skip
+        solution0;
     );
-    if not OpamClientConfig.(!r.show) &&
-       (download_only || confirmation ?ask requested action_graph)
-    then (
+    if OpamClientConfig.(!r.show) then
+      let _ = get_depexts t new_state0.installed in
+      (* Prints the msg about additional depexts to install *)
+      t, Aborted
+    else if download_only || confirmation ?ask names solution then (
       let t =
-        install_depexts t @@ OpamPackage.Set.inter
-          new_state.installed (OpamSolver.all_packages solution)
+        install_depexts t  @@ OpamPackage.Set.inter
+          new_state0.installed (OpamSolver.all_packages solution0)
       in
       let requested =
-        OpamPackage.packages_of_names new_state.installed requested
+        OpamPackage.packages_of_names new_state.installed names
       in
       let run_job =
         if OpamStateConfig.(!r.dryrun) || OpamClientConfig.(!r.fake)
@@ -1273,23 +1465,21 @@ let apply ?ask t ~requested ?add_roots ?(assume_built=false)
       t, Aborted
   )
 
-let resolve t action ~orphans ?reinstall ~requested request =
-  if OpamClientConfig.(!r.json_out <> None) then (
-    OpamJson.append "command-line"
-      (`A (List.map (fun s -> `String s) (Array.to_list Sys.argv)));
-    OpamJson.append "switch" (OpamSwitch.to_json t.switch)
-  );
+let resolve t action ?reinstall ~requested request =
+  if OpamClientConfig.(!r.json_out <> None) then
+    OpamJson.append "switch" (OpamSwitch.to_json t.switch);
+  OpamRepositoryState.check_last_update ();
   let universe =
     OpamSwitchState.universe t ~requested ?reinstall action
   in
   Json.output_request request action;
-  let r = OpamSolver.resolve universe ~orphans request in
+  let r = OpamSolver.resolve universe request in
   Json.output_solution t r;
   r
 
-let resolve_and_apply ?ask t action ~orphans ?reinstall ~requested ?add_roots
-    ?(assume_built=false) ?download_only ?force_remove request =
-  match resolve t action ~orphans ?reinstall ~requested request with
+let resolve_and_apply ?ask t action ?reinstall ~requested ?print_requested
+    ?add_roots ?(assume_built=false) ?download_only ?force_remove request =
+  match resolve t action ?reinstall ~requested request with
   | Conflicts cs ->
     log "conflict!";
     OpamConsole.msg "%s"
@@ -1299,7 +1489,8 @@ let resolve_and_apply ?ask t action ~orphans ?reinstall ~requested ?add_roots
   | Success solution ->
     let t, res =
       apply ?ask t
-        ~requested ?add_roots ~assume_built ?download_only ?force_remove
+        ~requested ?print_requested ?add_roots ~assume_built
+        ?download_only ?force_remove
         solution
     in
     t, Success res
