@@ -9,8 +9,6 @@
 (*                                                                        *)
 (**************************************************************************)
 
-open OpamCompat
-
 module type SET = sig
   include Set.S
   val map: (elt -> elt) -> t -> t
@@ -18,9 +16,10 @@ module type SET = sig
   val choose_one : t -> elt
   val choose_opt : t -> elt option
   val of_list: elt list -> t
+  val to_list_map: (elt -> 'b) -> t -> 'b list
   val to_string: t -> string
-  val to_json: t -> OpamJson.t
-  val of_json: OpamJson.t -> t option
+  val to_json: t OpamJson.encoder
+  val of_json: t OpamJson.decoder
   val find: (elt -> bool) -> t -> elt
   val find_opt: (elt -> bool) -> t -> elt option
   val safe_add: elt -> t -> t
@@ -36,8 +35,8 @@ end
 module type MAP = sig
   include Map.S
   val to_string: ('a -> string) -> 'a t -> string
-  val to_json: ('a -> OpamJson.t) -> 'a t -> OpamJson.t
-  val of_json: (OpamJson.t -> 'a option) -> OpamJson.t -> 'a t option
+  val to_json: 'a OpamJson.encoder -> 'a t OpamJson.encoder
+  val of_json: 'a OpamJson.decoder -> 'a t OpamJson.decoder
   val keys: 'a t -> key list
   val values: 'a t -> 'a list
   val find_opt: key -> 'a t -> 'a option
@@ -49,13 +48,16 @@ module type MAP = sig
   val update: key -> ('a -> 'a) -> 'a -> 'a t -> 'a t
   val map_reduce:
     ?default:'b -> (key -> 'a -> 'b) -> ('b -> 'b -> 'b) -> 'a t -> 'b
+  val filter_map: (key -> 'a -> 'b option) -> 'a t -> 'b t
 end
 module type ABSTRACT = sig
   type t
+  val compare: t -> t -> int
+  val equal: t -> t -> bool
   val of_string: string -> t
   val to_string: t -> string
-  val to_json: t -> OpamJson.t
-  val of_json: OpamJson.t -> t option
+  val to_json: t OpamJson.encoder
+  val of_json: t OpamJson.decoder
   module Set: SET with type elt = t
   module Map: MAP with type key = t
 end
@@ -63,8 +65,19 @@ end
 module type OrderedType = sig
   include Set.OrderedType
   val to_string: t -> string
-  val to_json: t -> OpamJson.t
-  val of_json: OpamJson.t -> t option
+  val to_json: t OpamJson.encoder
+  val of_json: t OpamJson.decoder
+end
+
+module OpamCompare = struct
+  external compare : 't -> 't -> int = "%compare"
+  external equal : 't -> 't -> bool = "%equal"
+  external (=) : 't -> 't -> bool = "%equal"
+  external (<>) : 't -> 't -> bool = "%notequal"
+  external (<) : 't -> 't -> bool = "%lessthan"
+  external (>) : 't -> 't -> bool = "%greaterthan"
+  external (<=) : 't -> 't -> bool = "%lessequal"
+  external (>=) : 't -> 't -> bool = "%greaterequal"
 end
 
 let max_print = 100
@@ -112,15 +125,13 @@ module OpamList = struct
   let to_string f =
     concat_map ~left:"{ " ~right:" }" ~nil:"{}" ", " f
 
-  let rec remove_duplicates_eq eq = function
-    | a::(b::_ as r) when eq a b -> remove_duplicates_eq eq r
-    | a::r -> a::remove_duplicates_eq eq r
+  let rec remove_duplicates eq = function
+    | a::(b::_ as r) when eq a b -> remove_duplicates eq r
+    | a::r -> a::remove_duplicates eq r
     | [] -> []
 
-  let remove_duplicates l = remove_duplicates_eq ( = ) l
-
   let sort_nodup cmp l =
-    remove_duplicates_eq (fun a b -> cmp a b = 0) (List.sort cmp l)
+    remove_duplicates (fun a b -> cmp a b = 0) (List.sort cmp l)
 
   let filter_map f l =
     let rec loop accu = function
@@ -139,6 +150,12 @@ module OpamList = struct
       | Some r -> r
       | None -> find_map f r
 
+  let rec find_map_opt f = function
+    | [] -> None
+    | x::r -> match f x with
+      | Some r -> Some r
+      | None -> find_map_opt f r
+
   let insert comp x l =
     let rec aux = function
       | [] -> [x]
@@ -151,28 +168,63 @@ module OpamList = struct
     | l when index <= 0 -> value :: l
     | x::l -> x :: insert_at (index - 1) value l
 
-  let rec assoc_opt x = function
-      [] -> None
-    | (a,b)::l -> if compare a x = 0 then Some b else assoc_opt x l
+  let rec assoc eq x = function
+    | [] -> raise Not_found
+    | (a,b)::r -> if eq a x then b else assoc eq x r
 
-  let pick_assoc x l =
+  let rec assoc_opt eq x = function
+    |  [] -> None
+    | (a,b)::l -> if eq a x then Some b else assoc_opt eq x l
+
+  let pick f l =
+    let rec aux acc = function
+      | [] -> None, l
+      | x::l ->
+        if f x then Some x, List.rev_append acc l
+        else aux (x::acc) l
+    in
+    aux [] l
+
+  let pick_assoc eq x l =
     let rec aux acc = function
       | [] -> None, l
       | (k,v) as b::r ->
-        if k = x then Some v, List.rev_append acc r
+        if eq k x then Some v, List.rev_append acc r
         else aux (b::acc) r
     in
     aux [] l
 
-  let update_assoc k v l =
+  let rec mem_assoc eq x = function
+    | [] -> false
+    | (a,_)::r -> eq a x || mem_assoc eq x r
+
+  let update_assoc eq k v l =
     let rec aux acc = function
       | [] -> List.rev ((k,v)::acc)
       | (k1,_) as b::r ->
-        if k1 = k then List.rev_append acc ((k,v)::r)
+        if eq k1 k then List.rev_append acc ((k,v)::r)
         else aux (b::acc) r
     in
     aux [] l
 
+  let remove_assoc eq k l =
+    let rec aux acc = function
+      | [] -> List.rev acc
+      | (k1,_) as b::r ->
+        if eq k1 k then List.rev_append acc r
+        else aux (b::acc) r
+    in
+    aux [] l
+
+  let fold_left_map f s l =
+    let s, l_rev =
+      List.fold_left (fun (s, l_rev) x ->
+          let s, y = f s x in
+          s, y :: l_rev)
+        (s, []) l
+    in
+    s, List.rev l_rev
+
 end
 
 
@@ -206,6 +258,9 @@ module Set = struct
     let of_list l =
       List.fold_left (fun set e -> add e set) empty l
 
+    let to_list_map f set =
+      fold (fun x acc -> f x :: acc) set []
+
     let to_string s =
       if S.cardinal s > max_print then
         Printf.sprintf "%d elements" (S.cardinal s)
@@ -303,6 +358,13 @@ module Map = struct
           add key (f key value) map
         ) map empty
 
+    let filter_map f map =
+      fold (fun key value map ->
+          match f key value with
+          | Some value -> add key value map
+          | None -> map
+        ) map empty
+
     let values map =
       List.rev (M.fold (fun _ v acc -> v :: acc) map [])
 
@@ -345,8 +407,8 @@ module Map = struct
             let get_pair = function
               | `O binding ->
                 begin match
-                    O.of_json (List.assoc "key" binding),
-                    value_of_json (List.assoc "value" binding)
+                    O.of_json (OpamList.assoc String.equal "key" binding),
+                    value_of_json (OpamList.assoc String.equal "value" binding)
                   with
                   | Some key, Some value -> (key, value)
                   | _ -> raise Not_found
@@ -386,6 +448,8 @@ end
 
 module AbstractString = struct
   type t = string
+  let compare = String.compare
+  let equal = String.equal
   let of_string x = x
   let to_string x = x
   let to_json x = `String x
@@ -406,7 +470,7 @@ end
 
 module OInt = struct
   type t = int
-  let compare = compare
+  let compare = Int.compare
   let to_string = string_of_int
   let to_json i = `String (string_of_int i)
   let of_json = function
@@ -444,12 +508,19 @@ module Option = struct
     | None -> dft
     | Some x -> f x
 
-  let compare cmp o1 o2 = match o1,o2 with
+  let compare cmp o1 o2 =
+    match o1,o2 with
     | None, None -> 0
     | Some _, None -> 1
     | None, Some _ -> -1
     | Some x1, Some x2 -> cmp x1 x2
 
+  let equal f o1 o2 =
+    match o1, o2 with
+    | Some o1, Some o2 -> f o1 o2
+    | None, None -> true
+    | _ , _  -> false
+
   let to_string ?(none="") f = function
     | Some x -> f x
     | None -> none
@@ -485,7 +556,7 @@ module OpamString = struct
 
   module OString = struct
     type t = string
-    let compare = compare
+    let compare = String.compare
     let to_string x = x
     let to_json x = `String x
     let of_json = function
@@ -516,6 +587,11 @@ module OpamString = struct
     let rec chk i = i >= x || suffix.[i] = s.[i+n-x] && chk (i+1) in
     chk 0
 
+  let for_all f s =
+    let len = String.length s in
+    let rec aux i = i >= len || f s.[i] && aux (i+1) in
+    aux 0
+
   let contains_char s c =
     try let _ = String.index s c in true
     with Not_found -> false
@@ -634,6 +710,26 @@ module OpamString = struct
       |_ -> []
     in List.rev (aux acc0 tokens)
 
+  let split_quoted path sep =
+    let length = String.length path in
+    let rec f acc index current last normal =
+      if (index : int) = length then
+        let current = current ^ String.sub path last (index - last) in
+        List.rev (if current <> "" then current::acc else acc)
+      else
+      let c = path.[index]
+      and next = succ index in
+      if (c : char) = sep && normal || c = '"' then
+        let current = current ^ String.sub path last (index - last) in
+        if c = '"' then
+          f acc next current next (not normal)
+        else
+        let acc = if current = "" then acc else current::acc in
+        f acc next "" next true
+      else
+        f acc next current last normal in
+    f [] 0 "" 0 true
+
   let fold_left f acc s =
     let acc = ref acc in
     for i = 0 to String.length s - 1 do acc := f !acc s.[i] done;
@@ -669,6 +765,15 @@ module OpamString = struct
       && length_s > from
       && String.sub full 0 length_s = s
 
+  let is_hex s =
+    try
+      String.iter (function
+          | '0'..'9' | 'A'..'F' | 'a'..'f' -> ()
+          | _ -> raise Exit)
+        s;
+      true
+    with Exit -> false
+
 end
 
 type warning_printer =
@@ -695,32 +800,88 @@ module Env = struct
       | curr::after -> aux (curr::before) after
     in aux [] v
 
-  let list =
-    let lazy_env = lazy (
-      let e = Unix.environment () in
-      List.rev_map (fun s ->
-          match OpamString.cut_at s '=' with
-          | None   -> s, ""
-          | Some p -> p
-        ) (Array.to_list e)
-    ) in
-    fun () -> Lazy.force lazy_env
-
-  let get =
-    if Sys.win32 then
-      fun n ->
-        let n = String.uppercase_ascii n in
-        snd (List.find (fun (k,_) -> String.uppercase_ascii k = n) (list ()))
-    else
-      fun n -> List.assoc n (list ())
-
-  let getopt n = try Some (get n) with Not_found -> None
-
   let escape_single_quotes ?(using_backslashes=false) =
     if using_backslashes then
       Re.(replace (compile (set "\\\'")) ~f:(fun g -> "\\"^Group.get g 0))
     else
       Re.(replace_string (compile (char '\'')) ~by:"'\"'\"'")
+
+  let escape_powershell =
+    (* escape single quotes with two single quotes.
+       https://docs.microsoft.com/en-us/powershell/module/microsoft.powershell.core/about/about_quoting_rules?view=powershell-7.1 *)
+    Re.(replace_string (compile (char '\'')) ~by:"''")
+
+  module Name = struct
+    module M = struct
+      include AbstractString
+
+      let compare =
+        if Sys.win32 then
+          fun l r ->
+            String.(compare (lowercase_ascii l) (lowercase_ascii r))
+        else
+          String.compare
+    end
+
+    type t = string
+
+    let of_string = M.of_string
+    let to_string = M.to_string
+    let of_json = M.of_json
+    let to_json = M.to_json
+    let compare = M.compare
+
+    let equal =
+      if Sys.win32 then
+        fun l r ->
+          String.(equal (lowercase_ascii l) (lowercase_ascii r))
+      else
+        String.equal
+
+    let equal_string = equal
+
+    module Set = Set.Make(M)
+    module Map = Map.Make(M)
+  end
+
+  let to_list env =
+    List.rev_map (fun s ->
+        match OpamString.cut_at s '=' with
+        | None   -> s, ""
+        | Some p -> p)
+      (Array.to_list env)
+
+  let raw_env = Unix.environment
+
+  let list =
+    let lazy_env = lazy (to_list (raw_env ())) in
+    fun () -> Lazy.force lazy_env
+
+  let cyg_env ~env ~cygbin ~git_location =
+    let f v =
+      match OpamString.cut_at v '=' with
+      | Some (path, c) when Name.equal_string path "path" ->
+        (match git_location with
+         | None ->
+           Printf.sprintf "%s=%s;%s" path cygbin c
+         | Some git_location ->
+           if String.equal git_location cygbin then
+             Printf.sprintf "%s=%s;%s" path cygbin c
+           else
+             Printf.sprintf "%s=%s;%s;%s" path git_location cygbin c)
+      | _ -> v
+    in
+    Array.map f env
+
+  let get_full n = List.find (fun (k,_) -> Name.equal k n) (list ())
+
+  let get n = snd (get_full n)
+
+  let getopt = Option.of_Not_found get
+
+  let getopt_full n =
+    try let (n, v) = get_full n in (n, Some v)
+    with Not_found -> (n, None)
 end
 
 
@@ -736,24 +897,8 @@ module OpamSys = struct
   let path_sep = if Sys.win32 then ';' else ':'
 
   let split_path_variable ?(clean=true) =
-    if Sys.win32 then fun path ->
-      let length = String.length path in
-      let rec f acc index current last normal =
-        if index = length then
-          let current = current ^ String.sub path last (index - last) in
-          List.rev (if current <> "" then current::acc else acc)
-        else let c = path.[index]
-          and next = succ index in
-          if c = ';' && normal || c = '"' then
-            let current = current ^ String.sub path last (index - last) in
-            if c = '"' then
-              f acc next current next (not normal)
-            else
-            let acc = if current = "" then acc else current::acc in
-            f acc next "" next true
-          else
-            f acc next current last normal in
-      f [] 0 "" 0 true
+    if Sys.win32 then
+      fun path -> OpamString.split_quoted path ';'
     else fun path ->
       let split = if clean then OpamString.split else OpamString.split_delim in
       split path path_sep
@@ -838,7 +983,14 @@ module OpamSys = struct
        forcing the environment in this function that is used
        before the .init() functions are called -- see
        OpamStateConfig.default. *)
-    let home = lazy (try Unix.getenv "HOME" with Not_found -> Sys.getcwd ()) in
+    let home = lazy (
+      try Unix.getenv "HOME"
+      with Not_found ->
+        if Sys.win32 then
+          OpamStubs.getPathToHome ()
+        else
+          Sys.getcwd ()
+    ) in
     fun () -> Lazy.force home
 
   let etc () = "/etc"
@@ -856,9 +1008,9 @@ module OpamSys = struct
         Hashtbl.add memo arg r;
         r
 
-  let system () =
-    (* CSIDL_SYSTEM = 0x25 *)
-    OpamStubs.(shGetFolderPath 0x25 SHGFP_TYPE_CURRENT)
+  let system =
+    let system = Lazy.from_fun OpamStubs.getPathToSystem in
+    fun () -> Lazy.force system
 
   type os =
     | Darwin
@@ -891,7 +1043,21 @@ module OpamSys = struct
     ) in
     fun () -> Lazy.force os
 
+  type powershell_host = Powershell_pwsh | Powershell
   type shell = SH_sh | SH_bash | SH_zsh | SH_csh | SH_fish
+    | SH_pwsh of powershell_host | SH_cmd
+
+  let all_shells =
+    [SH_sh; SH_bash;
+     SH_zsh;
+     SH_csh;
+     SH_fish;
+     SH_pwsh Powershell_pwsh;
+     SH_pwsh Powershell;
+     SH_cmd]
+
+  let windows_default_shell = SH_cmd
+  let unix_default_shell = SH_sh
 
   let shell_of_string = function
     | "tcsh"
@@ -900,6 +1066,8 @@ module OpamSys = struct
     | "zsh"  -> Some SH_zsh
     | "bash" -> Some SH_bash
     | "fish" -> Some SH_fish
+    | "pwsh" -> Some (SH_pwsh Powershell_pwsh)
+    | "dash"
     | "sh"   -> Some SH_sh
     | _      -> None
 
@@ -913,48 +1081,82 @@ module OpamSys = struct
     else
       fun x -> x
 
+  let chop_exe_suffix name =
+    Option.default name (Filename.chop_suffix_opt name ~suffix:".exe")
+
+  let windows_process_ancestry = Lazy.from_fun OpamStubs.getProcessAncestry
+
+  type shell_choice = Accept of shell
+
+  let windows_get_shell =
+    let categorize_process (_, image) =
+      match String.lowercase_ascii (Filename.basename image) with
+      | "powershell.exe" | "powershell_ise.exe" ->
+        Some (Accept (SH_pwsh Powershell))
+      | "pwsh.exe" -> Some (Accept (SH_pwsh Powershell_pwsh))
+      | "cmd.exe" -> Some (Accept SH_cmd)
+      | "" -> None
+      | name ->
+        Option.map
+          (fun shell -> Accept shell)
+          (shell_of_string (chop_exe_suffix name))
+    in
+    lazy (
+      let lazy ancestors = windows_process_ancestry in
+      match OpamList.filter_map categorize_process ancestors with
+      | [] -> None
+      | Accept most_relevant_shell :: _ -> Some most_relevant_shell
+    )
+
   let guess_shell_compat () =
-    let parent_guess =
-      if Sys.win32 then
-        None
-      else
-        let ppid = Unix.getppid () in
-        let dir = Filename.concat "/proc" (string_of_int ppid) in
+    let parent_guess () =
+      let ppid = Unix.getppid () in
+      let dir = Filename.concat "/proc" (string_of_int ppid) in
+      try
+        Some (Unix.readlink (Filename.concat dir "exe"))
+      with e ->
+        fatal e;
         try
-          Some (Unix.readlink (Filename.concat dir "exe"))
-        with e ->
-          fatal e;
-          try
-            with_process_in "ps"
-              (Printf.sprintf "-p %d -o comm= 2>/dev/null" ppid)
-              (fun ic -> Some (input_line ic))
-          with
-          | Unix.Unix_error _ | Sys_error _ | Failure _ | End_of_file | Not_found ->
-              try
-                let c = open_in_bin ("/proc/" ^ string_of_int ppid ^ "/cmdline") in
-                begin try
-                  let s = input_line c in
+          with_process_in "ps"
+            (Printf.sprintf "-p %d -o comm= 2>/dev/null" ppid)
+            (fun ic -> Some (input_line ic))
+        with
+        | Unix.Unix_error _ | Sys_error _ | Failure _ | End_of_file | Not_found ->
+            try
+              let c = open_in_bin ("/proc/" ^ string_of_int ppid ^ "/cmdline") in
+              begin try
+                let s = input_line c in
+                close_in c;
+                Some (String.sub s 0 (String.index s '\000'))
+              with
+              | Not_found ->
+                  None
+              | e ->
                   close_in c;
-                  Some (String.sub s 0 (String.index s '\000'))
-                with
-                | Not_found ->
-                    None
-                | e ->
-                    close_in c;
-                    fatal e; None
-                end
-              with e ->
-                fatal e; None
+                  fatal e; None
+              end
+            with e ->
+              fatal e; None
     in
     let test shell = shell_of_string (Filename.basename shell) in
-    let shell =
-      match Option.replace test parent_guess with
-      | None ->
+    if Sys.win32 then
+      let shell =
+        match Lazy.force windows_get_shell with
+        | None ->
           Option.of_Not_found Env.get "SHELL" |> Option.replace test
-      | some ->
+        | some ->
           some
-    in
-    Option.default SH_sh shell
+        in
+      Option.default windows_default_shell shell
+    else
+      let shell =
+        match Option.replace test (parent_guess ()) with
+        | None ->
+            Option.of_Not_found Env.get "SHELL" |> Option.replace test
+        | some ->
+            some
+      in
+      Option.default unix_default_shell shell
 
   let guess_dot_profile shell =
     let home f =
@@ -962,30 +1164,37 @@ module OpamSys = struct
       with Not_found -> f in
     match shell with
     | SH_fish ->
-      List.fold_left Filename.concat (home ".config") ["fish"; "config.fish"]
-    | SH_zsh  -> home ".zshrc"
+      Some (List.fold_left Filename.concat (home ".config") ["fish"; "config.fish"])
+    | SH_zsh  -> Some (home ".zshrc")
     | SH_bash ->
-      (try
-         List.find Sys.file_exists [
-           (* Bash looks up these 3 files in order and only loads the first,
-              for LOGIN shells *)
-           home ".bash_profile";
-           home ".bash_login";
-           home ".profile";
-           (* Bash loads .bashrc INSTEAD, for interactive NON login shells only;
-              but it's often included from the above.
-              We may include our variables in both to be sure ; for now we rely
-              on non-login shells inheriting their env from a login shell
-              somewhere... *)
-         ]
-       with Not_found ->
-         (* iff none of the above exist, creating this should be safe *)
-         home ".bash_profile")
+      let shell =
+        (try
+           List.find Sys.file_exists [
+             (* Bash looks up these 3 files in order and only loads the first,
+                for LOGIN shells *)
+             home ".bash_profile";
+             home ".bash_login";
+             home ".profile";
+             (* Bash loads .bashrc INSTEAD, for interactive NON login shells only;
+                but it's often included from the above.
+                We may include our variables in both to be sure ; for now we rely
+                on non-login shells inheriting their env from a login shell
+                somewhere... *)
+           ]
+         with Not_found ->
+           (* iff none of the above exist, creating this should be safe *)
+           home ".bash_profile")
+      in
+      Some shell
     | SH_csh ->
       let cshrc = home ".cshrc" in
       let tcshrc = home ".tcshrc" in
-      if Sys.file_exists cshrc then cshrc else tcshrc
-    | SH_sh -> home ".profile"
+      Some (if Sys.file_exists cshrc then cshrc else tcshrc)
+    | SH_pwsh _ ->
+      None
+    | SH_sh -> Some (home ".profile")
+    | SH_cmd ->
+      None
 
 
   let registered_at_exit = ref []
@@ -997,44 +1206,201 @@ module OpamSys = struct
       (fun f -> try f () with _ -> ())
       !registered_at_exit
 
-  let is_cygwin_variant =
+  let env_var env var =
+    let len = Array.length env in
+    let f = if Sys.win32 then String.uppercase_ascii else fun x -> x in
+    let prefix = f var^"=" in
+    let pfxlen = String.length prefix in
+    let rec aux i =
+      if (i : int) >= len then "" else
+      let s = env.(i) in
+      if OpamString.starts_with ~prefix (f s) then
+        String.sub s pfxlen (String.length s - pfxlen)
+      else aux (i+1)
+    in
+    aux 0
+
+  let is_external_cmd name =
+    let forward_to_back =
+      if Sys.win32 then
+        String.map (function '/' -> '\\' | c -> c)
+      else
+        fun x -> x
+    in
+    let name = forward_to_back name in
+    OpamString.contains_char name Filename.dir_sep.[0]
+
+  let resolve_in_path_t env name =
+    if not (Filename.is_relative name) || is_external_cmd name then
+      invalid_arg "OpamStd.Sys.resolve_in_path: bare command expected"
+    else
+    let path = split_path_variable (env_var env "PATH") in
+    List.filter_map (fun path ->
+        let candidate = Filename.concat path name in
+        (* TODO: use Sys.is_regular_file once opam requires OCaml >= 5.1 *)
+        match Sys.is_directory candidate with
+        | false -> Some candidate
+        | true | exception (Sys_error _) -> None)
+      path
+
+  (* OCaml 4.05.0 no longer follows the updated PATH to resolve commands. This
+     makes unqualified commands absolute as a workaround. *)
+  let resolve_command =
+    let check_perms =
+      if Sys.win32 then fun f ->
+        try (Unix.stat f).Unix.st_kind = Unix.S_REG
+        with e -> fatal e; false
+      else fun f ->
+        try
+          let {Unix.st_uid; st_gid; st_perm; st_kind; _} = Unix.stat f in
+          if st_kind <> Unix.S_REG then false else
+          let groups =
+            IntSet.of_list (Unix.getegid () :: Array.to_list (Unix.getgroups ()))
+          in
+          let mask =
+            if Unix.geteuid () = (st_uid : int) then
+              0o100
+            else if IntSet.mem st_gid groups then
+              0o010
+            else
+              0o001
+          in
+          if (st_perm land mask) <> 0 then
+            true
+          else
+          match OpamACL.get_acl_executable_info f st_uid with
+          | None -> false
+          | Some [] -> true
+          | Some gids ->
+            not (IntSet.is_empty (IntSet.inter (IntSet.of_list gids) groups))
+        with e -> fatal e; false
+    in
+    let resolve ?dir env name =
+      if not (Filename.is_relative name) then begin
+        (* absolute path *)
+        if not (Sys.file_exists name) then `Not_found
+        else if not (check_perms name) then `Denied
+        else `Cmd name
+      end else if is_external_cmd name then begin
+        (* relative path *)
+        let cmd = match dir with
+          | None -> name
+          | Some d -> Filename.concat d name
+        in
+        if not (Sys.file_exists cmd) then `Not_found
+        else if not (check_perms cmd) then `Denied
+        else `Cmd cmd
+      end else
+      (* bare command, lookup in PATH *)
+      let name =
+        if Sys.win32 && not (Filename.check_suffix name ".exe") then
+          name ^ ".exe"
+        else name
+      in
+      let possibles = resolve_in_path_t env name in
+      (* Following the shell sematics for looking up PATH, programs with the
+         expected name but not the right permissions are skipped silently.
+         Therefore, only two outcomes are possible in that case, [`Cmd ..] or
+         [`Not_found]. *)
+      match List.find check_perms possibles with
+      | cmdname -> `Cmd cmdname
+      | exception Not_found ->
+        if possibles = [] then
+          `Not_found
+        else
+          `Denied
+    in
+    fun ?env ?dir name ->
+      let env = match env with None -> Env.raw_env () | Some e -> e in
+      resolve env ?dir name
+
+  let resolve_in_path ?env name =
+    let env = match env with None -> Env.raw_env () | Some e -> e in
+    match resolve_in_path_t env name with
+    | result::_ -> Some result
+    | [] -> None
+
+  let get_windows_executable_variant =
     if Sys.win32 then
       let results = Hashtbl.create 17 in
-      let requires_cygwin name =
-        let cmd = Printf.sprintf "cygcheck \"%s\"" name in
-        let ((c, _, _) as process) = Unix.open_process_full cmd (Unix.environment ()) in
-        let rec f a =
-          match input_line c with
-          | x ->
-              if OpamString.ends_with ~suffix:"cygwin1.dll" (String.trim x) then
-                if OpamString.starts_with ~prefix:"  " x then
-                  f `Cygwin
-                else if a <> `Cygwin then
-                  f `CygLinked
-                else
-                  f a
-              else
-                f a
-          | exception e ->
-              Unix.close_process_full process |> ignore;
-              fatal e;
-              a
+      let requires_cygwin cygcheck name =
+        let env =
+          Env.cyg_env ~env:(Env.raw_env ()) ~cygbin:(Filename.dirname cygcheck)
+            ~git_location:None
         in
-        f `Native
+        let cmd = OpamCompat.Filename.quote_command cygcheck [name] in
+        let ((c, _, _) as process) = Unix.open_process_full cmd env in
+        let rec check_dll platform =
+          match input_line c with
+          | dll ->
+            (* Guard against any risk of stray \r characters *)
+            let tdll = String.trim dll in
+            if OpamString.ends_with ~suffix:"cygwin1.dll" tdll then
+              if OpamString.starts_with ~prefix:"  " dll then
+                check_dll `Cygwin
+              else if platform = `Native then
+                check_dll (`Tainted `Cygwin)
+              else
+                check_dll platform
+            else if OpamString.ends_with ~suffix:"msys-2.0.dll" tdll then
+              if OpamString.starts_with ~prefix:"  " dll then
+                check_dll `Msys2
+              else if platform = `Native then
+                check_dll (`Tainted `Msys2)
+              else
+                check_dll platform
+            else
+              check_dll platform
+          | exception e ->
+            Unix.close_process_full process |> ignore;
+            fatal e;
+            platform
+        in
+        check_dll `Native
       in
-      fun name ->
-        if Filename.is_relative name then
-          requires_cygwin name
-        else
-          try
-            Hashtbl.find results name
-          with Not_found ->
-            let result = requires_cygwin name
-            in
-              Hashtbl.add results name result;
+      fun ?search_in_first name ->
+        let cygcheck =
+          let open Option.Op in
+          let contains_cygcheck dir =
+            let cygcheck = Filename.concat dir "cygcheck.exe" in
+            if Sys.file_exists cygcheck then
+              Some cygcheck
+            else
+              None
+          in
+          search_in_first >>= contains_cygcheck
+          >>+ fun () ->
+            (* ~search_in_first not supplied, or cygcheck.exe not found in it;
+               now try general PATH-resolution. *)
+            match resolve_command "cygcheck.exe" with
+            | `Cmd cmd -> Some cmd
+            | `Not_found | `Denied -> None
+        in
+        match cygcheck with
+        | None -> `Native
+        | Some cygcheck ->
+          if Filename.is_relative name then
+            requires_cygwin cygcheck name
+          else
+            try Hashtbl.find results (cygcheck, name)
+            with Not_found ->
+              let result = requires_cygwin cygcheck name in
+              Hashtbl.add results (cygcheck, name) result;
               result
     else
-      fun _ -> `Native
+    fun ?search_in_first:_ _ -> `Native
+
+  let get_cygwin_variant ?search_in_first cmd =
+    (* Treat MSYS2's variant of `cygwin1.dll` called `msys-2.0.dll` equivalently.
+       Confer https://www.msys2.org/wiki/How-does-MSYS2-differ-from-Cygwin/ *)
+    match get_windows_executable_variant ?search_in_first cmd with
+    | `Native -> `Native
+    | `Cygwin
+    | `Msys2 -> `Cygwin
+    | `Tainted _ -> `CygLinked
+
+  let is_cygwin_variant ?search_in_first cmd =
+    get_cygwin_variant ?search_in_first cmd = `Cygwin
 
   exception Exit of int
   exception Exec of string * string array * string array
@@ -1071,7 +1437,7 @@ module OpamSys = struct
     `User_interrupt, 130;
   ]
 
-  let get_exit_code reason = List.assoc reason exit_codes
+  let get_exit_code reason = OpamList.assoc OpamCompare.equal reason exit_codes
 
   let exit_because reason = exit (get_exit_code reason)
 
@@ -1084,6 +1450,15 @@ module OpamSys = struct
       if !called then invalid_arg "Just what do you think you're doing, Dave?";
       called := true;
       console := printer
+
+  let is_valid_basename_char =
+    if Sys.win32 then
+      function
+      | '\000'..'\031'
+      | '<' | '>' | ':' | '"' | '/' | '\\' | '|' | '?' | '*' -> false
+      | _ -> true
+    else
+      fun c -> c <> '\000' && c <> '/'
 end
 
 
@@ -1111,47 +1486,58 @@ module Win32 = struct
   end
 
   let (set_parent_pid, parent_putenv) =
-    let ppid = ref (lazy (OpamStubs.(getCurrentProcessID () |> getParentProcessID))) in
+    let ppid = ref (OpamCompat.Lazy.map (function (_::(pid, _)::_) -> pid | _ -> 0l) OpamSys.windows_process_ancestry) in
     let parent_putenv = lazy (
-      let ppid = Lazy.force !ppid in
-      if OpamStubs.isWoW64 () <> OpamStubs.isWoW64Process ppid then
-        (*
-         * Expect to see opam-putenv.exe in the same directory as opam.exe,
-         * rather than PATH (allow for crazy users like developers who may have
-         * both builds of opam)
-         *)
-        let putenv_exe =
-          Filename.(concat (dirname Sys.executable_name) "opam-putenv.exe")
+      let {contents = lazy ppid} = ppid in
+      let our_architecture = OpamStubs.getProcessArchitecture None in
+      let their_architecture = OpamStubs.getProcessArchitecture (Some ppid) in
+      let no_opam_putenv =
+        let warning = lazy (
+          !console.warning "opam-putenv was not found - \
+                            OPAM is unable to alter environment variables";
+          false)
         in
-        let ctrl = ref stdout in
-        let quit_putenv () =
-          if !ctrl <> stdout then
-            let () = Printf.fprintf !ctrl "::QUIT\n%!" in
-            ctrl := stdout
-        in
-        at_exit quit_putenv;
-        if Sys.file_exists putenv_exe then
-          fun key value ->
-            if !ctrl = stdout then begin
-              let (inCh, outCh) = Unix.pipe () in
-              let _ =
-                Unix.create_process putenv_exe
-                                    [| putenv_exe; Int32.to_string ppid |]
-                                    inCh Unix.stdout Unix.stderr
-              in
-              ctrl := (Unix.out_channel_of_descr outCh);
-              set_binary_mode_out !ctrl true;
-            end;
-            Printf.fprintf !ctrl "%s\n%s\n%!" key value;
-            if key = "::QUIT" then ctrl := stdout;
-            true
-        else
-          let warning = lazy (
-            !console.warning "opam-putenv was not found - \
-                              OPAM is unable to alter environment variables";
-            false)
+        fun _ _ -> Lazy.force warning
+      in
+      if our_architecture <> their_architecture then
+        match their_architecture with
+        | OpamStubs.ARM | ARM64 | IA64 | Unknown ->
+          (* ARM support not yet implemented - just ensure we don't inject Intel
+             code into an ARM process! *)
+          no_opam_putenv
+        | AMD64 | Intel ->
+          (*
+           * Expect to see opam-putenv.exe in the same directory as opam.exe,
+           * rather than PATH (allow for crazy users like developers who may have
+           * both builds of opam)
+           *)
+          let putenv_exe =
+            Filename.(concat (dirname Sys.executable_name) "opam-putenv.exe")
           in
-          fun _ _ -> Lazy.force warning
+          let ctrl = ref stdout in
+          let quit_putenv () =
+            if !ctrl <> stdout then
+              let () = Printf.fprintf !ctrl "::QUIT\n%!" in
+              ctrl := stdout
+          in
+          at_exit quit_putenv;
+          if Sys.file_exists putenv_exe then
+            fun key value ->
+              if !ctrl = stdout then begin
+                let (inCh, outCh) = Unix.pipe () in
+                let _ =
+                  Unix.create_process putenv_exe
+                                      [| putenv_exe; Int32.to_string ppid |]
+                                      inCh Unix.stdout Unix.stderr
+                in
+                ctrl := (Unix.out_channel_of_descr outCh);
+                set_binary_mode_out !ctrl true;
+              end;
+              Printf.fprintf !ctrl "%s\n%s\n%!" key value;
+              if key = "::QUIT" then ctrl := stdout;
+              true
+          else
+            no_opam_putenv
       else
         function "::QUIT" -> fun _ -> true
         | key -> OpamStubs.process_putenv ppid key)
@@ -1413,6 +1799,7 @@ module Config = struct
   type when_ = [ `Always | `Never | `Auto ]
   type when_ext = [ `Extended | when_ ]
   type answer = [ `unsafe_yes | `all_yes | `all_no | `ask ]
+  type yes_answer = [ `unsafe_yes | `all_yes ]
 
   let env conv var =
     try Option.map conv (Env.getopt ("OPAM"^var))
@@ -1442,8 +1829,8 @@ module Config = struct
     env (function s ->
         if s = "" then 0 else
         match bool_of_string s with
-        | Some true -> 0
-        | Some false -> 1
+        | Some true -> 1
+        | Some false -> 0
         | None -> int_of_string s)
       var
 
@@ -1507,14 +1894,16 @@ module Config = struct
 
   module E = struct
     type t = ..
+    type t += REMOVED
     let (r : t list ref) = ref []
 
     let update v = r := v :: !r
     let updates l = r := l @ !r
 
     let find var = OpamList.find_map var !r
+    let value_t var = try Some (find var) with Not_found -> None
     let value var =
-      let l = lazy (try Some (find var); with Not_found -> None) in
+      let l = lazy (value_t var) in
       fun () -> Lazy.force l
   end
 
@@ -1524,3 +1913,4 @@ module List = OpamList
 module String = OpamString
 module Sys = OpamSys
 module Format = OpamFormat
+module Compare = OpamCompare
