@@ -161,6 +161,9 @@ val verifies: t -> OpamPackage.t -> bool
 (** Checks if a given set of (installed) packages satisfies a formula *)
 val satisfies_depends: OpamPackage.Set.t -> t -> bool
 
+(** Returns the set of names referred to in a formula *)
+val all_names: (OpamPackage.Name.t * 'a) formula -> OpamPackage.Name.Set.t
+
 (** Returns the subset of packages possibly matching the formula (i.e. including
     all disjunction cases) *)
 val packages: OpamPackage.Set.t -> t -> OpamPackage.Set.t
@@ -173,9 +176,15 @@ val compare_nc:
 (** Convert a formula to CNF *)
 val cnf_of_formula: 'a formula -> 'a formula
 
+(** Convert a formula to CNF, but as a nested list *)
+val formula_to_cnf: 'a formula -> 'a cnf
+
 (** Convert a formula to DNF *)
 val dnf_of_formula: 'a formula -> 'a formula
 
+(** Convert a formula to DNF, but as a nested list *)
+val formula_to_dnf: 'a formula -> 'a dnf
+
 (** Transform a formula where versions can be expressed using formulas
     to a flat atom formula *)
 val to_atom_formula: t -> atom formula
