@@ -148,16 +148,24 @@ let validate_and_add_to_cache label url cache_dir file checksums =
     OpamFilename.remove file;
     false
   with Not_found ->
-    (match cache_dir, checksums with
-     | Some dir, ck::_ ->
-       OpamFilename.copy ~src:file ~dst:(cache_file dir ck)
-       (* idea: hardlink to the other checksums? *)
+    (let checksums = OpamHash.sort checksums in
+     match cache_dir, checksums with
+     | Some dir, best_chks :: others_chks ->
+       OpamFilename.copy ~src:file ~dst:(cache_file dir best_chks);
+       List.iter (fun checksum ->
+           let target = cache_file dir best_chks in
+           let link = cache_file dir checksum in
+           try
+             OpamFilename.link ~relative:true ~target ~link
+           with Sys_error _ -> ())
+         others_chks;
      | _ -> ());
     true
 
 (* [cache_dir] used to add to cache only *)
 let pull_from_upstream
-    label ?(working_dir=false) ?subpath cache_dir destdir checksums url =
+    label ?full_fetch ?(working_dir=false) ?subpath
+    cache_dir destdir checksums url =
   let module B = (val url_backend url: OpamRepositoryBackend.S) in
   let cksum = match checksums with [] -> None | c::_ -> Some c in
   let text =
@@ -199,7 +207,7 @@ let pull_from_upstream
        )
      else url, B.pull_url
    in
-   pull ?cache_dir ?subpath destdir cksum url
+   pull ?full_fetch ?cache_dir ?subpath destdir cksum url
   )
   @@| function
   | (Result (Some file) | Up_to_date (Some file)) as ret ->
@@ -212,14 +220,17 @@ let pull_from_upstream
   | (Result None | Up_to_date None) as ret -> ret
   | Not_available _ as na -> na
 
-let pull_from_mirrors label ?working_dir ?subpath cache_dir destdir checksums urls =
+let pull_from_mirrors label ?full_fetch ?working_dir ?subpath
+    cache_dir destdir checksums urls =
   let rec aux = function
     | [] -> invalid_arg "pull_from_mirrors: empty mirror list"
     | [url] ->
-      pull_from_upstream label ?working_dir ?subpath cache_dir destdir checksums url
+      pull_from_upstream label ?full_fetch ?working_dir ?subpath
+        cache_dir destdir checksums url
       @@| fun r -> url, r
     | url::mirrors ->
-      pull_from_upstream label ?working_dir ?subpath cache_dir destdir checksums url
+      pull_from_upstream label ?full_fetch ?working_dir ?subpath
+        cache_dir destdir checksums url
       @@+ function
       | Not_available (_,s) ->
         OpamConsole.warning "%s: download of %s failed (%s), trying mirror"
@@ -237,20 +248,79 @@ let pull_from_mirrors label ?working_dir ?subpath cache_dir destdir checksums ur
     url, Not_available (Some m, m)
   | ret -> ret
 
-let pull_tree
-    label ?cache_dir ?(cache_urls=[]) ?working_dir ?subpath
-    local_dirname checksums remote_urls =
-  let extract_archive f s =
-    OpamFilename.cleandir local_dirname;
-    OpamFilename.extract_job f local_dirname @@+ function
-    | None -> Done (Up_to_date s)
-    | Some (Failure s) ->
-      Done (Not_available (Some s, "Could not extract archive:\n"^s))
-    | Some (OpamSystem.Process_error pe) ->
-      Done (Not_available (Some (OpamProcess.result_summary pe),
-                           OpamProcess.string_of_result pe))
-    | Some e -> Done (Not_available (None, Printexc.to_string e))
+(* handle subpathes *)
+let pull_tree_t
+    ?full_fetch ?cache_dir ?(cache_urls=[]) ?working_dir
+    dirnames checksums remote_urls =
+  let extract_archive =
+    let fallback success = function
+      | None -> success ()
+      | Some (Failure s) ->
+        Done (Not_available (Some s, "Could not extract archive:\n"^s))
+      | Some (OpamSystem.Process_error pe) ->
+        Done (Not_available (Some (OpamProcess.result_summary pe),
+                             OpamProcess.string_of_result pe))
+      | Some e -> Done (Not_available (None, Printexc.to_string e))
+    in
+    match dirnames with
+    | [ label, local_dirname, _subpath ] ->
+      (fun archive msg ->
+         OpamFilename.cleandir local_dirname;
+         let text = OpamProcess.make_command_text label "extract" in
+         OpamProcess.Job.with_text text @@
+         OpamFilename.extract_job archive local_dirname
+         @@+ fallback (fun () ->  Done (Up_to_date msg)))
+    | _ ->
+      fun archive msg ->
+        OpamFilename.with_tmp_dir_job @@ fun tmpdir ->
+        let copies () =
+          OpamParallel.map ~jobs:3
+            ~command:(fun (label, local_dirname, _subpath) ->
+                let text = OpamProcess.make_command_text label label in
+                OpamProcess.Job.with_text text @@
+                (try
+                   OpamFilename.cleandir local_dirname;
+                   OpamFilename.copy_dir ~src:tmpdir ~dst:local_dirname;
+                   Done (Up_to_date label)
+                 with OpamSystem.Process_error r ->
+                   Done (Not_available
+                           (Some label, OpamProcess.result_summary r))))
+            dirnames
+        in
+        let text =
+          let label =
+            match dirnames with
+            | [(label1, _, _); (label2, _, _)] ->
+              label1 ^ ", " ^ label2
+            | (label, _, _)::rest ->
+              Printf.sprintf "%s + %d others" label (List.length rest)
+            | [] -> assert false in
+          OpamProcess.make_command_text label "extract"
+        in
+        OpamProcess.Job.with_text text @@
+        OpamFilename.extract_job archive tmpdir
+        @@+ fallback (fun () ->
+            let failing =
+              OpamStd.List.filter_map (function
+                  | Result _ | Up_to_date _ -> None
+                  | Not_available (Some s,l) -> Some (s,l)
+                  | Not_available (None, _) -> assert false
+                ) (copies ())
+            in
+            if failing = [] then Done (Up_to_date msg) else
+            let simple =
+              Printf.sprintf "Failed to copy source of %s"
+                (OpamStd.Format.pretty_list (List.map fst failing))
+            in
+            let long =
+              Printf.sprintf "Failed to copy source of:\n%s"
+                (OpamStd.Format.itemize (fun (nv, msg) ->
+                     Printf.sprintf "%s: %s" nv msg)
+                    failing)
+            in
+            Done (Not_available (Some simple, long)))
   in
+  let label = OpamStd.List.concat_map ", " (fun (x,_,_) -> x) dirnames in
   (match cache_dir with
    | Some cache_dir ->
      let text = OpamProcess.make_command_text label "dl" in
@@ -277,18 +347,46 @@ let pull_tree
           Some ("missing checksum"),
           label ^ ": Missing checksum, and `--require-checksums` was set."))
     else
-      pull_from_mirrors label ?working_dir ?subpath cache_dir local_dirname checksums
-        remote_urls
+      OpamFilename.with_tmp_dir_job @@ fun tmpdir ->
+      let extract url archive =
+        match dirnames with
+        | [_] ->
+          let tmp_archive = OpamFilename.(create tmpdir (basename archive)) in
+          OpamFilename.move ~src:archive ~dst:tmp_archive;
+          extract_archive tmp_archive url
+        | _ -> extract_archive archive url
+      in
+      let pull label checksums remote_urls =
+        match dirnames with
+        | [ label, local_dirname, subpath ] ->
+          pull_from_mirrors label ?full_fetch ?working_dir ?subpath
+            cache_dir local_dirname checksums remote_urls
+          @@| fun (url, res) ->
+          (OpamUrl.to_string_w_subpath subpath url),
+          res
+        | _ ->
+          pull_from_mirrors label ?full_fetch ?working_dir cache_dir tmpdir
+            checksums remote_urls
+          @@| fun (url, res) -> OpamUrl.to_string url, res
+      in
+      pull label checksums remote_urls
       @@+ function
       | _, Up_to_date None -> Done (Up_to_date "no changes")
       | url, (Up_to_date (Some archive) | Result (Some archive)) ->
-        OpamFilename.with_tmp_dir_job @@ fun tmpdir ->
-        let tmp_archive = OpamFilename.(create tmpdir (basename archive)) in
-        OpamFilename.move ~src:archive ~dst:tmp_archive;
-        extract_archive tmp_archive (OpamUrl.to_string url)
-      | url, Result None -> Done (Result (OpamUrl.to_string url))
+        extract url archive
+      | url, Result None -> Done (Result url)
       | _, (Not_available _ as na) -> Done na
 
+
+let pull_tree label
+    ?full_fetch ?cache_dir ?(cache_urls=[]) ?working_dir ?subpath
+    local_dirname  =
+  pull_tree_t ?full_fetch ?cache_dir ~cache_urls ?working_dir
+    [label, local_dirname, subpath]
+
+let pull_shared_tree ?cache_dir ?(cache_urls=[]) dirnames checksums remote_urls =
+  pull_tree_t ?cache_dir ~cache_urls dirnames checksums remote_urls
+
 let revision dirname url =
   let kind = url.OpamUrl.backend in
   let module B = (val find_backend_by_kind kind: OpamRepositoryBackend.S) in
@@ -456,7 +554,8 @@ let update repo repo_root =
   | Update_empty ->
     log "update empty, no validation performed";
     apply_repo_update repo repo_root Update_empty @@+ fun () ->
-    B.repo_update_complete repo_root repo.repo_url
+    B.repo_update_complete repo_root repo.repo_url @@+ fun () ->
+    Done `No_changes
   | (Update_full _ | Update_patch _) as upd ->
     OpamProcess.Job.catch (fun exn ->
         cleanup_repo_update upd;
@@ -468,7 +567,8 @@ let update repo repo_root =
       failwith "Invalid repository signatures, update aborted"
     | true ->
       apply_repo_update repo repo_root upd @@+ fun () ->
-      B.repo_update_complete repo_root repo.repo_url
+      B.repo_update_complete repo_root repo.repo_url @@+ fun () ->
+      Done `Changes
 
 let on_local_version_control url ~default f =
   match url.OpamUrl.backend with
