@@ -31,8 +31,8 @@ module type S = sig
 
   val name: OpamUrl.backend
 
-  (** [pull_url local_dir checksum remote_url] pulls the contents of
-      [remote_url] into [local_dir].
+  (** [pull_url ?full_fetch ?cache_dir ?subpath local_dir checksum remote_url]
+      pulls the contents of [remote_url] into [local_dir].
 
       Two kinds of results are allowed:
 
@@ -43,10 +43,17 @@ module type S = sig
         been synchronised with its own, and [None] is returned
 
       [checksum] can be used for retrieval but is NOT checked by this
-      function. *)
+      function.
+
+      If [full_fetch] is set to false, VCS repository is retrieved with shallow
+      history (by default, full history).
+      If [cache_dir] is given, the directory is used by VCS tool as a its cache
+      directory.
+      If [subpath] is given, only that [subpath] of the url is retrieved. *)
   val pull_url:
-    ?cache_dir:dirname -> ?subpath:string -> dirname -> OpamHash.t option -> url ->
-    filename option download OpamProcess.job
+    ?full_fetch:bool ->
+    ?cache_dir:dirname -> ?subpath:subpath -> dirname -> OpamHash.t option ->
+    url -> filename option download OpamProcess.job
 
   (** [pull_repo_update] fetches the remote update from [url] to the local
       repository at [dirname], but does not apply it, allowing for further
@@ -73,7 +80,8 @@ module type S = sig
       [pull_url], then remove deleted files, and finally copy via rsync
       unversioned & modified-uncommitted files. *)
   val sync_dirty:
-    ?subpath:string -> dirname -> url -> filename option download OpamProcess.job
+    ?subpath:subpath -> dirname -> url ->
+    filename option download OpamProcess.job
 
   (** [get_remote_url ?hash dirname] return the distant url of repo [dirname], \
       if found. When [hash] is specified, it checks that this hash (branch or \
