@@ -14,26 +14,21 @@ open OpamTypes
 open OpamTypesBase
 
 let log ?level fmt = OpamConsole.log "opam-file" ?level fmt
+let slog = OpamConsole.slog
 
 open OpamFile.OPAM
 
 let is_valid_license_id s =
-  let memplus s =
-    let s = OpamStd.String.remove_suffix ~suffix:"+" s in
-    OpamStd.String.Set.mem s OpamSpdxList.licenses
-  in
-  match OpamStd.String.split (String.lowercase_ascii s) ' ' with
-  | [s] -> memplus s
-  | [s; "with"; e] ->
-    memplus s && OpamStd.String.Set.mem e OpamSpdxList.exceptions
-  | _ -> false
+  match Spdx_licenses.parse s with
+  | Ok _ -> true
+  | Error _ -> false
 
 (** manipulation utilities *)
 
 let names_of_formula flag f =
   OpamPackageVar.filter_depends_formula
-    ~build:true ~post:true ~dev:true ~test:flag ~doc:flag ~default:false
-    ~env:OpamStd.Option.none f
+    ~build:true ~post:true ~dev:true ~test:flag ~doc:flag ~dev_setup:flag
+    ~default:false ~env:OpamStd.Option.none f
   |> OpamFormula.atoms
   |> List.map fst
   |> OpamPackage.Name.Set.of_list
@@ -124,6 +119,91 @@ let map_all_filters f t =
   with_deprecated_build_test (map_commands t.deprecated_build_test) |>
   with_deprecated_build_doc (map_commands t.deprecated_build_doc)
 
+(* unguarded_commands_variables is an alternative implementation of
+   OpamFilter.commands_variables which excludes package variables which are
+   guarded by an unambiguous {package:installed} filter. That is, at each level,
+   if assuming !package:installed reduces the filter to false, then the uses of
+   package:variable are not returned. This allows expressions like:
+   ["--with-foo=%{foo:share}%" {foo:installed}] or even
+   ["--with-foo"] {foo:installed & foo:bar != "baz"} not to trigger warning 41
+   if the package is not explicitly depended on. *)
+
+let unguarded_commands_variables commands =
+  let is_installed_variable filter guarded_packages v =
+    match OpamVariable.Full.package v with
+    | None -> guarded_packages
+    | (Some name) as package ->
+      let is_installed var =
+        String.equal "installed"
+          (OpamVariable.to_string (OpamVariable.Full.variable var))
+      in
+      let env var =
+        if Option.equal OpamPackage.Name.equal
+            (OpamVariable.Full.package var) package &&
+           is_installed var then
+          Some (B false)
+        else
+          None
+      in
+      if is_installed v &&
+         OpamFilter.partial_eval env filter = FBool false then
+        OpamPackage.Name.Set.add name guarded_packages
+      else
+        guarded_packages
+  in
+  let filter_guarded variables guarded_packages =
+    let is_unguarded v =
+      match OpamVariable.Full.package v with
+      | Some package ->
+        not (OpamPackage.Name.Set.mem package guarded_packages)
+      | None -> true
+    in
+    List.filter is_unguarded variables
+  in
+  let unguarded_packages_from_filter guarded_packages = function
+    | None -> guarded_packages, []
+    | Some f ->
+      let filter_variables = OpamFilter.variables f in
+      let guarded_packages =
+        List.fold_left (is_installed_variable f)
+          guarded_packages filter_variables
+      in
+      guarded_packages, filter_guarded filter_variables guarded_packages
+  in
+  let unguarded_argument_variables guarded_packages (argument, filter) =
+    let guarded_packages, filter_variables =
+      unguarded_packages_from_filter guarded_packages filter
+    in
+    let variables_from_arguments =
+      filter_guarded (OpamFilter.simple_arg_variables argument) guarded_packages
+    in
+    guarded_packages, variables_from_arguments @ filter_variables
+  in
+  let unguarded_command_variables guarded_packages (command, filter) =
+    let filter_guarded_packages, filter_variables =
+      unguarded_packages_from_filter OpamPackage.Name.Set.empty filter
+    in
+    let add_argument (guarded_packages, acc) argument =
+      let guarded_packages, unguarded_variables =
+        unguarded_argument_variables guarded_packages argument
+      in
+      guarded_packages, unguarded_variables @ acc
+    in
+    let command_guarded_packages, unguarded_variables =
+      List.fold_left add_argument (filter_guarded_packages, filter_variables)
+        command
+    in
+    OpamPackage.Name.Set.union guarded_packages command_guarded_packages,
+    unguarded_variables
+  in
+  let f (guarded_packages, acc) c =
+    let guarded_packages, unguarded_variables =
+      unguarded_command_variables guarded_packages c
+    in
+    guarded_packages, (unguarded_variables @ acc)
+  in
+  List.fold_left f (OpamPackage.Name.Set.empty, []) commands
+
 (* Returns all variables from all commands (or on given [command]) and all filters *)
 let all_variables ?exclude_post ?command t =
   let commands =
@@ -135,6 +215,18 @@ let all_variables ?exclude_post ?command t =
   List.fold_left (fun acc f -> OpamFilter.variables f @ acc)
     [] (all_filters ?exclude_post t)
 
+(* As all_variables, but any commands or arguments which are fully guarded by
+   package:installed are excluded; used for Warning 41 so that
+   ["%{foo:share}%" {foo:installed}] doesn't trigger a warning on foo *)
+let all_unguarded_variables ?exclude_post t =
+  let guarded_packages, unguarded_commands_variables =
+    unguarded_commands_variables (all_commands t)
+  in
+  guarded_packages,
+  unguarded_commands_variables @
+  List.fold_left (fun acc f -> OpamFilter.variables f @ acc)
+    [] (all_filters ?exclude_post t)
+
 let map_all_variables f t =
   let map_fld (x, flt) = x, OpamFilter.map_variables f flt in
   let map_optfld = function
@@ -297,17 +389,17 @@ let t_lint ?check_extra_files ?(check_upstream=false) ?(all=false) t =
   let all_expanded_strings = all_expanded_strings t in
   let all_depends = all_depends t in
   (* Upstream is checked only if it is an archive and non vcs backend *)
-  let url_is_archive =
+  let url_vcs =
     let open OpamStd.Option.Op in
     t.url >>| OpamFile.URL.url >>| (fun u ->
         match u.OpamUrl.backend with
-        | #OpamUrl.version_control -> false
-        | _ -> OpamSystem.is_archive (OpamUrl.base_url u))
+        | #OpamUrl.version_control -> true
+        | _ -> false)
   in
   let check_upstream =
     check_upstream &&
     not (OpamFile.OPAM.has_flag Pkgflag_Conf t) &&
-    url_is_archive = Some true
+    url_vcs = Some false
   in
   let warnings = [
     cond 20 `Warning
@@ -369,12 +461,13 @@ let t_lint ?check_extra_files ?(check_upstream=false) ?(all=false) t =
        (unk_flags <> []));
     (let filtered_vars =
        OpamFilter.variables_of_filtered_formula t.depends @
-       OpamFilter.variables_of_filtered_formula t.depopts
+       OpamFilter.variables_of_filtered_formula t.depopts @
+       OpamFilter.variables_of_filtered_formula t.conflicts
        |> List.filter (fun v -> not (OpamVariable.Full.is_global v))
        |> List.map OpamVariable.Full.to_string
      in
      cond 29 `Error
-       "Package dependencies mention package variables"
+       "Package dependencies or conflicts mention package variables"
        ~detail:filtered_vars
        (filtered_vars <> []));
 (*
@@ -460,18 +553,32 @@ let t_lint ?check_extra_files ?(check_upstream=false) ?(all=false) t =
        ~detail:alpha_flags
        (alpha_flags <> []));
 *)
-    (let undep_pkgs =
-       List.fold_left
-         (fun acc v ->
-            match OpamVariable.Full.package v with
-            | Some n when
-                t.OpamFile.OPAM.name <> Some n &&
-                not (OpamPackage.Name.Set.mem n all_depends) &&
-                OpamVariable.(Full.variable v <> of_string "installed")
-              ->
-              OpamPackage.Name.Set.add n acc
-            | _ -> acc)
-         OpamPackage.Name.Set.empty (all_variables ~exclude_post:true t)
+    (let all_mentioned_packages =
+       OpamPackage.Name.Set.union
+         (OpamFormula.all_names t.depends)
+         (OpamFormula.all_names t.depopts)
+     in
+     let undep_pkgs =
+       let guarded_packages, all_unguarded_variables =
+         all_unguarded_variables ~exclude_post:true t
+       in
+       let first_lot =
+         List.fold_left
+           (fun acc v ->
+              match OpamVariable.Full.package v with
+              | Some n when
+                  t.OpamFile.OPAM.name <> Some n &&
+                  not (OpamPackage.Name.Set.mem n all_depends) &&
+                  OpamVariable.(Full.variable v <> of_string "installed")
+                ->
+                OpamPackage.Name.Set.add n acc
+              | _ -> acc)
+           OpamPackage.Name.Set.empty all_unguarded_variables
+       in
+       let second_lot =
+         OpamPackage.Name.Set.diff guarded_packages all_mentioned_packages
+       in
+       OpamPackage.Name.Set.union first_lot second_lot
      in
      cond 41 `Warning
        "Some packages are mentioned in package scripts or features, but \
@@ -530,14 +637,14 @@ let t_lint ?check_extra_files ?(check_upstream=false) ?(all=false) t =
        (t.install <> [] || t.remove <> [] || t.url <> None ||
         t.extra_sources <> []));
     cond 47 `Warning
-      "Synopsis (or description first line) should start with a capital and \
-       not end with a dot"
+      "Synopsis should start with a capital and not end with a dot"
       (let valid_re =
          Re.(compile (seq [bos; diff any (alt [blank; lower]); rep any;
                            diff any (alt [blank; char '.']); eos]))
        in
        match t.descr with None -> false | Some d ->
-         not (Re.execp valid_re (OpamFile.Descr.synopsis d)));
+         let synopsis = OpamFile.Descr.synopsis d in
+         synopsis <> "" && not (Re.execp valid_re synopsis));
     cond 48 `Warning
       "The fields 'build-test:' and 'build-doc:' are deprecated, and should be \
        replaced by uses of the 'with-test' and 'with-doc' filter variables in \
@@ -603,10 +710,15 @@ let t_lint ?check_extra_files ?(check_upstream=false) ?(all=false) t =
        | Some fs, Some [] -> List.map fst fs
        | Some efiles, Some ffiles ->
          OpamStd.List.filter_map (fun (n, _) ->
-             if List.mem_assoc n ffiles then None else Some n)
+             if OpamStd.List.mem_assoc OpamFilename.Base.equal
+                 n ffiles then
+               None else Some n)
            efiles @
          OpamStd.List.filter_map (fun (n, check_f) ->
-             try if check_f (List.assoc n efiles) then None else Some n
+             try
+               if check_f (OpamStd.List.assoc OpamFilename.Base.equal
+                             n efiles) then
+                 None else Some n
              with Not_found -> Some n)
            ffiles
      in
@@ -651,12 +763,16 @@ let t_lint ?check_extra_files ?(check_upstream=false) ?(all=false) t =
                       used norm)
                   bad_os_arch_values)
        (bad_os_arch_values <> []));
+    (* Retired, since `OPAM_LAST_ENV` allows environment updates to be reliably
+       reverted. *)
+(*
     cond 56 `Warning
       "It is discouraged for non-compiler packages to use 'setenv:'"
       (t.env <> [] && not (has_flag Pkgflag_Compiler t));
+*)
     cond 57 `Error
-      "Synopsis and description must not be both empty"
-      (t.descr = None || t.descr = Some OpamFile.Descr.empty);
+      "Synopsis must not be empty"
+      (match t.descr with None -> true | Some d -> String.equal (OpamFile.Descr.synopsis d) "");
     (let vars = all_variables ~exclude_post:false ~command:[] t in
      let exists svar =
        List.exists (fun v -> v = OpamVariable.Full.of_string svar) vars
@@ -749,7 +865,7 @@ let t_lint ?check_extra_files ?(check_upstream=false) ?(all=false) t =
        List.filter (fun s -> not (is_valid_license_id s)) t.license
      in
      cond 62 `Warning
-       "License doesn't adhere to the SPDX standard, see https://spdx.org/licenses/"
+       "License doesn't adhere to the SPDX standard, see https://spdx.org/licenses/ "
        ~detail:bad_licenses
        (bad_licenses <> []));
 (*
@@ -864,11 +980,37 @@ let t_lint ?check_extra_files ?(check_upstream=false) ?(all=false) t =
               |> List.map OpamHash.to_string
               |> OpamStd.Format.pretty_list)])
           t.url)
-      (match t.url with
-       | None -> false
-       | Some urlf ->
-         (OpamFile.URL.checksum urlf <> [])
-         && url_is_archive <> Some true);
+      (url_vcs = Some true
+       && OpamStd.Option.Op.(t.url >>| fun u -> OpamFile.URL.checksum u <> [])
+          = Some true);
+    cond 68 `Warning
+      "Missing field 'license'"
+      (t.license = []);
+    (let vars =
+       List.flatten @@
+       List.map (fun s ->
+           List.filter_map
+             (fun f ->
+                try
+                  let _ =
+                    OpamTypesBase.filter_ident_of_string_interp
+                      ~accept:false f
+                  in
+                  None
+                with OpamTypesBase.Parse_variable (p,v) ->
+                  Some (p,v)
+             )
+             (OpamFilter.extract_variables_from_string s))
+         all_expanded_strings
+     in
+     cond 69 `Warning
+       "Package name in variable in string interpolation contains several \
+        '+', use"
+       ~detail:(List.map (fun (p,v) ->
+           Printf.sprintf "'?%s:%s:' instead of '%s:%s'"
+             p v p v)
+           vars)
+       (vars <> []));
   ]
   in
   format_errors @
@@ -1118,9 +1260,12 @@ let add_aux_files ?dir ~files_subdir_hashes opam =
       match OpamFile.OPAM.extra_files opam, extra_files with
       | None, None -> opam
       | None, Some ef ->
-        log ~level:2 "Missing extra-files field for %s, adding them."
-          (OpamStd.List.concat_map ", "
-             (fun (_,f) -> OpamFilename.Base.to_string f) ef);
+        log ~level:2
+          "Missing extra-files field for %a for %a, adding them."
+          (slog @@ OpamStd.List.concat_map ", "
+             (fun (_,f) -> OpamFilename.Base.to_string f)) ef
+          OpamStd.Op.(slog @@ OpamPackage.to_string @* OpamFile.OPAM.package)
+          opam;
         let ef =
           List.map
             (fun (file, basename) ->
@@ -1138,7 +1283,8 @@ let add_aux_files ?dir ~files_subdir_hashes opam =
       | Some oef, Some ef ->
         let wr_check, nf_opam, rest =
           List.fold_left (fun (wr_check, nf_opam, rest) (file, basename) ->
-              match OpamStd.List.pick_assoc basename rest with
+              match OpamStd.List.pick_assoc
+                      OpamFilename.Base.equal basename rest with
               | None, rest ->
                 wr_check, (basename::nf_opam), rest
               | Some ohash, rest ->
