@@ -57,14 +57,13 @@ val is_dev_package: 'a switch_state -> OpamFile.OPAM.t -> bool
 (** The defaults are [true] for [build], false for [dev] and [post], and
     defined by OpamStateConfig for [test] and [bool]. *)
 val filter_depends_formula:
-  ?build:bool -> ?post:bool -> ?test:bool -> ?doc:bool -> ?dev:bool ->
-  ?default:bool -> env:OpamFilter.env ->
+  ?build:bool -> ?post:bool -> ?test:bool -> ?doc:bool -> ?dev_setup:bool ->
+  ?dev:bool -> ?default:bool -> env:OpamFilter.env ->
   filtered_formula -> formula
 
 (** Assumes [filter_default=false] by default, i.e. dependencies with undefined
     filters are discarded. *)
 val all_depends:
-  ?build:bool -> ?post:bool -> ?test:bool -> ?doc:bool -> ?dev:bool ->
-  ?filter_default:bool ->
-  ?depopts:bool ->
+  ?build:bool -> ?post:bool -> ?test:bool -> ?doc:bool -> ?dev_setup:bool ->
+  ?dev:bool -> ?filter_default:bool -> ?depopts:bool ->
   'a switch_state -> OpamFile.OPAM.t -> formula
