@@ -29,6 +29,7 @@ module E : sig
     | ROOT of string option
     | SWITCH of string option
     | UNLOCKBASE of bool option
+    | WITHDEVSETUP of bool option
     | WITHDOC of bool option
     | WITHTEST of bool option
   val root: unit -> string option
@@ -37,12 +38,15 @@ end
 
 type t = private {
   root_dir: OpamFilename.Dir.t;
+  original_root_dir: OpamFilename.Dir.t;
+  root_from: provenance;
   current_switch: OpamSwitch.t option;
   switch_from: provenance;
   jobs: int Lazy.t;
   dl_jobs: int;
   build_test: bool;
   build_doc: bool;
+  dev_setup: bool;
   dryrun: bool;
   makecmd: string Lazy.t;
   ignore_constraints_on: name_set;
@@ -54,12 +58,15 @@ type t = private {
 
 type 'a options_fun =
   ?root_dir:OpamFilename.Dir.t ->
+  ?original_root_dir:OpamFilename.Dir.t ->
+  ?root_from:provenance ->
   ?current_switch:OpamSwitch.t ->
   ?switch_from:provenance ->
   ?jobs:(int Lazy.t) ->
   ?dl_jobs:int ->
   ?build_test:bool ->
   ?build_doc:bool ->
+  ?dev_setup:bool ->
   ?dryrun:bool ->
   ?makecmd:string Lazy.t ->
   ?ignore_constraints_on:name_set ->
@@ -73,9 +80,15 @@ include OpamStd.Config.Sig
   with type t := t
    and type 'a options_fun := 'a options_fun
 
+(** Default state values *)
+val default : t
+
 (** Get the initial opam root value (from default, env or optional argument).
     This allows one to get it before doing the init, which is useful to get the
     configuration file used to fill some options to init() *)
+val opamroot_with_provenance: ?root_dir:dirname -> unit -> provenance * dirname
+
+(** [opamroot ?root_dir () = snd (opamroot_with_provenance ?root_dir ()] *)
 val opamroot: ?root_dir:dirname -> unit -> dirname
 
 (** Loads the global configuration file, protecting against concurrent writes *)
