@@ -13,7 +13,7 @@ open OpamTypes
 open OpamStd.Op
 open OpamPackage.Set.Op
 
-let log fmt = OpamConsole.log "STATE" fmt
+let log ?level fmt = OpamConsole.log ?level "STATE" fmt
 let slog = OpamConsole.slog
 
 open OpamStateTypes
@@ -34,24 +34,28 @@ let load_switch_config ?lock_kind gt switch =
        (OpamSwitch.to_string switch);
      OpamFile.Switch_config.empty)
 
-let compute_available_packages gt switch switch_config ~pinned ~opams =
-  (* remove all versions of pinned packages, but the pinned-to version *)
-  let pinned_names = OpamPackage.names_of_packages pinned in
-  let opams =
-    OpamPackage.Map.filter
-      (fun nv _ ->
-         not (OpamPackage.Name.Set.mem nv.name pinned_names) ||
-         OpamPackage.Set.mem nv pinned)
-      opams
-  in
-  let avail_map =
+let filter_available_packages gt switch switch_config ~opams =
+  OpamPackage.keys @@
     OpamPackage.Map.filter (fun package opam ->
         OpamFilter.eval_to_bool ~default:false
           (OpamPackageVar.resolve_switch_raw ~package gt switch switch_config)
           (OpamFile.OPAM.available opam))
       opams
+
+let compute_available_and_pinned_packages gt switch switch_config ~pinned ~opams =
+  (* remove all versions of pinned packages, but the pinned-to version *)
+  let pinned_names = OpamPackage.names_of_packages pinned in
+  let (opams, pinned) =
+    OpamPackage.Map.partition
+      (fun nv _ ->
+         not (OpamPackage.Name.Set.mem nv.name pinned_names) ||
+         OpamPackage.Set.mem nv pinned)
+      opams
   in
-  OpamPackage.keys avail_map
+  (filter_available_packages gt switch switch_config ~opams, pinned)
+
+let compute_available_packages gt switch switch_config ~pinned ~opams =
+  fst @@ compute_available_and_pinned_packages gt switch switch_config ~pinned ~opams
 
 let repos_list_raw rt switch_config =
   let global, repos =
@@ -87,7 +91,7 @@ let infer_switch_invariant_raw
   let env nv v =
     if List.mem v OpamPackageVar.predefined_depends_variables then
       match OpamVariable.Full.to_string v with
-      | "dev" | "with-test" | "with-doc" -> Some (B false)
+      | "dev" | "with-test" | "with-doc" | "with-dev-setup" -> Some (B false)
       | _ -> None
     else
       OpamPackageVar.resolve_switch_raw ~package:nv gt switch switch_config v
@@ -115,7 +119,7 @@ let infer_switch_invariant_raw
             deps dmap
         in
         dmap)
-      (OpamPackage.packages_of_names (Lazy.force available_packages) @@
+      (OpamPackage.packages_of_names available_packages @@
        OpamPackage.names_of_packages @@
        compiler)
       OpamPackage.Map.empty
@@ -133,7 +137,7 @@ let infer_switch_invariant_raw
   match List.sort (fun (_, c1) (_, c2) -> compare c1 c2) counts with
   | (nv, _) :: _ ->
     let versions =
-      OpamPackage.packages_of_name (Lazy.force available_packages) nv.name
+      OpamPackage.packages_of_name available_packages nv.name
     in
     let n = OpamPackage.Set.cardinal versions in
     if n <= 1 then
@@ -153,9 +157,10 @@ let infer_switch_invariant st =
         st.installed
     else st.compiler_packages
   in
+  let lazy available_packages = st.available_packages in
   infer_switch_invariant_raw
     st.switch_global st.switch st.switch_config st.opams
-    st.packages compiler_packages st.installed_roots st.available_packages
+    st.packages compiler_packages st.installed_roots available_packages
 
 let depexts_raw ~env nv opams =
   try
@@ -173,7 +178,9 @@ module Installed_cache = OpamCached.Make(struct
     let name = "installed"
   end)
 
-let depexts_status_of_packages_raw ~depexts global_config switch_config packages =
+let depexts_status_of_packages_raw
+    ~depexts ?env global_config switch_config packages =
+  if OpamPackage.Set.is_empty packages then OpamPackage.Map.empty else
   let open OpamSysPkg.Set.Op in
   let syspkg_set, syspkg_map =
     OpamPackage.Set.fold (fun nv (set, map) ->
@@ -190,7 +197,7 @@ let depexts_status_of_packages_raw ~depexts global_config switch_config packages
   in
   let syspkg_set = syspkg_set -- bypass in
   let ret =
-    match OpamSysInteract.packages_status syspkg_set with
+    match OpamSysInteract.packages_status ?env global_config syspkg_set with
     | avail, not_found ->
       let avail, not_found =
         if OpamStateConfig.(!r.no_depexts) then
@@ -223,6 +230,7 @@ let depexts_unavailable_raw sys_packages nv =
   | _ -> None
 
 let load lock_kind gt rt switch =
+  OpamFormatUpgrade.as_necessary_repo_switch_light_upgrade lock_kind `Switch gt;
   let chrono = OpamConsole.timer () in
   log "LOAD-SWITCH-STATE %@ %a" (slog OpamSwitch.to_string) switch;
   if not (OpamGlobalState.switch_exists gt switch) then
@@ -330,7 +338,7 @@ let load lock_kind gt rt switch =
     OpamPackage.Map.union (fun _ x -> x) repos_package_index pinned_opams
   in
   let available_packages =
-    lazy (compute_available_packages gt switch switch_config
+    lazy (compute_available_and_pinned_packages gt switch switch_config
             ~pinned ~opams)
   in
   let opams =
@@ -366,7 +374,7 @@ let load lock_kind gt rt switch =
     let changed =
       OpamPackage.Map.merge (fun _ opam_new opam_installed ->
           match opam_new, opam_installed with
-          | Some r, Some i when not (OpamFile.OPAM.effectively_equal i r) ->
+          | Some r, Some i when not (OpamFile.OPAM.effectively_equal ~modulo_state:true i r) ->
             Some ()
           | _ -> None)
         opams installed_opams
@@ -376,32 +384,15 @@ let load lock_kind gt rt switch =
       (slog OpamPackage.Set.to_string) changed;
     changed
   ) in
-  (* Detect and initialise missing switch description *)
-  let switch_config =
-    if switch_config <> OpamFile.Switch_config.empty &&
-       switch_config.synopsis = "" then
-      let synopsis =
-        match OpamPackage.Set.elements (compiler_packages %% installed_roots)
-        with
-        | [] -> OpamSwitch.to_string switch
-        | [nv] ->
-          let open OpamStd.Option.Op in
-          (OpamPackage.Map.find_opt nv opams >>= OpamFile.OPAM.synopsis) +!
-          OpamPackage.to_string nv
-        | pkgs -> OpamStd.List.concat_map " " OpamPackage.to_string pkgs
-      in
-      let conf = { switch_config with synopsis } in
-      if lock_kind = `Lock_write then (* auto-repair *)
-        OpamFile.Switch_config.write
-          (OpamPath.Switch.switch_config gt.root switch)
-          conf;
-      conf
-    else switch_config
-  in
   let switch_config, switch_invariant =
     match switch_config.invariant with
     | Some invariant -> switch_config, invariant
     | None ->
+      let available_packages =
+        let lazy (available_packages, pinned) = available_packages in
+        OpamPackage.Set.union available_packages @@
+          filter_available_packages gt switch switch_config ~opams:pinned
+      in
       let invariant =
         infer_switch_invariant_raw
           gt switch switch_config opams
@@ -427,6 +418,24 @@ let load lock_kind gt rt switch =
           switch_config;
       switch_config, invariant
   in
+  (* Detect and initialise missing switch description *)
+  let switch_config =
+    if switch_config <> OpamFile.Switch_config.empty &&
+       switch_config.synopsis = "" then
+      let synopsis =
+        if switch_invariant = OpamFormula.Empty then
+          OpamSwitch.to_string switch
+        else
+          OpamFormula.to_string switch_invariant
+      in
+      let conf = { switch_config with synopsis } in
+      if lock_kind = `Lock_write then (* auto-repair *)
+        OpamFile.Switch_config.write
+          (OpamPath.Switch.switch_config gt.root switch)
+          conf;
+      conf
+    else switch_config
+  in
   let conf_files =
     let conf_files =
       OpamFilename.files (OpamPath.Switch.config_dir gt.root switch)
@@ -451,15 +460,14 @@ let load lock_kind gt rt switch =
   let ext_files_changed = lazy (
     OpamPackage.Name.Map.fold (fun name conf acc ->
         let nv = OpamPackage.package_of_name installed name in
+        let path = lazy (
+          OpamStd.Sys.split_path_variable (OpamStd.Env.get "PATH")
+          |> List.map OpamFilename.Dir.of_string
+        ) in
         if
           List.exists (fun (file, hash) ->
               let exists = OpamFilename.exists file in
-              let should_exist =
-                let count_not_zero c =
-                  function '0' -> c | _ -> succ c
-                in
-                OpamStd.String.fold_left count_not_zero 0 (OpamHash.contents hash) <> 0
-              in
+              let should_exist = OpamHash.is_null hash in
               let changed =
                 exists <> should_exist ||
                 exists && not (OpamHash.check_file (OpamFilename.to_string file) hash)
@@ -477,7 +485,21 @@ let load lock_kind gt rt switch =
                   "File %s, which package %s depends upon, was changed on your \
                    system.\n\
                    The package will need to be reinstalled."
-                  (OpamFilename.to_string file) (OpamPackage.to_string nv);
+                  (OpamFilename.to_string file) (OpamPackage.to_string nv)
+              else
+                (let exec_not_in_path =
+                   should_exist &&
+                   OpamFilename.is_exec file &&
+                   let dirname = OpamFilename.dirname file in
+                   not @@ List.exists (OpamFilename.Dir.equal dirname)
+                     (Lazy.force path)
+                 in
+                 if exec_not_in_path then
+                   OpamConsole.warning
+                     "File %s, which package %s depends upon, \
+                      is no longer in your %s."
+                     (OpamFilename.to_string file) (OpamPackage.to_string nv)
+                     (OpamConsole.colorise `bold "PATH"));
               changed)
             (OpamFile.Dot_config.file_depends conf)
         then OpamPackage.Set.add nv acc
@@ -486,12 +508,14 @@ let load lock_kind gt rt switch =
       OpamPackage.Set.empty
   ) in
   (* depext check *)
+  let available_packages = OpamCompat.Lazy.map fst available_packages in
   let sys_packages =
     if not (OpamFile.Config.depext gt.config)
     || OpamStateConfig.(!r.no_depexts) then
       lazy OpamPackage.Map.empty
     else lazy (
       depexts_status_of_packages_raw gt.config switch_config
+        ~env:gt.global_variables
         (Lazy.force available_packages)
         ~depexts:(fun package ->
             let env =
@@ -561,6 +585,12 @@ let load lock_kind gt rt switch =
                (OpamPackage.Map.bindings missing_map)));
     changed
   ) in
+  let available_packages = lazy (
+    let chrono = OpamConsole.timer () in
+    let r = Lazy.force available_packages in
+    log ~level:2 "Availability of packages computed in %.3fs." (chrono ());
+    r
+  ) in
   let reinstall = lazy (
     OpamFile.PkgList.safe_read (OpamPath.Switch.reinstall gt.root switch) ++
     Lazy.force changed ++
@@ -568,7 +598,9 @@ let load lock_kind gt rt switch =
     Lazy.force sys_packages_changed
   ) in
   let invalidated = lazy (
-    Lazy.force ext_files_changed ++ Lazy.force sys_packages_changed
+    Lazy.force changed ++
+    Lazy.force ext_files_changed ++
+    Lazy.force sys_packages_changed
     -- Lazy.force available_packages
   ) in
   let st = {
@@ -675,7 +707,8 @@ let primary_url_with_subpath st nv =
     let url = OpamFile.URL.url urlf in
     match OpamFile.URL.subpath urlf with
     | None -> Some url
-    | Some subpath -> Some (OpamUrl.Op.(url / subpath))
+    | Some subpath ->
+      Some OpamUrl.Op.(url / (OpamFilename.SubPath.to_string subpath))
 
 let files st nv =
   match opam_opt st nv with
@@ -709,7 +742,9 @@ let is_dev_package st nv =
     if OpamPackage.Set.mem nv st.pinned then
       opam_opt st nv
     else
-      OpamPackage.Map.find_opt nv st.repos_package_index
+      match OpamPackage.Map.find_opt nv st.repos_package_index with
+      | None -> opam_opt st nv
+      | some -> some
   in
   match opam_opt with
   | Some opam -> OpamPackageVar.is_dev_package st opam
@@ -738,7 +773,7 @@ let depexts st nv =
 
 let depexts_status_of_packages st set =
   depexts_status_of_packages_raw st.switch_global.config st.switch_config set
-    ~depexts:(depexts st)
+    ~env:st.switch_global.global_variables ~depexts:(depexts st)
 
 let depexts_unavailable st nv =
   depexts_unavailable_raw (Lazy.force st.sys_packages) nv
@@ -836,48 +871,102 @@ let get_conflicts st packages opams_map =
     (fun package -> OpamPackageVar.resolve_switch ~package st)
     packages opams_map
 
-let can_upgrade_to_avoid_version name st =
-  OpamPackage.Set.exists (fun pkg ->
-      OpamPackage.Name.equal (OpamPackage.name pkg) name &&
-      OpamFile.OPAM.has_flag Pkgflag_AvoidVersion (OpamPackage.Map.find pkg st.opams)
-    ) st.installed
+let avoid_version st nv =
+  let open OpamStd.Option.Op in
+  let opam = opam st nv in
+  let has_avoid_flag opam =
+    OpamFile.OPAM.has_flag Pkgflag_AvoidVersion opam
+    || OpamFile.OPAM.has_flag Pkgflag_Deprecated opam
+  in
+  has_avoid_flag opam
+  && not ((OpamPackage.package_of_name_opt st.installed nv.name >>=
+           (fun nv -> OpamPackage.Map.find_opt nv st.installed_opams) >>|
+           has_avoid_flag)
+          +! false)
+
+let undefined_filter_variable nv v =
+  (if OpamFormatConfig.(!r.strict) then
+     OpamConsole.error_and_exit `File_error
+       "Undefined filter variable %s in dependencies of %s"
+   else
+     log
+       "ERR: Undefined filter variable %s in dependencies of %s")
+    (OpamVariable.Full.to_string v) (OpamPackage.to_string nv)
+
+let package_env_t st ~force_dev_deps ~test ~doc ~dev_setup
+    ~requested_allpkgs ?(err_undefined=true) nv v =
+  if List.mem v OpamPackageVar.predefined_depends_variables then
+    match OpamVariable.Full.to_string v with
+    | "dev" ->
+      Some (B (force_dev_deps || is_dev_package st nv))
+    | "with-test" ->
+      Some (B (test && OpamPackage.Set.mem nv requested_allpkgs))
+    | "with-doc" ->
+      Some (B (doc && OpamPackage.Set.mem nv requested_allpkgs))
+    | "with-dev-setup" ->
+      Some (B (dev_setup && OpamPackage.Set.mem nv requested_allpkgs))
+    | _ -> None (* Computation delayed to the solver *)
+  else
+  let r = OpamPackageVar.resolve_switch ~package:nv st v in
+  if err_undefined && r = None then
+    undefined_filter_variable nv v;
+  r
+
+let get_dependencies_t st ~force_dev_deps ~test ~doc ~dev_setup
+    ~requested_allpkgs deps opams =
+  let filter_undefined nv =
+    let warn_undefined v =
+      if not (List.mem v OpamPackageVar.predefined_depends_variables) then
+        undefined_filter_variable nv v
+    in
+    OpamFormula.map (fun (name, fc) ->
+        let fc =
+          OpamFormula.map (function
+              | Constraint (_, FIdent (_, v, _))
+              | Constraint (_, FUndef (FIdent (_, v, _))) ->
+                (if OpamFormatConfig.(!r.strict) then
+                   OpamConsole.error_and_exit `File_error
+                 else OpamConsole.warning)
+                  "Undefined filter variable %s in dependencies of %s"
+                  (OpamVariable.to_string v) (OpamPackage.to_string nv);
+                Atom (Filter (FBool false))
+              | (Filter filter) as f ->
+                List.iter warn_undefined (OpamFilter.variables filter);
+                Atom f
+              | f -> Atom f)
+            fc
+        in
+        Atom (name, fc))
+  in
+  OpamPackage.Map.mapi (fun nv opam ->
+      OpamFilter.partial_filter_formula
+        (package_env_t st ~force_dev_deps ~test ~doc
+           ~dev_setup ~requested_allpkgs ~err_undefined:false nv)
+        (deps opam)
+      |> filter_undefined nv) opams
 
 let universe st
     ?(test=OpamStateConfig.(!r.build_test))
     ?(doc=OpamStateConfig.(!r.build_doc))
+    ?(dev_setup=OpamStateConfig.(!r.dev_setup))
     ?(force_dev_deps=false)
     ?reinstall
     ~requested
     user_action =
+  let chrono = OpamConsole.timer () in
+  let names = OpamPackage.names_of_packages requested in
   let requested_allpkgs =
-    OpamPackage.packages_of_names st.packages requested
+    OpamPackage.packages_of_names st.packages names
   in
-  let env nv v =
-    if List.mem v OpamPackageVar.predefined_depends_variables then
-      match OpamVariable.Full.to_string v with
-      | "dev" ->
-        Some (B (force_dev_deps || is_dev_package st nv))
-      | "with-test" ->
-        Some (B (test && OpamPackage.Set.mem nv requested_allpkgs))
-      | "with-doc" ->
-        Some (B (doc && OpamPackage.Set.mem nv requested_allpkgs))
-      | _ -> None (* Computation delayed to the solver *)
-    else
-    let r = OpamPackageVar.resolve_switch ~package:nv st v in
-    if r = None then
-      (if OpamFormatConfig.(!r.strict) then
-         OpamConsole.error_and_exit `File_error
-           "undefined filter variable in dependencies of %s: %s"
-       else
-         log
-           "ERR: undefined filter variable in dependencies of %s: %s")
-        (OpamPackage.to_string nv) (OpamVariable.Full.to_string v);
-    r
+  let env =
+    package_env_t st
+      ~force_dev_deps ~test ~doc ~dev_setup
+      ~requested_allpkgs
   in
-  let get_deps f opams =
-    OpamPackage.Map.mapi (fun nv opam ->
-        OpamFilter.partial_filter_formula (env nv) (f opam)
-      ) opams
+  let get_deps =
+    get_dependencies_t st
+      ~force_dev_deps ~test ~doc ~dev_setup
+      ~requested_allpkgs
   in
   let u_depends =
     let depend =
@@ -899,7 +988,6 @@ let universe st
   in
   let u_depopts = get_deps OpamFile.OPAM.depopts st.opams in
   let u_conflicts = get_conflicts st st.packages st.opams in
-  let base = st.compiler_packages in
   let u_invariant =
     if OpamStateConfig.(!r.unlock_base) then OpamFormula.Empty
     else st.switch_invariant
@@ -939,12 +1027,7 @@ let universe st
       OpamPackage.Set.empty
   in
   let avoid_versions =
-    OpamPackage.Map.fold (fun nv opam acc ->
-        if OpamFile.OPAM.has_flag Pkgflag_AvoidVersion opam &&
-           not (can_upgrade_to_avoid_version (OpamFile.OPAM.name opam) st)
-        then OpamPackage.Set.add nv acc else acc)
-      st.opams
-      OpamPackage.Set.empty
+    OpamPackage.Set.filter (avoid_version st) u_available
   in
   let u =
 {
@@ -957,14 +1040,14 @@ let universe st
   u_conflicts;
   u_installed_roots = st.installed_roots;
   u_pinned    = OpamPinned.packages st;
-  u_base      = base;
   u_invariant;
   u_reinstall;
-  u_attrs     = ["opam-query", requested_allpkgs;
+  u_attrs     = ["opam-query", requested;
                  "missing-depexts", missing_depexts;
                  "avoid-version", avoid_versions];
 }
   in
+  log ~level:2 "Universe load: %.3fs" (chrono ());
   u
 
 let dump_pef_state st oc =
@@ -974,7 +1057,7 @@ let dump_pef_state st oc =
     Printf.fprintf oc "version: %s\n" (OpamPackage.version_to_string nv);
     let installed = OpamPackage.Set.mem nv st.installed in
     (* let root = OpamPackage.Set.mem nv st.installed_roots in *)
-    let base = OpamPackage.Set.mem nv st.compiler_packages in
+    let inv = OpamPackage.Set.mem nv st.compiler_packages in
     let pinned = OpamPackage.Set.mem nv st.pinned in
     let available = OpamPackage.Set.mem nv (Lazy.force st.available_packages) in
     let reinstall = OpamPackage.Set.mem nv (Lazy.force st.reinstall) in
@@ -983,7 +1066,7 @@ let dump_pef_state st oc =
     Printf.fprintf oc "available: %b\n" available;
     if installed then output_string oc "installed: true\n";
     if pinned then output_string oc "pinned: true\n";
-    if base then output_string oc "base: true\n";
+    if inv then output_string oc "invariant-pkg: true\n";
     if reinstall then output_string oc "reinstall: true\n";
 
     (* metadata (resolved for the current switch) *)
@@ -1161,15 +1244,28 @@ let update_pin nv opam st =
     OpamStd.Option.default nv.version (OpamFile.OPAM.version_opt opam)
   in
   let nv = OpamPackage.create nv.name version in
-  update_package_metadata nv opam @@
-  { st with
-    pinned =
-      OpamPackage.Set.add nv
-        (OpamPackage.filter_name_out st.pinned nv.name);
-    available_packages = lazy (
-      OpamPackage.filter_name_out (Lazy.force st.available_packages) nv.name
-    );
-  }
+  let pinned =
+    OpamPackage.Set.add nv (OpamPackage.filter_name_out st.pinned nv.name)
+  in
+  let available_packages = lazy (
+    OpamPackage.filter_name_out (Lazy.force st.available_packages) nv.name
+  ) in
+  let st =
+    update_package_metadata nv opam { st with pinned; available_packages }
+  in
+  if not (OpamFile.Config.depext st.switch_global.config)
+  || OpamSysPkg.Set.is_empty (depexts st nv)
+  then st else
+  let sys_packages = lazy (
+    OpamPackage.Map.union (fun _ n -> n)
+      (Lazy.force st.sys_packages)
+      (depexts_status_of_packages st (OpamPackage.Set.singleton nv))
+  ) in
+  let available_packages = lazy (
+    OpamPackage.Set.filter (fun nv -> depexts_unavailable st nv = None)
+      (Lazy.force st.available_packages)
+  ) in
+  { st with sys_packages; available_packages }
 
 let do_backup lock st = match lock with
   | `Lock_write ->
@@ -1230,3 +1326,168 @@ let update_repositories gt update_fun switch =
   OpamFile.Switch_config.write
     (OpamPath.Switch.switch_config gt.root switch)
     conf
+
+
+(* dependencies computation *)
+
+let dependencies_filter_to_formula_t ~build ~post st nv =
+  let env v =
+    if List.mem v OpamPackageVar.predefined_depends_variables then
+      match OpamVariable.Full.to_string v with
+      | "build" -> Some (B build)
+      | "post" -> Some (B post)
+      | _ -> None
+    else
+      OpamPackageVar.resolve_switch ~package:nv st v
+  in
+  OpamFilter.filter_formula ~default:true env
+
+let dependencies_t st base_deps_compute deps_compute
+    ~depopts ~installed ?(unavailable=false) packages =
+  if OpamPackage.Set.is_empty packages then OpamPackage.Set.empty else
+  let base =
+    packages ++
+    if installed then st.installed
+    else if unavailable then st.packages
+    else Lazy.force st.available_packages
+  in
+  log ~level:3 "dependencies packages=%a"
+    (slog OpamPackage.Set.to_string) packages;
+  let timer = OpamConsole.timer () in
+  let base_depends =
+    let filter = base_deps_compute base in
+    let get_deps =
+      get_dependencies_t st
+        ~force_dev_deps:false ~test:false ~doc:false
+        ~dev_setup:false ~requested_allpkgs:packages
+    in
+    let opams =
+      OpamPackage.Set.fold (fun pkg opams ->
+          OpamPackage.Map.add pkg (OpamPackage.Map.find pkg st.opams) opams)
+        base OpamPackage.Map.empty
+    in
+    let u_depends = get_deps OpamFile.OPAM.depends opams in
+    let depends =
+      OpamPackage.Map.filter_map filter u_depends
+    in
+    if depopts then
+      let u_depopts = get_deps OpamFile.OPAM.depopts opams in
+      let depopts =
+        OpamPackage.Map.filter_map filter u_depopts
+      in
+      OpamPackage.Map.union (fun d d' -> OpamFormula.And (d, d'))
+        depopts depends
+    else
+      depends
+  in
+  let result = deps_compute base base_depends packages in
+  log "dependencies (%.3f) result=%a" (timer ())
+    (slog OpamPackage.Set.to_string) result;
+  result
+
+let dependencies st ~build ~post =
+  dependencies_t st
+    (fun base nv ff ->
+       if OpamPackage.Set.mem nv base then Some ff else None)
+    (fun base base_depends packages ->
+       let open OpamPackage.Set.Op in
+       let get_deps nvs =
+         OpamPackage.Set.fold (fun nv set ->
+             let depends_formula =
+               dependencies_filter_to_formula_t ~build ~post st nv
+                 (OpamPackage.Map.find nv base_depends)
+             in
+             if depends_formula = Empty then set else
+             let deps = OpamFormula.packages base depends_formula in
+             if OpamPackage.Set.is_empty deps then set else
+               deps ++ set)
+           nvs OpamPackage.Set.empty
+       in
+       let rec aux all deps =
+         let new_deps = get_deps deps in
+         if OpamPackage.Set.is_empty new_deps then all
+         else aux (all ++ new_deps) (new_deps -- all)
+       in
+       aux packages packages)
+
+let reverse_dependencies st ~build ~post =
+  dependencies_t st
+    (fun base nv ff ->
+       if OpamPackage.Set.mem nv base then
+         Some (dependencies_filter_to_formula_t ~build ~post st nv ff)
+       else None)
+    (fun base base_depends packages ->
+       let base_int_pkg =
+         OpamPackage.Set.fold (fun nv map ->
+             OpamStd.IntMap.add (Hashtbl.hash nv) nv map)
+           base OpamStd.IntMap.empty
+       in
+       let rev_deps =
+         OpamPackage.Map.fold (fun nv depends_formula rev_deps ->
+             let depends =
+               OpamPackage.Set.fold (fun nv deps ->
+                   Hashtbl.hash nv :: deps)
+                 (OpamFormula.packages base depends_formula) []
+             in
+             List.fold_left (fun rev_deps rev_nv ->
+                 OpamStd.IntMap.update rev_nv
+                   (fun l -> Hashtbl.hash nv :: l)
+                   [] rev_deps)
+               rev_deps depends)
+           base_depends OpamStd.IntMap.empty
+       in
+       let open OpamStd.IntSet.Op in
+       let get_revdeps all done_ packages =
+         OpamStd.IntSet.fold (fun nv (all, done_, remaining) ->
+             if OpamStd.IntSet.mem nv done_ then
+               all, done_, remaining
+             else
+             match OpamStd.IntMap.find_opt nv rev_deps with
+             | Some deps ->
+               let deps = (OpamStd.IntSet.of_list deps) in
+               deps ++ all,
+               OpamStd.IntSet.add nv done_,
+               deps ++ remaining
+             | None -> all, done_, remaining)
+           packages (all, done_, OpamStd.IntSet.empty)
+       in
+       let rec aux all done_ remaining =
+         let all, done_, remaining =
+           get_revdeps all done_ remaining
+         in
+         if OpamStd.IntSet.is_empty remaining then all else
+           aux all done_ remaining
+       in
+       let int_revdeps =
+         let packages =
+           OpamPackage.Set.fold (fun nv set ->
+               OpamStd.IntSet.add (Hashtbl.hash nv) set)
+             packages OpamStd.IntSet.empty
+         in
+         aux OpamStd.IntSet.empty OpamStd.IntSet.empty packages
+       in
+       OpamStd.IntSet.fold (fun hash result ->
+           match OpamStd.IntMap.find_opt hash base_int_pkg with
+           | Some nv -> OpamPackage.Set.add nv result
+           | None -> OpamStd.Sys.exit_because `Internal_error)
+         int_revdeps packages)
+
+(* invariant computation *)
+
+let invariant_root_packages st =
+  OpamPackage.Set.filter (OpamFormula.verifies st.switch_invariant) st.installed
+
+let compute_invariant_packages st =
+  let pkgs = invariant_root_packages st in
+  dependencies ~build:false ~post:true ~depopts:false ~installed:true
+    ~unavailable:false st pkgs
+
+let compiler_packages st =
+  let compiler_packages =
+    OpamPackage.Set.filter (fun nv ->
+        try OpamFile.OPAM.has_flag Pkgflag_Compiler (opam st nv)
+        with Not_found -> false)
+      st.installed
+  in
+  dependencies ~build:true ~post:false ~depopts:true ~installed:true
+    ~unavailable:false st compiler_packages
