@@ -50,15 +50,15 @@ let eval_redirect gt repo repo_root =
 let repository rt repo =
   let max_loop = 10 in
   let gt = rt.repos_global in
-  if repo.repo_url = OpamUrl.empty then Done (fun rt -> rt) else
+  if repo.repo_url = OpamUrl.empty then Done None else
   let repo_root = OpamRepositoryState.get_repo_root rt repo in
   (* Recursively traverse redirection links, but stop after 10 steps or if
      we cycle back to the initial repo. *)
-  let rec job r n =
+  let rec job r redirect n =
     if n = 0 then
       (OpamConsole.warning "%s: Too many redirections, stopping."
          (OpamRepositoryName.to_string repo.repo_name);
-       Done r)
+       Done (r, `No_changes))
     else
     let text =
       OpamProcess.make_command_text ~color:`blue
@@ -66,13 +66,14 @@ let repository rt repo =
         OpamUrl.(string_of_backend repo.repo_url.backend)
     in
     OpamProcess.Job.with_text text @@
-    OpamRepository.update r repo_root @@+ fun () ->
+    OpamRepository.update r repo_root @@+ fun has_changes ->
+    let has_changes = if redirect then `Changes else has_changes in
     if n <> max_loop && r = repo then
       (OpamConsole.warning "%s: Cyclic redirections, stopping."
          (OpamRepositoryName.to_string repo.repo_name);
-       Done r)
+       Done (r, has_changes))
     else match eval_redirect gt r repo_root with
-      | None -> Done r
+      | None -> Done (r, has_changes)
       | Some (new_url, f) ->
         OpamFilename.cleandir repo_root;
         let reason = match f with
@@ -84,9 +85,9 @@ let repository rt repo =
           (OpamConsole.colorise `bold "permanently")
           (OpamUrl.to_string new_url)
           reason;
-        job { r with repo_url = new_url } (n-1)
+        job { r with repo_url = new_url } true (n-1)
   in
-  job repo max_loop @@+ fun repo ->
+  job repo false max_loop @@+ fun (repo, has_changes) ->
   let repo_file_path = OpamRepositoryPath.repo repo_root in
   if not (OpamFile.exists repo_file_path) then
     OpamConsole.warning
@@ -94,58 +95,69 @@ let repository rt repo =
        compatible with this version of opam."
       (OpamRepositoryName.to_string repo.repo_name)
       (OpamUrl.to_string repo.repo_url);
-  let repo_file = OpamFile.Repo.safe_read repo_file_path in
-  let repo_file = OpamFile.Repo.with_root_url repo.repo_url repo_file in
-  let repo_vers =
-    OpamStd.Option.default OpamFile.Repo.format_version @@
-    OpamFile.Repo.opam_version repo_file
-  in
-  if not OpamFormatConfig.(!r.skip_version_checks) &&
-     OpamVersion.compare repo_vers OpamVersion.current > 0 then
-    Printf.ksprintf failwith
-      "repository format version is %s, and this is only opam %s"
-      (OpamVersion.to_string repo_vers)
-      (OpamVersion.to_string OpamVersion.current);
-  List.iter (fun (msg, filter) ->
-      if OpamFilter.opt_eval_to_bool (OpamPackageVar.resolve_global gt) filter
-      then
-        OpamConsole.formatted_msg ~indent:4 "%s (at %s): %s\n"
-          (OpamConsole.colorise' [`bold;`green]
-             (OpamRepositoryName.to_string repo.repo_name))
-          (OpamConsole.colorise `bold (OpamUrl.to_string repo.repo_url))
-          msg)
-    (OpamFile.Repo.announce repo_file);
-  OpamFilename.make_tar_gz_job
-    (OpamRepositoryPath.tar gt.root repo.repo_name)
-    repo_root
-  @@+ function
-  | Some e ->
-    OpamStd.Exn.fatal e;
-    Printf.ksprintf failwith
-      "Failed to regenerate local repository archive: %s"
-      (Printexc.to_string e)
-  | None ->
-    let opams =
-      OpamRepositoryState.load_opams_from_dir repo.repo_name repo_root
+  match has_changes with
+  | `No_changes ->
+    log "Repository did not change: nothing to do.";
+    Done None
+  | `Changes ->
+    log "Repository has new changes";
+    let repo_file = OpamFile.Repo.safe_read repo_file_path in
+    let repo_file = OpamFile.Repo.with_root_url repo.repo_url repo_file in
+    let repo_vers =
+      OpamStd.Option.default OpamFile.Repo.format_version @@
+      OpamFile.Repo.opam_version repo_file
     in
-    let local_dir = OpamRepositoryPath.root gt.root repo.repo_name in
-    if OpamFilename.exists_dir local_dir then
-      (* Mark the obsolete local directory for deletion once we complete: it's
-         no longer needed once we have a tar.gz *)
-      Hashtbl.add rt.repos_tmp repo.repo_name (lazy local_dir);
-    Done (
-      (* Return an update function to make parallel execution possible *)
-      fun rt ->
-        { rt with
-          repositories =
-            OpamRepositoryName.Map.add repo.repo_name repo rt.repositories;
-          repos_definitions =
-            OpamRepositoryName.Map.add repo.repo_name repo_file
-              rt.repos_definitions;
-          repo_opams =
-            OpamRepositoryName.Map.add repo.repo_name opams rt.repo_opams;
-        }
-    )
+    if not OpamFormatConfig.(!r.skip_version_checks) &&
+       OpamVersion.compare repo_vers OpamVersion.current > 0 then
+      Printf.ksprintf failwith
+        "repository format version is %s, and this is only opam %s"
+        (OpamVersion.to_string repo_vers)
+        (OpamVersion.to_string OpamVersion.current);
+    List.iter (fun (msg, filter) ->
+        if OpamFilter.opt_eval_to_bool (OpamPackageVar.resolve_global gt) filter
+        then
+          OpamConsole.formatted_msg ~indent:4 "%s (at %s): %s\n"
+            (OpamConsole.colorise' [`bold;`green]
+               (OpamRepositoryName.to_string repo.repo_name))
+            (OpamConsole.colorise `bold (OpamUrl.to_string repo.repo_url))
+            msg)
+      (OpamFile.Repo.announce repo_file);
+    let tarred_repo = OpamRepositoryPath.tar gt.root repo.repo_name in
+    (if OpamRepositoryConfig.(!r.repo_tarring) then
+       OpamFilename.make_tar_gz_job tarred_repo repo_root
+     else Done None)
+    @@+ function
+    | Some e ->
+      OpamStd.Exn.fatal e;
+      Printf.ksprintf failwith
+        "Failed to regenerate local repository archive: %s"
+        (Printexc.to_string e)
+    | None ->
+      let opams =
+        OpamRepositoryState.load_opams_from_dir repo.repo_name repo_root
+      in
+      let local_dir = OpamRepositoryPath.root gt.root repo.repo_name in
+      if OpamRepositoryConfig.(!r.repo_tarring) then
+        (if OpamFilename.exists_dir local_dir then
+           (* Mark the obsolete local directory for deletion once we complete: it's
+              no longer needed once we have a tar.gz *)
+           Hashtbl.add rt.repos_tmp repo.repo_name (lazy local_dir))
+      else if OpamFilename.exists tarred_repo then
+        (OpamFilename.move_dir ~src:repo_root ~dst:local_dir;
+         OpamFilename.remove tarred_repo);
+      Done (Some (
+          (* Return an update function to make parallel execution possible *)
+          fun rt ->
+            { rt with
+              repositories =
+                OpamRepositoryName.Map.add repo.repo_name repo rt.repositories;
+              repos_definitions =
+                OpamRepositoryName.Map.add repo.repo_name repo_file
+                  rt.repos_definitions;
+              repo_opams =
+                OpamRepositoryName.Map.add repo.repo_name opams rt.repo_opams;
+            }
+        ))
 
 let repositories rt repos =
   let command repo =
@@ -155,36 +167,51 @@ let repositories rt repos =
          OpamConsole.error "Could not update repository %S: %s"
            (OpamRepositoryName.to_string repo.repo_name)
            (match ex with Failure s -> s | ex -> Printexc.to_string ex);
-         Done ([repo], fun t -> t)) @@
+         Done ([repo], None)) @@
     fun () -> repository rt repo @@|
     fun f -> [], f
   in
+  let merge (failed1, f1) (failed2, f2) =
+    failed1 @ failed2,
+    match f1, f2 with
+    | None, None -> None
+    | Some f1, Some f2 -> Some (f1 @* f2)
+    | Some f, None | None, Some f -> Some f
+  in
   let failed, rt_update =
     OpamParallel.reduce
       ~jobs:OpamStateConfig.(!r.dl_jobs)
-      ~command
-      ~merge:(fun (failed1, f1) (failed2, f2) -> failed1 @ failed2, f1 @* f2)
-      ~nil:([], fun x -> x)
+      ~command ~merge
+      ~nil:([], None)
       ~dry_run:OpamStateConfig.(!r.dryrun)
       repos
   in
-  let rt = rt_update rt in
-  OpamRepositoryState.write_config rt;
-  OpamRepositoryState.Cache.save rt;
+  let rt =
+    match rt_update with
+    | Some rt_update ->
+      let rt = rt_update rt in
+      OpamRepositoryState.write_config rt;
+      OpamRepositoryState.Cache.save rt;
+      rt
+    | None -> rt
+  in
   failed, rt
 
 let fetch_dev_package url srcdir ?(working_dir=false) ?subpath nv =
   let remote_url = OpamFile.URL.url url in
   let mirrors = remote_url :: OpamFile.URL.mirrors url in
   let checksum = OpamFile.URL.checksum url in
-  log "updating %a%a" (slog OpamUrl.to_string) remote_url
-    (slog (OpamStd.Option.map_default (fun s -> " ("^s^")") "")) subpath;
+  log "updating %a" (slog (OpamUrl.to_string_w_subpath subpath)) remote_url;
+(*
+    (slog (OpamStd.Option.to_string OpamFilename.SubPath.pretty_string))
+    subpath;
+*)
   OpamRepository.pull_tree
     ~cache_dir:(OpamRepositoryPath.download_cache OpamStateConfig.(!r.root_dir))
     (OpamPackage.to_string nv) srcdir checksum ~working_dir ?subpath mirrors
   @@| OpamRepository.report_fetch_result nv
 
-let pinned_package st ?version ?(working_dir=false) name =
+let pinned_package st ?version ?(autolock=false) ?(working_dir=false) name =
   log "update-pinned-package %s%a" (OpamPackage.Name.to_string name)
     (slog @@ function true -> " (working dir)" | false -> "") working_dir;
   let open OpamStd.Option.Op in
@@ -211,19 +238,17 @@ let pinned_package st ?version ?(working_dir=false) name =
       then OpamFileTools.add_aux_files ~files_subdir_hashes:true opam
       else opam
     in
+    let locked = if autolock then OpamFile.OPAM.locked opam else None in
     (* append subpath to source dir to retrieve opam files *)
-    let srcdir_find =
-      OpamStd.Option.map_default
-        (fun x -> OpamFilename.Op.(srcdir / x)) srcdir subpath
-    in
+    let srcdir_find = OpamFilename.SubPath.(srcdir /? subpath) in
     let old_source_opam_hash, old_source_opam =
-      match OpamPinned.find_opam_file_in_source name srcdir_find with
+      match OpamPinned.find_opam_file_in_source ?locked name srcdir_find with
       | None -> None, None
-      | Some f ->
+      | Some (f, lock) ->
         Some (OpamHash.compute (OpamFile.to_string f)),
         try
           Some (OpamFile.OPAM.read f |> OpamFile.OPAM.with_name name |>
-                add_extra_files srcdir f)
+                add_extra_files srcdir f |> OpamFile.OPAM.with_locked_opt lock)
         with e -> OpamStd.Exn.fatal e; None
     in
     let repo_opam =
@@ -248,14 +273,17 @@ let pinned_package st ?version ?(working_dir=false) name =
          OpamRepository.is_dirty ?subpath url
          @@| function false -> () | true ->
            OpamConsole.note
-             "Ignoring uncommitted changes in %s%s (`--working-dir' not active)."
+             "Ignoring uncommitted changes in %s%s \
+              (`--working-dir' not specified or specified with no argument)."
              url.OpamUrl.path
-             (match subpath with None -> "" | Some s -> "/" ^ s))
+             (OpamStd.Option.to_string (fun sp ->
+                  Filename.dir_sep ^ OpamFilename.SubPath.to_string sp) subpath))
     @@+ fun () ->
     (* Do the update *)
     fetch_dev_package urlf srcdir ~working_dir ?subpath nv @@+ fun result ->
     let new_source_opam =
-      OpamPinned.find_opam_file_in_source name srcdir_find >>= fun f ->
+      OpamPinned.find_opam_file_in_source ?locked name srcdir_find
+      >>= fun (f, lock) ->
       let warns, opam_opt = OpamFileTools.lint_file f in
       let warns, opam_opt = match opam_opt with
         | Some opam0 ->
@@ -277,6 +305,7 @@ let pinned_package st ?version ?(working_dir=false) name =
          OpamConsole.errmsg "%s\n"
            (OpamFileTools.warns_to_string warns));
       opam_opt >>| OpamFile.OPAM.with_name name >>| add_extra_files srcdir f
+      >>| OpamFile.OPAM.with_locked_opt lock
     in
     let equal_opam a b =
       let cleanup_opam o =
@@ -285,7 +314,7 @@ let pinned_package st ?version ?(working_dir=false) name =
              Some ((OpamFile.OPAM.version_opt o)
                    +! (OpamSwitchState.get_package st name |> OpamPackage.version))
            with Not_found -> None)
-          +! (OpamPackage.Version.of_string "~dev")
+          +! OpamPackage.Version.default
         in
         o |> OpamFile.OPAM.with_url_opt None
         |> OpamFile.OPAM.with_version v
@@ -386,11 +415,11 @@ let pinned_package st ?version ?(working_dir=false) name =
     | Result  _, _ ->
       Done ((fun st -> st), true)
 
-let dev_package st ?working_dir nv =
+let dev_package st ?autolock ?working_dir nv =
   log "update-dev-package %a" (slog OpamPackage.to_string) nv;
   if OpamSwitchState.is_pinned st nv.name &&
      not (OpamSwitchState.is_version_pinned st nv.name) then
-    pinned_package st ~version:nv.version ?working_dir nv.name
+    pinned_package st ?autolock ~version:nv.version ?working_dir nv.name
   else
   match OpamSwitchState.url st nv with
   | None     -> Done ((fun st -> st), false)
@@ -398,18 +427,19 @@ let dev_package st ?working_dir nv =
     if (OpamFile.URL.url url).OpamUrl.backend = `http then
       Done ((fun st -> st), false)
     else
-      fetch_dev_package url (OpamSwitchState.source_dir st nv) ?working_dir nv
+      fetch_dev_package url (OpamSwitchState.source_dir st nv)
+        ?subpath:(OpamFile.URL.subpath url) ?working_dir nv
       @@| fun result ->
       (fun st -> st), match result with Result () -> true | _ -> false
 
-let dev_packages st ?(working_dir=OpamPackage.Set.empty) packages =
+let dev_packages st ?autolock ?(working_dir=OpamPackage.Set.empty) packages =
   log "update-dev-packages";
   let command nv =
     let working_dir = OpamPackage.Set.mem nv working_dir in
     OpamProcess.Job.ignore_errors
       ~default:(false, (fun st -> st), OpamPackage.Set.empty)
     @@ fun () ->
-    dev_package st ~working_dir nv @@| fun (st_update, changed) ->
+    dev_package st ?autolock ~working_dir nv @@| fun (st_update, changed) ->
     true, st_update, match changed with
     | true -> OpamPackage.Set.singleton nv
     | false -> OpamPackage.Set.empty
@@ -440,14 +470,14 @@ let dev_packages st ?(working_dir=OpamPackage.Set.empty) packages =
       selections1;
   success, st, updated_set
 
-let pinned_packages st ?(working_dir=OpamPackage.Name.Set.empty) names =
+let pinned_packages st ?autolock ?(working_dir=OpamPackage.Name.Set.empty) names =
   log "update-pinned-packages";
   let command name =
     let working_dir = OpamPackage.Name.Set.mem name working_dir in
     OpamProcess.Job.ignore_errors
       ~default:((fun st -> st), OpamPackage.Name.Set.empty)
     @@ fun () ->
-    pinned_package st ~working_dir name @@| fun (st_update, changed) ->
+    pinned_package st ?autolock ~working_dir name @@| fun (st_update, changed) ->
     st_update,
     match changed with
     | true -> OpamPackage.Name.Set.singleton name
@@ -474,31 +504,39 @@ let pinned_packages st ?(working_dir=OpamPackage.Name.Set.empty) names =
   OpamSwitchAction.add_to_reinstall st ~unpinned_only:false updates,
   updates
 
-let active_caches st nv =
+let active_caches st nvs =
   let global_cache = OpamFile.Config.dl_cache st.switch_global.config in
   let rt = st.switch_repos in
+  let repos_list = OpamSwitchState.repos_list st in
   let repo_cache =
-    match OpamRepositoryState.find_package_opt rt
-            (OpamSwitchState.repos_list st) nv
-    with
-    | None -> []
-    | Some (repo, _) ->
-      let repo_def = OpamRepositoryName.Map.find repo rt.repos_definitions in
-      let root_url = match OpamFile.Repo.root_url repo_def with
-        | None -> OpamSystem.internal_error "repo file of unknown origin"
-        | Some u -> u
-      in
-      OpamStd.List.filter_map (fun rel ->
-          if OpamStd.String.contains ~sub:"://" rel
-          then
-            let r = OpamUrl.parse_opt ~handle_suffix:false rel in
-            if r = None then
-              OpamConsole.warning "Invalid cache url %s, skipping" rel;
-            r
-          else Some OpamUrl.Op.(root_url / rel))
-        (OpamFile.Repo.dl_cache repo_def)
+    List.fold_left (fun (repos, caches as acc) nv ->
+        match OpamRepositoryState.find_package_opt rt repos_list nv with
+        | None -> acc
+        | Some (repo, _) ->
+          if List.mem repo repos then acc else
+          let repo_def = OpamRepositoryName.Map.find repo rt.repos_definitions in
+          let root_url = match OpamFile.Repo.root_url repo_def with
+            | None -> OpamSystem.internal_error "repo file of unknown origin"
+            | Some u -> u
+          in
+          let cache =
+            OpamStd.List.filter_map (fun rel ->
+                if OpamStd.String.contains ~sub:"://" rel
+                then
+                  let r = OpamUrl.parse_opt ~handle_suffix:false rel in
+                  if r = None then
+                    OpamConsole.warning "Invalid cache url %s, skipping" rel;
+                  r
+                else Some OpamUrl.Op.(root_url / rel))
+              (OpamFile.Repo.dl_cache repo_def)
+          in
+          repo::repos, cache::caches)
+      ([],[]) nvs
+    |> snd
+    |> List.rev
+    |> List.flatten
   in
-  repo_cache @ global_cache
+  global_cache @ repo_cache
 
 let cleanup_source st old_opam_opt new_opam =
   let open OpamStd.Option.Op in
@@ -520,37 +558,66 @@ let cleanup_source st old_opam_opt new_opam =
     OpamFilename.rmdir
       (OpamSwitchState.source_dir st (OpamFile.OPAM.package new_opam))
 
-let download_package_source st nv dirname =
-  let opam = OpamSwitchState.opam st nv in
+let download_package_source_t st url nv_dirs =
   let cache_dir = OpamRepositoryPath.download_cache st.switch_global.root in
-  let cache_urls = active_caches st nv in
-
+  let cache_urls = active_caches st (List.map fst nv_dirs) in
   let fetch_source_job =
-    match OpamFile.OPAM.url opam with
-    | None   -> Done None
-    | Some u ->
-      (OpamRepository.pull_tree (OpamPackage.to_string nv)
-        ~cache_dir ~cache_urls ?subpath:(OpamFile.URL.subpath u)
-        dirname
-        (OpamFile.URL.checksum u)
-        (OpamFile.URL.url u :: OpamFile.URL.mirrors u))
-      @@| fun r -> Some r
+    match url with
+    | None -> Done None
+    | Some url ->
+      let dirnames =
+        List.map (fun (nv, dir) ->
+            OpamPackage.to_string nv, dir,
+            OpamStd.Option.Op.(OpamSwitchState.opam st nv
+                               |> OpamFile.OPAM.url
+                               >>= OpamFile.URL.subpath))
+          nv_dirs
+      in
+      let checksums = OpamFile.URL.checksum url in
+      (OpamRepository.pull_shared_tree ~cache_dir ~cache_urls
+         dirnames checksums
+         (OpamFile.URL.url url :: OpamFile.URL.mirrors url))
+      @@+ function
+      | Not_available (_s,_l) as source_result
+        when OpamFile.Config.swh_fallback st.switch_global.config ->
+        (OpamDownload.SWHID.archive_fallback url dirnames
+         @@| function
+         | Result None -> Some source_result
+         | Result (Some r) -> Some (Result r)
+         | Not_available (s,l) -> Some (Not_available (s,l))
+         | Up_to_date _ -> assert false)
+      | source_result -> Done (Some source_result)
   in
-  let fetch_extra_source_job (name, u) = function
-    | (_, Not_available _) :: _ as err -> Done err
+  let fetch_extra_source_job (nv, name, u) = function
+    | (_, _, Not_available _) :: _ as err -> Done err
     | ret ->
       (OpamRepository.pull_file_to_cache
          (OpamPackage.to_string nv ^"/"^ OpamFilename.Base.to_string name)
          ~cache_dir ~cache_urls
          (OpamFile.URL.checksum u)
          (OpamFile.URL.url u :: OpamFile.URL.mirrors u))
-      @@| fun r -> (OpamFilename.Base.to_string name, r) :: ret
+      @@| fun r -> (nv, OpamFilename.Base.to_string name, r) :: ret
   in
   fetch_source_job @@+ function
   | Some (Not_available _) as r -> Done (r, [])
   | r ->
     OpamProcess.Job.seq
       (List.map fetch_extra_source_job
-         (OpamFile.OPAM.extra_sources opam))
+         (List.flatten @@ List.map (fun (nv,_) ->
+              List.map (fun (n,u) -> nv, n, u)
+                (OpamFile.OPAM.extra_sources (OpamSwitchState.opam st nv)))
+             nv_dirs))
       []
     @@| fun r1 -> r, r1
+
+let download_shared_package_source st url nvs =
+  download_package_source_t st url
+    (List.map (fun nv -> nv, OpamSwitchState.source_dir st nv) nvs)
+
+let download_package_source st nv dirname =
+  download_package_source_t st
+    (OpamFile.OPAM.url (OpamSwitchState.opam st nv))
+    [nv, dirname]
+  @@| fun (sources, extra_sources) ->
+  sources,
+  List.map (fun (_nv, name, failure) -> name, failure) extra_sources
