@@ -10,8 +10,8 @@ let default_repo = "opam-repo-"^null_hash
 let diff_rule base_name ~condition =
   Format.sprintf
     {|
-(alias
- (name reftest-%s)%s
+(rule
+ (alias reftest-%s)%s
  (action
   (diff %s.test %s.out)))
 
@@ -35,17 +35,18 @@ let run_rule ~base_name ~archive_hash ~condition =
 (rule
  (targets %s)
  (deps %s)%s
+ (package opam)
  (action
   (with-stdout-to
    %%{targets}
-   (run ./run.exe %%{bin:opam} %%{dep:%s.test} %%{read-lines:testing-env}))))
+   (run ./run.exe %%{exe:../../src/client/opamMain.exe.exe} %%{dep:%s.test} %%{read-lines:testing-env}))))
 |} (base_name^".out") (opamroot_directory ~archive_hash) condition base_name
 
 let archive_download_rule archive_hash =
    Format.sprintf {|
 (rule
  (targets %s)
- (action (run wget --quiet -O %%{targets} https://github.com/ocaml/opam-repository/archive/%s.tar.gz)))
+ (action (run curl --silent -Lo %%{targets} https://github.com/ocaml/opam-repository/archive/%s.tar.gz)))
 |} (tgz_name ~archive_hash) archive_hash
 
 let default_repo_rule =
@@ -74,6 +75,7 @@ let opam_init_rule archive_hash =
   Format.sprintf {|
 (rule
   (targets %s)
+  (deps opam-root-version)
   (action
    (progn
     (ignore-stdout
