@@ -36,14 +36,18 @@
        (`foo'bar'` is not translated to `foobar`)
      * Variable expansion in arguments (`$FOO` or `${FOO}`). Undefined variables
        are left as-is
-     * rewrites: `| 'REGEXP' -> 'STR'` (can be repeated; set `STR` to `\c` to
-       clear the line)
-     * `| grep REGEXP`
-     * `| grep -v REGEXP`
-     * `| unordered` compares lines without considering their ordering
+     * rewrites:
+       * `| 'REGEXP' -> 'STR'` (can be repeated; set `STR` to `\c` to
+         clear the line)
+       * `| grep REGEXP`
+       * `| grep -v REGEXP`
+       * `| unordered` compares lines without considering their ordering
+       * `| sed-cmd command` replaces full path resolved command by `command`
      * variables from command outputs: `cmd args >$ VAR`
      * `### : comment`
-     * opam-cat: prints a nromalised opam file
+     * `opam-cat file`: prints a normalised opam file
+     * `json-cat file`: print a human readable opam output json file, with
+       replacement of some duration and temporary files names
    - if you need more shell power, create a script using <FILENAME> then run it.
      Or just use `sh -c`... but beware for compatibility.
 
@@ -56,6 +60,11 @@ type test = {
   commands: (string * string list) list;
 }
 
+type opam = {
+  as_called : string;
+  as_seen_in_opam : string;
+}
+
 let cmd_prompt = "### "
 let no_opam_repo = "N0REP0"
 let default_repo = "REPO"
@@ -112,6 +121,7 @@ let base_env =
   propagate "TMPDIR" @
   propagate "TMP" @
   propagate "TEMP" @
+  (if Sys.win32 then ["SHELL", "/bin/sh"] else []) @
   [
     "OPAMKEEPBUILDDIR", "1";
     "OPAMCOLOR", "never";
@@ -182,12 +192,15 @@ let filters_of_var =
       Sed x)
 
 let command
-    ?(allowed_codes = [0]) ?(vars=[]) ?(silent=false) ?(filter=[])
+    ?(allowed_codes = [0]) ?(vars=[]) ?(silent=false) ?(filter=[]) ?(sort=false)
     cmd args =
   let env =
     Array.of_list @@
     List.map (fun (var, value) -> Printf.sprintf "%s=%s" var value) @@
-    (base_env @ vars)
+    (List.filter (fun (v,_) ->
+         List.find_opt (fun (v',_) -> String.equal v v') vars = None)
+        base_env)
+    @ vars
   in
   let input, stdout = Unix.pipe () in
   Unix.set_close_on_exec input;
@@ -212,23 +225,30 @@ let command
       Unix.stdin stdout stdout
   in
   Unix.close stdout;
-  let out_buf = Buffer.create 273 in
-  let rec filter_output ?(first=true) ic =
+  let rec filter_output out_buf ic =
     match input_line ic with
     | s ->
-      let s = str_replace_path ~escape:`Unescape OpamSystem.back_to_forward filter s in
-      if s = "\\c" then filter_output ~first ic
+      let s =
+        str_replace_path ~escape:`Unescape OpamSystem.back_to_forward filter s
+      in
+      if s = "\\c" then filter_output out_buf ic
       else
-        (if not first then Buffer.add_char out_buf '\n';
-         Buffer.add_string out_buf s;
-         if not silent then print_endline s;
-         filter_output ~first:false ic)
-    | exception End_of_file -> ()
+        (let out_buf = s::out_buf in
+         if not silent && not sort then
+           print_endline s;
+         filter_output out_buf ic)
+    | exception End_of_file -> out_buf
   in
-  filter_output ic;
+  let out_buf = filter_output [] ic in
   let ret = waitpid pid in
   close_in ic;
-  let out = Buffer.contents out_buf in
+  let out =
+    if sort then List.sort String.compare out_buf
+    else List.rev out_buf
+  in
+  if not silent && sort then
+    List.iter print_endline out;
+  let out = String.concat "\n" out in
   if not (List.mem ret allowed_codes) then
     raise (Command_failure (ret, String.concat " " (cmd :: args), out))
   else
@@ -289,13 +309,16 @@ type command =
   | Pin_file_content of string
   | Cat of { files: string list;
              filter: (Re.t * filt_sort) list; }
+  | Json of { files: string list;
+             filter: (Re.t * filt_sort) list; }
   | Run of { env: (string * string) list;
              cmd: string;
              args: string list; (* still escaped *)
              filter: (Re.t * filt_sort) list;
              output: string option;
-             unordered: bool; }
-  | Export of (string * string) list
+             unordered: bool;
+             sort: bool;}
+  | Export of (string * [`eq | `pluseq | `eqplus] * string) list
   | Comment of string
 
 module Parse = struct
@@ -333,7 +356,7 @@ module Parse = struct
   let re_varbind =
     seq [
       group @@ seq [alpha; rep (alt [alnum; set "_-"])];
-      char '=';
+      group @@ alt [ str "+="; str "=+"; char '=' ];
       group @@ re_str_atom;
       rep space;
     ]
@@ -372,7 +395,14 @@ module Parse = struct
     else
     let varbinds, pos =
       let gr = exec (compile @@ rep re_varbind) str in
-      List.map (fun gr -> Group.get gr 1, get_str (Group.get gr 2))
+      List.map (fun gr ->
+          Group.get gr 1,
+          (match Group.get gr 2 with
+           | "=" -> `eq
+           | "+=" -> `pluseq
+           | "=+" -> `eqplus
+           | _ -> assert false),
+          get_str (Group.get gr 3))
         (all (compile @@ re_varbind) (Group.get gr 0)),
       Group.stop gr 0
     in
@@ -398,43 +428,122 @@ module Parse = struct
         failwith (Printf.sprintf "Bad POSIX regexp: %s" re)
     in
     let rec get_args_rewr acc = function
-      | [] -> List.rev acc, false, [], None
+      | [] -> List.rev acc, false, false, [], None
       | ("|"|">$") :: _ as rewr ->
-        let rec get_rewr (unordered, acc) = function
+        let rec get_rewr (unordered, sort, acc) = function
           | "|" :: re :: "->" :: str :: r ->
-            get_rewr (unordered, (posix_re re, Sed (get_str str)) :: acc) r
+            get_rewr (unordered, sort, (posix_re re, Sed (get_str str)) :: acc) r
           | "|" :: "grep" :: "-v" :: re :: r ->
-            get_rewr (unordered, (posix_re re, GrepV) :: acc) r
+            get_rewr (unordered, sort, (posix_re re, GrepV) :: acc) r
           | "|" :: "grep" :: re :: r ->
-            get_rewr (unordered, (posix_re re, Grep) :: acc) r
+            get_rewr (unordered, sort, (posix_re re, Grep) :: acc) r
+          | "|" :: "sort" :: r ->
+            if acc <> [] then
+              Printf.printf "Warning: sort should appear _before_ any filters\n%!";
+            get_rewr (unordered, true, acc) r
           | "|" :: "unordered" :: r ->
-            get_rewr (true, acc) r
+            get_rewr (true, sort, acc) r
+          | "|" :: "sed-cmd" :: cmd :: r ->
+            let sandbox =
+              (* Sandbox prefix
+                 >[...] /tmp/build_592d92_dune/opam-reftest-2b89f9/OPAM/opam-init/hooks/sandbox.sh "build" "cmd" <
+                 >[...] ${BASEDIR}/opam-init/hooks/sandbox.sh "build" "cmd" <
+                 -->
+                 >[...] cmd <
+              *)
+              seq [
+                alt [ char '/'; Re.str "${" ];
+                non_greedy @@ rep1 any; Re.str "sandbox.sh";
+                space;
+                char '"';
+                alt @@ List.map Re.str [ "build"; "install"; "remove" ];
+                char '"';
+                space;
+                char '"'; Re.str cmd; char '"';
+                space;
+              ]
+            in
+            let with_quote_set s = set ("\"'"^s) in
+            let opt_quoted r = [
+              seq @@ [ char '"'] @  r @ [ char '"'; rep1 space ];
+              seq @@ r @ [ rep1 space ];
+            ] in
+            let unix_prefix =
+              (* Unix & Mac command prefix
+                 >[...] /usr/bin/cmd <
+                 >[...] /usr/bin/cmd <
+                 -->
+                 >[...] cmd <
+              *)
+              opt_quoted @@ [
+                rep1 @@ seq [ char '/'; rep1 @@ diff any (with_quote_set "/ ") ];
+                char '/';
+                Re.str cmd;
+              ]
+            in
+            let win_prefix =
+              (* Windows command prefix
+                 >[...] C:\Windows\system32\cmd.exe <
+                 >[...] C:\Windows\system32\cmd <
+                 >[...] C:\Windows\system 32\cmd <
+                 -->
+                 >[...] cmd <
+              *)
+              opt_quoted @@ [
+                alpha; char ':';
+                rep1 @@ seq [ char '\\'; opt @@ char '\\';
+                              rep1 @@ diff any (with_quote_set "\\") ];
+                char '\\'; opt @@ char '\\';
+                Re.str cmd;
+                opt @@ Re.str ".exe";
+              ] in
+            let re = alt @@ sandbox :: unix_prefix @ win_prefix in
+            let str = Printf.sprintf "%s " cmd in
+            get_rewr (unordered, sort, (re, Sed str) :: acc) r
           | ">$" :: output :: [] ->
-            unordered, List.rev acc, Some (get_str output)
+            unordered, sort, List.rev acc, Some (get_str output)
           | [] ->
-            unordered, List.rev acc, None
+            unordered, sort, List.rev acc, None
           | r ->
             Printf.printf
               "Bad rewrite %S, expecting '| RE -> STR' or '>$ VAR'\n%!"
               (String.concat " " r);
-            unordered, List.rev acc, None
+            unordered, sort, List.rev acc, None
         in
-        let unordered, rewr, out = get_rewr (false, []) rewr in
-        List.rev acc, unordered, rewr, out
+        let unordered, sort, rewr, out = get_rewr (false, false, []) rewr in
+        List.rev acc, unordered, sort, rewr, out
       | arg :: r -> get_args_rewr (arg :: acc) r
     in
-    let args, unordered, rewr, output = get_args_rewr [] args in
+    let args, unordered, sort, rewr, output = get_args_rewr [] args in
     match cmd with
     | Some "opam-cat" ->
       Cat { files = args; filter = rewr; }
+    | Some "json-cat" ->
+      Json { files = args; filter = rewr; }
     | Some cmd ->
+      let env, plus =
+        List.fold_left (fun (env,plus) (v,op,value) ->
+            match op with
+            | `eq -> (v,value)::env, plus
+            | `pluseq -> env, (v^"+="^value)::plus
+            | `eqplus -> env, (v^"=+"^value)::plus)
+          ([],[]) varbinds
+      in
+      (match plus with
+       | [] -> ()
+       | _ ->
+         OpamConsole.error
+           "variable bindings at the beginning of a command does not \
+            support '+=' or '=+' operators: %s"
+           (OpamStd.Format.pretty_list plus));
       Run {
-        env = varbinds;
+        env;
         cmd;
         args;
         filter = rewr;
         output;
         unordered;
+        sort;
       }
     | None ->
       Export varbinds
@@ -443,8 +552,13 @@ end
 let parse_command = Parse.command
 
 let common_filters ?opam dir =
-   let tmpdir = Filename.get_temp_dir_name () in
+   let tmpdir = OpamSystem.real_path (Filename.get_temp_dir_name ()) in
    let open Re in
+   let dir_to_regex dir =
+     if Sys.win32 then
+       [str dir; str (OpamSystem.back_to_forward dir); str (OpamSystem.apply_cygpath dir)]
+     else
+       [str dir] in
    [
      seq [ bol;
            alt [ str "#=== ERROR";
@@ -458,11 +572,10 @@ let common_filters ?opam dir =
      GrepV;
      seq [bol; str cmd_prompt],
      Sed "##% ";
-     alt [str dir; str (OpamSystem.back_to_forward dir)],
+     alt (dir_to_regex dir),
      Sed "${BASEDIR}";
-     seq [opt (str "/private");
-          alt [str tmpdir;
-               str (OpamSystem.back_to_forward tmpdir)];
+     seq [
+          alt (dir_to_regex tmpdir);
           rep (set "/\\");
           str "opam-";
           rep1 (alt [xdigit; char '-'])],
@@ -470,12 +583,12 @@ let common_filters ?opam dir =
    ] @
    (match opam with
     | None -> []
-    | Some opam -> [ str opam, Sed "${OPAM}" ])
+    | Some opam -> [ str opam.as_seen_in_opam, Sed "${OPAM}" ])
 
-let run_cmd ~opam ~dir ?(vars=[]) ?(filter=[]) ?(silent=false) cmd args =
+let run_cmd ~opam ~dir ?(vars=[]) ?(filter=[]) ?(silent=false) ?(sort=false) cmd args =
   let filter = filter @ common_filters ~opam dir in
   let var_filters = filters_of_var vars in
-  let cmd = if cmd = "opam" then opam else cmd in
+  let cmd = if cmd = "opam" then opam.as_called else cmd in
   let args =
     List.map (fun a ->
         let expanded =
@@ -487,7 +600,7 @@ let run_cmd ~opam ~dir ?(vars=[]) ?(filter=[]) ?(silent=false) cmd args =
         Parse.get_str expanded)
       args
   in
-  try command ~vars ~filter ~silent cmd args, None
+  try command ~vars ~filter ~silent ~sort cmd args, None
   with Command_failure (n,_, out) -> out, Some n
 
 let write_file ~path ~contents =
@@ -550,8 +663,8 @@ let print_opamfile file =
     in
     OpamPrinter.FullPos.Normalise.opamfile mangled
   with
-  | Sys_error _ -> Printf.sprintf "# %s not found" file
-  | e -> Printf.sprintf "# Error on file %s: %s" file (Printexc.to_string e)
+  | Sys_error _ -> Printf.sprintf "# %s not found\n" file
+  | e -> Printf.sprintf "# Error on file %s: %s\n" file (Printexc.to_string e)
 
 let template_opamfile =
   OpamParser.FullPos.string {|
@@ -561,17 +674,35 @@ description: "Two words."
 authors: "the testing team"
 homepage: "egapemoh"
 maintainer: "maint@tain.er"
-license: "ISC"
-dev-repo: "hg+htpps://to@lo.ck"
+license: "MIT"
+dev-repo: "hg+https://pkg@op.am"
 bug-reports: "https://nobug"
 |} "<nofile>"
 
+let print_file ~filters reader names =
+  let files =
+    match names with
+    | [file] -> [reader file]
+    | files  ->
+      List.map
+        (fun f -> Printf.sprintf "=> %s <=\n%s" f (reader f))
+        files
+  in
+  List.map (fun s -> OpamStd.String.split s '\n') files
+  |> List.flatten
+  |> List.iter (fun s ->
+      let s =
+        str_replace_path OpamSystem.back_to_forward
+          filters s
+      in
+      if not (String.equal s "\\c") then print_string (s^"\n"))
+
 let run_test ?(vars=[]) ~opam t =
   let old_cwd = Sys.getcwd () in
   let opamroot0 = Filename.concat old_cwd ("root-"^t.repo_hash) in
   with_temp_dir @@ fun dir ->
+  let dir = OpamSystem.real_path dir in
   Sys.chdir dir;
-  let dir = Sys.getcwd () in (* because it may need to be normalised on macOS *)
   let opamroot = Filename.concat dir "OPAM" in
   if Sys.win32 then
     ignore @@ command ~allowed_codes:[0; 1] ~silent:true
@@ -580,18 +711,19 @@ let run_test ?(vars=[]) ~opam t =
   else
     ignore @@ command "cp" ["-PR"; opamroot0; opamroot];
   let vars = [
-    "OPAM", opam;
+    "OPAM", opam.as_seen_in_opam;
     "OPAMROOT", opamroot;
     "BASEDIR", dir;
+    "PATH", Sys.getenv "PATH";
   ] @ vars
   in
   if t.repo_hash = no_opam_repo then
     (mkdir_p (default_repo^"/packages");
      write_file ~path:(default_repo^"/repo") ~contents:{|opam-version: "2.0"|};
-     ignore @@ command opam ~silent:true
+     ignore @@ command opam.as_called ~silent:true
        [ "repository"; "set-url"; "default"; "./"^default_repo;
          "--root"; opamroot]);
-  ignore @@ command ~silent:true opam
+  ignore @@ command ~silent:true opam.as_called
     ["var"; "--quiet"; "--root"; opamroot; "--global"; "--cli=2.1";
      "sys-ocaml-version=4.08.0"];
   print_endline t.repo_hash;
@@ -653,33 +785,81 @@ let run_test ?(vars=[]) ~opam t =
           vars
         | Export bindings ->
           List.fold_left
-            (fun vars (v, r) ->
-               let r =
+            (fun vars -> fun (v, op, r) ->
+               let r' =
                  str_replace_path ~escape:`Backslashes
                    OpamSystem.forward_to_back (filters_of_var vars) r
                in
-               (v, r) :: List.filter (fun (w, _) -> not (String.equal v w)) vars)
+               let value =
+                 match op with
+                 | `eq -> r'
+                 | (`pluseq | `eqplus) as op ->
+                   match List.find_opt (fun (v',_) -> String.equal v v') (base_env @ vars) with
+                   | Some (_,c) ->
+                     let sep = if Sys.win32 then ";" else ":" in
+                     (match op with
+                      | `pluseq -> r'^sep^c
+                      | `eqplus -> c^sep^r')
+                   | None -> r'
+               in
+               (v, value) :: List.filter (fun (w, _) -> not (String.equal v w)) vars)
             vars bindings
         | Cat { files; filter } ->
           let files =
             List.map (fun s -> Re.(replace_string (compile @@ str "$OPAMROOT")
                                      ~by:opamroot s)) files
           in
-          let s =
-            match files with
-            | [file] -> print_opamfile file
-            | files  ->
-              OpamStd.List.concat_map "\n"
-                (fun f -> Printf.sprintf "=> %s <=\n%s" f (print_opamfile f))
-                files
-          in
-          print_string (str_replace_path OpamSystem.back_to_forward
-                          (filter @ common_filters dir) s);
+          print_file ~filters:(filter @ common_filters dir)
+            print_opamfile files;
           vars
-        | Run {env; cmd; args; filter; output; unordered} ->
+        | Json { files; filter } ->
+          let files =
+            List.map (fun s ->
+                Re.(replace_string (compile @@ str "$OPAMROOT") ~by:opamroot s))
+              files
+          in
+          let json_filters =
+            let open Re in
+            let hex_output ext =
+              seq [ char '-'; rep1 digit;
+                    char '-'; repn xdigit 6 (Some 6);
+                    str ext ],
+              Sed ext
+            in
+            [
+              seq [ str {|"duration": |};
+                    rep1 digit; char '.';
+                    rep1 @@ alt [digit ; char 'e'; char '-'] ],
+              Sed {|"duration": 6.2831853071|};
+              hex_output ".env";
+              hex_output ".out";
+              str OpamVersion.(to_string (full ())), Sed "currentv";
+            ]
+          in
+          let to_string f =
+            let ic = open_in f in
+            let content =
+              let rec aux content =
+                try aux (input_line ic :: content)
+                with End_of_file -> content
+              in
+              aux []
+              |> List.rev
+              |> String.concat "\n"
+            in
+            close_in ic;
+            match OpamJson.of_string content with
+            | Some json ->
+              OpamJson.to_string ~minify:false json ^ "\n"
+            | None -> "# Return Error reading json\n"^content
+          in
+          print_file ~filters:(filter @ common_filters ~opam dir @ json_filters)
+            to_string files;
+          vars
+        | Run {env; cmd; args; filter; output; unordered; sort} ->
           let silent = output <> None || unordered in
           let r, errcode =
-            run_cmd ~opam ~dir ~vars:(vars @ env) ~filter ~silent cmd args
+            run_cmd ~opam ~dir ~vars:(vars @ env) ~filter ~silent ~sort cmd args
           in
           (if unordered then
              (* print lines from Result, but respecting order from Expect *)
@@ -715,6 +895,13 @@ let () =
   match Array.to_list Sys.argv with
   | _ :: opam :: input :: env ->
     let opam = OpamFilename.(to_string (of_string opam)) in
+    let opam =
+      (* NOTE: We need that to be able to have the same output from Sys.executable_name when calling the opam binary *)
+      let opam_without_double_exe = Filename.chop_suffix opam ".exe" in
+      if Sys.cygwin
+      then {as_called = opam; as_seen_in_opam = opam_without_double_exe}
+      else {as_called = opam_without_double_exe; as_seen_in_opam = opam_without_double_exe}
+    in
     let vars =
       List.map (fun s -> match OpamStd.String.cut_at s '=' with
           | Some (var, value) -> var, value
